@@ -18,6 +18,8 @@
 #include <NTL/lip.h>
 #include <NTL/tools.h>
 #include <NTL/vector.h>
+#include <NTL/SmartPtr.h>
+#include <NTL/sp_arith.h>
 
 NTL_OPEN_NNS
 
@@ -30,16 +32,23 @@ public:
 typedef ZZ_p residue_type;
 typedef ZZX poly_type;
 
-NTL_verylong rep; // This is currently public for "emergency" situations
-                   // May be private in future versions.
 
 
-ZZ() : rep(0) { }
-// initial value is 0.
+class Deleter {
+public:
+   static void apply(NTL_verylong& p) { NTL_zfree(&p); }
+};
 
-explicit ZZ(long a) : rep(0) { *this = a; }
+WrappedPtr<NTL_verylong_body, Deleter> rep;
+// This is currently public for "emergency" situations
+// May be private in future versions.
 
-ZZ(INIT_SIZE_TYPE, long k) : rep(0)
+ZZ() { }
+
+
+explicit ZZ(long a) { *this = a; }
+
+ZZ(INIT_SIZE_TYPE, long k) 
 // initial value is 0, but space is pre-allocated so that numbers
 // x with x.size() <= k can be stored without re-allocation.
 // Call with ZZ(INIT_SIZE, k).
@@ -51,7 +60,7 @@ ZZ(INIT_SIZE_TYPE, long k) : rep(0)
    NTL_zsetlength(&rep, k); 
 }
 
-ZZ(const ZZ& a) : rep(0)
+ZZ(const ZZ& a) 
 // initial value is a.
 
 {
@@ -59,11 +68,11 @@ ZZ(const ZZ& a) : rep(0)
 }
 
 
-ZZ(INIT_VAL_TYPE, long a) : rep(0) { NTL_zintoz(a, &rep); }
-ZZ(INIT_VAL_TYPE, int a) : rep(0) { NTL_zintoz(a, &rep); }
+ZZ(INIT_VAL_TYPE, long a)  { NTL_zintoz(a, &rep); }
+ZZ(INIT_VAL_TYPE, int a)  { NTL_zintoz(a, &rep); }
 
-ZZ(INIT_VAL_TYPE, unsigned long a) : rep(0) { NTL_zuintoz(a, &rep); }
-ZZ(INIT_VAL_TYPE, unsigned int a) : rep(0) { NTL_zuintoz((unsigned long) a, &rep); }
+ZZ(INIT_VAL_TYPE, unsigned long a)  { NTL_zuintoz(a, &rep); }
+ZZ(INIT_VAL_TYPE, unsigned int a)  { NTL_zuintoz((unsigned long) a, &rep); }
 
 inline ZZ(INIT_VAL_TYPE, const char *);
 inline ZZ(INIT_VAL_TYPE, float);
@@ -75,13 +84,15 @@ ZZ& operator=(const ZZ& a) { NTL_zcopy(a.rep, &rep); return *this; }
 ZZ& operator=(long a) { NTL_zintoz(a, &rep); return *this; }
 
 
-~ZZ() { NTL_zfree(&rep); }
-
 void kill()
 // force the space held by this ZZ to be released.
 // The value then becomes 0.
 
-{ NTL_zfree(&rep); }
+{ rep.kill(); }
+
+
+void swap(ZZ& x)
+{ NTL_zswap(&rep, &x.rep); }
 
 void SetSize(long k)
 // pre-allocates space for k-digit numbers (base 2^NTL_ZZ_NBITS);  
@@ -115,7 +126,7 @@ long WideSinglePrecision() const
 static const ZZ& zero();
 
 
-ZZ(ZZ& x, INIT_TRANS_TYPE) { rep = x.rep; x.rep = 0; }
+ZZ(ZZ& x, INIT_TRANS_TYPE) { rep.swap(x.rep); }
 // used to cheaply hand off memory management of return value,
 // without copying, assuming compiler implements the
 // "return value optimization".  This is probably obsolete by
@@ -127,21 +138,21 @@ ZZ(ZZ& x, INIT_TRANS_TYPE) { rep = x.rep; x.rep = 0; }
 
 // mainly for internal consumption by ZZWatcher
 
-void release() { if (MaxAlloc() > NTL_RELEASE_THRESH) kill(); }
+void KillBig() { if (MaxAlloc() > NTL_RELEASE_THRESH) kill(); }
 
 };
 
 
 class ZZWatcher {
 public:
-   ZZ *watched;
+   ZZ& watched;
    explicit
-   ZZWatcher(ZZ *_watched) : watched(_watched) {}
+   ZZWatcher(ZZ& _watched) : watched(_watched) {}
 
-   ~ZZWatcher() { watched->release(); }
+   ~ZZWatcher() { watched.KillBig(); }
 };
 
-#define NTL_ZZRegister(x) NTL_THREAD_LOCAL static ZZ x; ZZWatcher _WATCHER__ ## x(&x)
+#define NTL_ZZRegister(x) NTL_TLS_LOCAL(ZZ, x); ZZWatcher _WATCHER__ ## x(x)
 
 
 
@@ -164,7 +175,7 @@ inline void set(ZZ& x)
 inline void swap(ZZ& x, ZZ& y)
 // swap the values of x and y (swaps pointers only)
 
-   { NTL_zswap(&x.rep, &y.rep); }
+   { x.swap(y); }
 
 
 inline double log(const ZZ& a)
@@ -197,15 +208,15 @@ inline void conv(ZZ& x, unsigned int a) { NTL_zuintoz((unsigned long)(a), &x.rep
 inline ZZ to_ZZ(unsigned int a) { return ZZ(INIT_VAL, a); }
 
 void conv(ZZ& x, const char *s);
-inline ZZ::ZZ(INIT_VAL_TYPE, const char *s) : rep(0) { conv(*this, s); }
+inline ZZ::ZZ(INIT_VAL_TYPE, const char *s)  { conv(*this, s); }
 inline ZZ to_ZZ(const char *s) { return ZZ(INIT_VAL, s); }
 
 inline void conv(ZZ& x, double a) { NTL_zdoubtoz(a, &x.rep); }
-inline ZZ::ZZ(INIT_VAL_TYPE, double a) : rep(0) { conv(*this, a); }
+inline ZZ::ZZ(INIT_VAL_TYPE, double a)  { conv(*this, a); }
 inline ZZ to_ZZ(double a) { return ZZ(INIT_VAL, a); }
 
 inline void conv(ZZ& x, float a) { NTL_zdoubtoz(double(a), &x.rep); }
-inline ZZ::ZZ(INIT_VAL_TYPE, float a) : rep(0) { conv(*this, a); }
+inline ZZ::ZZ(INIT_VAL_TYPE, float a)  { conv(*this, a); }
 inline ZZ to_ZZ(float a) { return ZZ(INIT_VAL, a); }
 
 inline void conv(long& x, const ZZ& a) { x = NTL_ztoint(a.rep); }
@@ -480,38 +491,104 @@ MulSubFrom(ZZ& x, const ZZ& a, const ZZ& b)
 }
 
 
+
 // Special routines for implementing CRT in ZZ_pX arithmetic
+// These are verbose, but fairly boilerplate
 
 
-inline void ZZ_p_crt_struct_init(void **crt_struct, long n, const ZZ& p, 
-                                 const long *primes)
-    { NTL_crt_struct_init(crt_struct, n, p.rep, primes); }
 
-inline void ZZ_p_crt_struct_insert(void *crt_struct, long i, const ZZ& m)
-   { NTL_crt_struct_insert(crt_struct, i, m.rep); }
+class ZZ_CRTStructAdapter;
+class ZZ_RemStructAdapter;
 
-inline void ZZ_p_crt_struct_free(void *crt_struct)
-   { NTL_crt_struct_free(crt_struct); }
+class ZZ_TmpVecAdapter {
+public:
+   UniquePtr<_ntl_tmp_vec> rep;
 
-inline void ZZ_p_crt_struct_eval(void *crt_struct, ZZ& t, const long *a)
-   { NTL_crt_struct_eval(crt_struct, &t.rep, a); }
-
-inline long ZZ_p_crt_struct_special(void *crt_struct)
-   { return NTL_crt_struct_special(crt_struct); }
-
-// Special routines for fast remaindering
+   inline void fetch(const ZZ_CRTStructAdapter&);
+   inline void fetch(ZZ_CRTStructAdapter&);
+   inline void fetch(const ZZ_RemStructAdapter&);
+};
 
 
-inline void ZZ_p_rem_struct_init(void **rem_struct, long n, 
-                                 const ZZ& p, long *primes)
-   { NTL_rem_struct_init(rem_struct, n, p.rep, primes); }
+class ZZ_CRTStructAdapter {
+public:
+   UniquePtr<_ntl_crt_struct> rep;
 
-inline void ZZ_p_rem_struct_free(void *rem_struct)
-   { NTL_rem_struct_free(rem_struct); }
+   void init(long n, const ZZ& p, long (*primes)(long))
+   {
+      rep.reset(_ntl_crt_struct_build(n, p.rep, primes));
+   }
+
+   void insert(long i, const ZZ& m)
+   {
+       rep->insert(i, m.rep);
+   }
+
+   void eval(ZZ& t, const long *a, ZZ_TmpVecAdapter& tmp_vec) const
+   {
+      rep->eval(&t.rep, a, tmp_vec.rep.get());
+   }
+
+   bool special() const
+   { 
+      return rep->special();
+   }
+};
 
 
-inline void ZZ_p_rem_struct_eval(void *rem_struct, long *x, const ZZ& a)
-   { NTL_rem_struct_eval(rem_struct, x, a.rep); }
+class ZZ_RemStructAdapter {
+public:
+   UniquePtr<_ntl_rem_struct> rep;
+
+   void init(long n, const ZZ& p, long (*primes)(long))
+   {
+      rep.reset(_ntl_rem_struct_build(n, p.rep, primes));
+   }
+
+   void eval(long *x, const ZZ& a, ZZ_TmpVecAdapter& tmp_vec) const
+   {
+      rep->eval(x, a.rep, tmp_vec.rep.get());
+   }
+};
+
+
+inline void ZZ_TmpVecAdapter::fetch(const ZZ_CRTStructAdapter& crt_struct)
+{
+   rep.reset(crt_struct.rep->fetch()); 
+}
+
+inline void ZZ_TmpVecAdapter::fetch(ZZ_CRTStructAdapter& crt_struct)
+{
+   rep.reset(crt_struct.rep->extract()); // EXTRACT!!
+}
+
+
+inline void ZZ_TmpVecAdapter::fetch(const ZZ_RemStructAdapter& rem_struct)
+{
+   rep.reset(rem_struct.rep->fetch()); 
+}
+
+
+// montgomery
+class ZZ_ReduceStructAdapter {
+public:
+   UniquePtr<_ntl_reduce_struct> rep;
+
+   void init(const ZZ& p, const ZZ& excess)
+   {
+      rep.reset(_ntl_reduce_struct_build(p.rep, excess.rep));
+   }
+
+   void eval(ZZ& x, ZZ& a) const
+   {
+      rep->eval(&x.rep, &a.rep);
+   }
+
+   void adjust(ZZ& x) const
+   {
+      rep->adjust(&x.rep);
+   }
+};
 
 
 
@@ -599,6 +676,30 @@ inline ZZ& operator%=(ZZ& x, const ZZ& b)
    { rem(x, x, b); return x; } 
 
 
+// preconditioned single-precision variant
+// not documented for now...
+
+
+class PreconditionedRemainder {
+private:
+   long p;
+   UniquePtr<_ntl_general_rem_one_struct> pinfo;
+
+public:
+   PreconditionedRemainder(long _p, long sz) : p(_p)
+   {
+      pinfo.reset(_ntl_general_rem_one_struct_build(p, sz));
+   }
+
+
+   long operator()(const ZZ& a) 
+   {
+      return _ntl_general_rem_one_struct_apply(a.rep, p, pinfo.get());
+   }
+};
+      
+
+
 /**********************************************************
 
                         GCD's
@@ -874,8 +975,87 @@ inline long ZZ_RoundCorrection(const ZZ& a, long k, long residual)
 ************************************************************/
 
 
+// ================ NEW PRG STUFF =================
+
+
+// Low-level key-derivation 
+
+
+void DeriveKey(unsigned char *key, long klen,  
+               const unsigned char *data, long dlen);
+
+
+
+// Low-level chacha stuff
+
+#define NTL_PRG_KEYLEN (32)
+
+class RandomStream {
+private:
+   _ntl_uint32 state[16];
+   unsigned char buf[64];
+   long pos;
+
+   void do_get(unsigned char *res, long n); 
+
+public:
+   explicit
+   RandomStream(const unsigned char *key);
+
+   // No default constructor 
+   // default copy and assignment
+
+   void get(unsigned char *res, long n) 
+   {
+      // optimize short reads
+      if (n >= 0 && n <= 64-pos) {
+         long i;
+         for (i = 0; i < n; i++) {
+            res[i] = buf[pos+i];
+         }
+         pos += n;
+      }
+      else {
+         do_get(res, n);
+      }
+   }
+
+};
+
+   
+
+
+RandomStream& GetCurrentRandomStream();
+// get reference to the current random by stream --
+// if SetSeed has not been called, it is called with
+// a default value (which should be unique to each
+// process/thread
+
+
 void SetSeed(const ZZ& s);
+void SetSeed(const unsigned char *data, long dlen);
+void SetSeed(const RandomStream& s);
 // initialize random number generator
+// in the first two version, a PRG key is derived from
+// the data using DeriveKey.
+
+
+// RAII for saving/restoring current state of PRG
+
+class RandomStreamPush {
+private: 
+   RandomStream saved;
+
+   RandomStreamPush(const RandomStreamPush&); // disable
+   void operator=(const RandomStreamPush&); // disable
+
+public:
+   RandomStreamPush() : saved(GetCurrentRandomStream()) { }
+   ~RandomStreamPush() { SetSeed(saved); } 
+
+};
+
+
 
 
 void RandomBnd(ZZ& x, const ZZ& n);
@@ -903,10 +1083,16 @@ inline ZZ RandomBits_ZZ(long NumBits)
 // single-precision version of the above
 
 long RandomBnd(long n);
+inline void RandomBnd(long& x, long n) { x = RandomBnd(n); }
 
 long RandomLen_long(long l);
+inline void RandomLen(long& x, long l) { x = RandomLen_long(l); }
 
 long RandomBits_long(long l);
+inline void RandomBits(long& x, long l) { x = RandomBits_long(l); }
+
+
+// specialty routines
 
 unsigned long RandomWord();
 unsigned long RandomBits_ulong(long l);
@@ -1148,24 +1334,24 @@ inline void SqrMod(ZZ& x, const ZZ& a, const ZZ& n)
 inline ZZ SqrMod(const ZZ& a, const ZZ& n)
    {  ZZ x; SqrMod(x, a, n); NTL_OPT_RETURN(ZZ, x); }
 
-inline void InvMod(ZZ& x, const ZZ& a, const ZZ& n)
-// x = a^{-1} mod n, 0 <= x < n
-// error is raised occurs if inverse not defined
-
-   { NTL_zinvmod(a.rep, n.rep, &x.rep); }
+void InvMod(ZZ& x, const ZZ& a, const ZZ& n);
+// defined in ZZ.c in terms of InvModStatus
 
 inline ZZ InvMod(const ZZ& a, const ZZ& n)
    {  ZZ x; InvMod(x, a, n); NTL_OPT_RETURN(ZZ, x); }
 
 
 inline long InvModStatus(ZZ& x, const ZZ& a, const ZZ& n)
-// if gcd(a,b) = 1, then ReturnValue = 0, x = a^{-1} mod n
+// if gcd(a,n) = 1, then ReturnValue = 0, x = a^{-1} mod n
 // otherwise, ReturnValue = 1, x = gcd(a, n)
 
   { return NTL_zinv(a.rep, n.rep, &x.rep); }
 
 
-inline void PowerMod(ZZ& x, const ZZ& a, const ZZ& e, const ZZ& n)
+void PowerMod(ZZ& x, const ZZ& a, const ZZ& e, const ZZ& n);
+// defined in ZZ.c in terms of LowLevelPowerMod
+
+inline void LowLevelPowerMod(ZZ& x, const ZZ& a, const ZZ& e, const ZZ& n)
    { NTL_zpowermod(a.rep, e.rep, n.rep, &x.rep); }
 
 inline ZZ PowerMod(const ZZ& a, const ZZ& e, const ZZ& n)
@@ -1224,7 +1410,7 @@ inline ZZ SqrRootMod(const ZZ& a, const ZZ& n)
 
 class PrimeSeq {
 
-char *movesieve;
+const char *movesieve;
 Vec<char> movesieve_mem;
 long pindex;
 long pshift;
@@ -1269,604 +1455,56 @@ NTL_SNS ostream& operator<<(NTL_SNS ostream& s, const ZZ& a);
 
 
 
-/****************************************************************
-
-    Single-precision modular arithmetic
-
-*****************************************************************/
-
-
-/*
-these routines implement single-precision modular arithmetic.
-If n is the modulus, all inputs should be in the range 0..n-1.
-The number n itself should be in the range 1..2^{NTL_SP_NBITS}-1.
-*/
-
-// I've declared these "static" so that the installation wizard
-// has more flexibility, without worrying about the (rather esoteric)
-// possibility of the linker complaining when the definitions
-// are inconsistent across severeal files.
-// Maybe an unnamed namespace would be better.
-
-// DIRT: undocumented feature: in all of these MulMod routines,
-// the first argument, a, need only be in the range
-// 0..2^{NTL_SP_NBITS}-1.  This is assumption is used internally
-// in some NT routines...I've tried to mark all such uses with a
-// DIRT comment.  I may decide to make this feature part
-// of the documented interface at some point in the future.
-
-
-
-
-
-
-
-
-
-static inline long AddMod(long a, long b, long n)
-// return (a+b)%n
-
-{
-   long res = a + b;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING) && !defined(NTL_CLEAN_INT))
-   res -= n;
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-   return res;
-#elif (defined(NTL_AVOID_BRANCHING))
-   res -= n;
-   res += (long) ((-(((unsigned long) res) >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n));
-   return res;
-#else
-   if (res >= n)
-      return res - n;
-   else
-      return res;
-#endif
-}
-
-static inline long SubMod(long a, long b, long n)
-// return (a-b)%n
-
-{
-   long res = a - b;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING) && !defined(NTL_CLEAN_INT))
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-   return res;
-#elif (defined(NTL_AVOID_BRANCHING))
-   res += (long) ((-(((unsigned long) res) >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n));
-   return res;
-#else
-   if (res < 0)
-      return res + n;
-   else
-      return res;
-#endif
-}
-
-static inline long NegateMod(long a, long n)
-{
-   return SubMod(0, a, n);
-}
-
-
-
-#if (defined(NTL_CLEAN_INT) || (defined(NTL_AVOID_BRANCHING)  && !NTL_ARITH_RIGHT_SHIFT))
-#define NTL_CLEAN_SPMM
-#endif
-
-
-#if (!defined(NTL_CLEAN_SPMM))
-
-
-/*
- * The default MulMod code.
- */
-
-static inline long MulMod(long a, long b, long n)
-{
-   long q, res;
-
-   q  = (long) ((((double) a) * ((double) b)) / ((double) n)); 
-   res = a*b - q*n;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-   res -= n;
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-#else
-   if (res >= n)
-      res -= n;
-   else if (res < 0)
-      res += n;
-#endif
-   return res;
-}
-
-static inline long MulMod(long a, long b, long n, double ninv)
-{
-   long q, res;
-
-   q  = (long) ((((double) a) * ((double) b)) * ninv); 
-   res = a*b - q*n;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-   res -= n;
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-#else
-   if (res >= n)
-      res -= n;
-   else if (res < 0)
-      res += n;
-#endif
-   return res;
-}
-
-
-static inline long MulMod2(long a, long b, long n, double bninv)
-{
-   long q, res;
-
-   q  = (long) (((double) a) * bninv);
-   res = a*b - q*n;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-   res -= n;
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-#else
-   if (res >= n)
-      res -= n;
-   else if (res < 0)
-      res += n;
-#endif
-   return res;
-}
-
-static inline long MulDivRem(long& qq, long a, long b, long n, double bninv)
-{
-   long q, res;
-
-   q  = (long) (((double) a) * bninv);
-   res = a*b - q*n;
-   if (res >= n) {
-      res -= n;
-      q++;
-   } else if (res < 0) {
-      res += n;
-      q--;
-   }
-
-   qq = q;
-   return res;
-}
-
-#else
-
-/*
- * NTL_CLEAN_INT set: these versions of MulMod are completely portable,
- * assuming IEEE floating point arithmetic.
- */
-
-static inline long MulMod(long a, long b, long n)
-{  
-   long q;
-   unsigned long res;
-
-   q  = (long) ((((double) a) * ((double) b)) / ((double) n)); 
-
-   res = ((unsigned long) a)*((unsigned long) b) - 
-         ((unsigned long) q)*((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-   res -= ((unsigned long) n);
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-#else
-   if (res >> (NTL_BITS_PER_LONG-1))
-      res += ((unsigned long) n);
-   else if (((long) res) >= n)
-      res -= ((unsigned long) n);
-#endif
- 
-   return ((long) res);
-}
-
-static inline long MulMod(long a, long b, long n, double ninv)
-{
-   long q; 
-   unsigned long res;
-
-   q  = (long) ((((double) a) * ((double) b)) * ninv); 
-
-   res = ((unsigned long) a)*((unsigned long) b) - 
-         ((unsigned long) q)*((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-   res -= ((unsigned long) n);
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-#else
-   if (res >> (NTL_BITS_PER_LONG-1))
-      res += ((unsigned long) n);
-   else if (((long) res) >= n)
-      res -= ((unsigned long) n);
-#endif
- 
-   return ((long) res);
-}
-
-
-static inline long MulMod2(long a, long b, long n, double bninv)
-{
-   long q;
-   unsigned long res;
-
-   q  = (long) (((double) a) * bninv);
-
-   res = ((unsigned long) a)*((unsigned long) b) - 
-         ((unsigned long) q)*((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-   res -= ((unsigned long) n);
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-#else
-   if (res >> (NTL_BITS_PER_LONG-1))
-      res += ((unsigned long) n);
-   else if (((long) res) >= n)
-      res -= ((unsigned long) n);
-#endif
-
- 
-   return ((long) res);
-}
-
-static inline long MulDivRem(long& qq, long a, long b, long n, double bninv)
-{
-   long q; 
-   unsigned long res;
-
-   q  = (long) (((double) a) * bninv);
-   res = ((unsigned long) a)*((unsigned long) b) - 
-         ((unsigned long) q)*((unsigned long) n);
-
-   if (res >> (NTL_BITS_PER_LONG-1)) {
-      res += n;
-      q--;
-   } else if (((long) res) >= n) {
-      res -= n;
-      q++;
-   }
-
-   qq = q;
-   return ((long) res);
-}
-
-
-#endif
-
-
-
-
-// These MulMod routines (with preconditioning) are sometimes
-// significantly faster.  There are four possible implementations:
-//  - default: uses MulMod2 above (lots of floating point)
-//  - NTL_SPMM_ULL: uses unsigned long long (if possible)
-//  - NTL_SPMM_ASM: uses assembly language (if possible)
-//  - NTL_SPMM_UL: uses only unsigned long arithmetic (portable, slower).
-
-#if ((defined(NTL_SPMM_ULL) || defined(NTL_SPMM_ASM)))
-
-
-// unsigned long long / asm versions
-
-typedef unsigned long mulmod_precon_t;
-
-#define NTL_SPMM_VEC_T vec_ulong
-
-#if (!defined(NTL_CLEAN_SPMM))
-
-static inline unsigned long PrepMulModPrecon(long b, long n, double ninv)
-{
-   long q, r;
-
-   q  = (long) ( (((double) b) * NTL_SP_FBOUND) * ninv ); 
-   r = (b << NTL_SP_NBITS) - q*n;
-
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   q += 1 + (r >> (NTL_BITS_PER_LONG-1)) + ((r - n) >> (NTL_BITS_PER_LONG-1));
-#else
-   if (r >= n)
-      q++;
-   else if (r < 0)
-      q--;
-#endif
-
-   return ((unsigned long) q) << (NTL_BITS_PER_LONG - NTL_SP_NBITS);
-}
-
-
-#else
-
-/*
- * clean int version -- this should be completely portable.
- */
-
-
-static inline unsigned long PrepMulModPrecon(long b, long n, double ninv)
-{
-   unsigned long q, r;
-
-   q = (long) ( (((double) b) * NTL_SP_FBOUND) * ninv ); 
-   r = (((unsigned long) b) << NTL_SP_NBITS ) - q * ((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   q += 1UL - (r >> (NTL_BITS_PER_LONG-1)) - ((r - ((unsigned long) n)) >> (NTL_BITS_PER_LONG-1));
-#else
-   if (r >> (NTL_BITS_PER_LONG-1))
-      q--;
-   else if (((long) r) >= n)
-      q++;
-#endif
-
-   return q << (NTL_BITS_PER_LONG - NTL_SP_NBITS);
-}
-
-
-
-#endif
-
-
-
-
-#if (defined(NTL_SPMM_ULL))
-
-static inline unsigned long MulHiUL(unsigned long a, unsigned long b)
-{
-   return (((NTL_ULL_TYPE)(a)) * ((NTL_ULL_TYPE)(b))) >> NTL_BITS_PER_LONG;
-} 
-
-#else 
-
-// assmbly code versions
-
-#include <NTL/SPMM_ASM.h>
-
-
-#endif
-
-
-
-
-
-   
-
-#if (!defined(NTL_CLEAN_SPMM))
-
-
-
-static inline long MulModPrecon(long a, long b, long n, unsigned long bninv)
-{
-   long q, res;
-   
-   q = (long) MulHiUL(a, bninv);
-
-   res = a*b - q*n;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   res -= n;
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-#else
-   if (res >= n)
-      res -= n;
-#endif
-   return res;
-}
-
-
-#else
-
-static inline long MulModPrecon(long a, long b, long n, unsigned long bninv)
-{
-   unsigned long q, res;
-
-   
-   q = MulHiUL(a, bninv);
-
-   res = ((unsigned long) a)*((unsigned long) b) - q*((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   res -= ((unsigned long) n);
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-#else
-   if (((long) res) >= n)
-      res -= ((unsigned long) n);
-#endif
-
-   return (long) res;
-}
-
-#endif
-
-
-
-#elif (defined(NTL_SPMM_UL))
-
-// plain, portable (but slower) int version
-
-typedef long mulmod_precon_t;
-
-#define NTL_SPMM_VEC_T vec_long
-
-
-#if (!defined(NTL_CLEAN_SPMM))
-
-static inline long PrepMulModPrecon(long b, long n, double ninv)
-{
-   long q, r;
-
-   q  = (long) ( (((double) b) * NTL_SP_FBOUND) * ninv ); 
-   r = (b << NTL_SP_NBITS) - q*n;
-
-
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   q += 1 + (r >> (NTL_BITS_PER_LONG-1)) + ((r - n) >> (NTL_BITS_PER_LONG-1));
-#else
-   if (r >= n)
-      q++;
-   else if (r < 0)
-      q--;
-#endif
-
-   return q;
-}
-
-
-#else
-
-static inline long PrepMulModPrecon(long b, long n, double ninv)
-{
-   unsigned long q, r;
-
-   q = (long) ( (((double) b) * NTL_SP_FBOUND) * ninv ); 
-   r = (((unsigned long) b) << NTL_SP_NBITS ) - q * ((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   q += 1UL - (r >> (NTL_BITS_PER_LONG-1)) - ((r - ((unsigned long) n)) >> (NTL_BITS_PER_LONG-1));
-#else
-   if (r >> (NTL_BITS_PER_LONG-1))
-      q--;
-   else if (((long) r) >= n)
-      q++;
-#endif
-
-   return ((long) q);
-}
-
-
-#endif
-
-
-
-
-static inline long MulHiSP(long b, long d)
-{
-   unsigned long _b1 = b & ((1UL << (NTL_SP_NBITS/2)) - 1UL);
-   unsigned long _d1 = d & ((1UL << (NTL_SP_NBITS/2)) - 1UL);
-   unsigned long _bd,_b1d1,_m,_aa;
-   unsigned long _ld = (d>>(NTL_SP_NBITS/2));
-   unsigned long _lb = (b>>(NTL_SP_NBITS/2));
-
-   _bd=_lb*_ld;
-   _b1d1=_b1*_d1;
-   _m=(_lb+_b1)*(_ld+_d1) - _bd - _b1d1;
-   _aa = ( _b1d1+ ((_m&((1UL << (NTL_SP_NBITS/2)) - 1UL))<<(NTL_SP_NBITS/2)));
-   return (_aa >> NTL_SP_NBITS) + _bd + (_m>>(NTL_SP_NBITS/2));
-}
-
-
-#if (!defined(NTL_CLEAN_SPMM))
-
-static inline long MulModPrecon(long a, long b, long n, long bninv)
-{
-
-   long q, res;
-
-   q = MulHiSP(a, bninv);
-
-   res = a*b - q*n;
-#if (NTL_ARITH_RIGHT_SHIFT && defined(NTL_AVOID_BRANCHING))
-   res -= n;
-   res += (res >> (NTL_BITS_PER_LONG-1)) & n;
-#else
-   if (res >= n)
-      res -= n;
-#endif
-   return res;
-}
-
-
-
-#else
-
-
-static inline long MulModPrecon(long a, long b, long n, long bninv)
-{
-
-   unsigned long q, res;
-
-   q = MulHiSP(a, bninv);
-
-   res = ((unsigned long) a)*((unsigned long) b) - q*((unsigned long) n);
-
-#if (defined(NTL_AVOID_BRANCHING))
-   res -= ((unsigned long) n);
-   res += (-(res >> (NTL_BITS_PER_LONG-1))) & ((unsigned long) n);
-#else
-   if (((long) res) >= n)
-      res -= ((unsigned long) n);
-#endif
-
-   return (long) res;
-}
-
-
-#endif
-
-
-
-
-#else
-
-// default, double version
-
-typedef double mulmod_precon_t;
-
-#define NTL_SPMM_VEC_T vec_double
-
-static inline double PrepMulModPrecon(long b, long n, double ninv)
-{
-   return ((double) b) * ninv;
-}
-
-static inline long MulModPrecon(long a, long b, long n, double bninv)
-{
-   return MulMod2(a, b, n, bninv);
-}
-
-
-#endif
+// Some additional SP arithmetic routines, not defined in sp_arith.h
 
 
 long InvMod(long a, long n);
 // computes a^{-1} mod n.  Error is raised if undefined.
 
+long InvModStatus(long& x, long a, long n);
+// if gcd(a,n) = 1, then ReturnValue = 0, x = a^{-1} mod n
+// otherwise, ReturnValue = 1, x = gcd(a, n)
+
 long PowerMod(long a, long e, long n);
 // computes a^e mod n, e >= 0
 
 
-inline
-void VectorMulModPrecon(long k, long *x, const long *a, long b, long n, 
-                        mulmod_precon_t bninv)
-{
-   for (long i = 0; i < k; i++)
-      x[i] = MulModPrecon(a[i], b, n, bninv);
-}
+// Error handling
 
-inline
-void VectorMulMod(long k, long *x, const long *a, long b, long n, 
-                  double ninv)
-{
-   mulmod_precon_t bninv;
-   bninv = PrepMulModPrecon(b, n, ninv);
-   VectorMulModPrecon(k, x, a, b, n, bninv);
-}
+#ifdef NTL_EXCEPTIONS
+
+class InvModErrorObject : public ArithmeticErrorObject {
+private:
+   SmartPtr<ZZ> a_ptr;
+   SmartPtr<ZZ> n_ptr;
+public:
+   InvModErrorObject(const char *s, const ZZ& a, const ZZ& n)
+      : ArithmeticErrorObject(s) , a_ptr(MakeSmart<ZZ>(a)),
+        n_ptr(MakeSmart<ZZ>(n)) { }
+
+   const ZZ& get_a() const { return *a_ptr; }
+   const ZZ& get_n() const { return *n_ptr; }
+};
+
+#else
+
+// We need this alt definition to keep pre-C++11 
+// compilers happy (NTL_EXCEPTIONS should only be used
+// with C++11 compilers).
+
+class InvModErrorObject : public ArithmeticErrorObject {
+public:
+   InvModErrorObject(const char *s, const ZZ& a, const ZZ& n)
+      : ArithmeticErrorObject(s) { }
+
+   const ZZ& get_a() const { return ZZ::zero(); }
+   const ZZ& get_n() const { return ZZ::zero(); }
+};
+
+#endif
 
 
-inline 
-void VectorMulMod(long k, long *x, const long *a, long b, long n)
-{
-   double ninv = 1/((double) n);
-   VectorMulMod(k, x, a, b, n, ninv);
-}
-
+void InvModError(const char *s, const ZZ& a, const ZZ& n); 
 
 NTL_CLOSE_NNS
 
