@@ -16,43 +16,45 @@
 #endif
 
 
-// some crossover points...choice depends
-// if we are using the gf2x lib or not
-
-#ifdef NTL_GF2X_LIB
-
-#define NTL_GF2X_GCD_CROSSOVER (400L*NTL_BITS_PER_LONG) 
-#define NTL_GF2X_HalfGCD_CROSSOVER (6L*NTL_BITS_PER_LONG)
-#define NTL_GF2X_BERMASS_CROSSOVER (200L*NTL_BITS_PER_LONG)
-
+// simple scaling factor for some crossover points:
+// we use a lower crossover of the underlying multiplication
+// is faster  
+#if (defined(NTL_GF2X_LIB) || defined(NTL_PCLMUL))
+#define XOVER_SCALE (1L)
 #else
-
-#define NTL_GF2X_GCD_CROSSOVER (900L*NTL_BITS_PER_LONG) 
-#define NTL_GF2X_HalfGCD_CROSSOVER (6L*NTL_BITS_PER_LONG)
-#define NTL_GF2X_BERMASS_CROSSOVER (450L*NTL_BITS_PER_LONG)
-
+#define XOVER_SCALE (2L)
 #endif
 
 
+
+#define NTL_GF2X_GCD_CROSSOVER (XOVER_SCALE*300L*NTL_BITS_PER_LONG) 
+
+#define NTL_GF2X_BERMASS_CROSSOVER (XOVER_SCALE*200L*NTL_BITS_PER_LONG)
+
+#define NTL_GF2X_HalfGCD_CROSSOVER (4L*NTL_BITS_PER_LONG)
+
+
+
 NTL_START_IMPL
 
 
 
-NTL_THREAD_LOCAL
-static vec_GF2X stab;  // used by PlainDivRem and PlainRem
+NTL_TLS_GLOBAL_DECL(vec_GF2X, stab)  
+// used by PlainDivRem and PlainRem
 
-NTL_THREAD_LOCAL
-static WordVector GF2X_rembuf;
+NTL_TLS_GLOBAL_DECL(WordVector, GF2X_rembuf)
 
 
 void PlainDivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2X& b)
 {
+   NTL_TLS_GLOBAL_ACCESS(stab);
+
    long da, sa, posa, db, sb, posb, dq, sq, posq;
 
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("GF2X: division by zero");
+   if (db < 0) ArithmeticError("GF2X: division by zero");
 
    if (da < db) {
       r = a;
@@ -69,6 +71,9 @@ void PlainDivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2X& b)
    sq = dq/NTL_BITS_PER_LONG + 1;
    posq = dq - NTL_BITS_PER_LONG*(sq-1);
 
+   NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+   WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
    _ntl_ulong *ap;
    if (&r == &a)
       ap = r.xrep.elts();
@@ -80,6 +85,13 @@ void PlainDivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2X& b)
    stab.SetLength(NTL_BITS_PER_LONG);
    long i;
 
+   NTL_SCOPE(guard) {
+      for (i = 0; i <= min(dq, NTL_BITS_PER_LONG-1); i++) {
+         WordVector& st = stab[((_ntl_ulong)(posb+i))%NTL_BITS_PER_LONG].xrep;
+         st.KillBig();
+      }
+   };
+
    stab[posb] = b;
    for (i = 1; i <= min(dq, NTL_BITS_PER_LONG-1); i++) 
       MulByX(stab[((_ntl_ulong)(posb+i))%NTL_BITS_PER_LONG], 
@@ -138,10 +150,10 @@ void PlainDivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2X& b)
    }
    r.normalize();
 
-   GF2X_rembuf.release();
+   guard.relax();
    for (i = 0; i <= min(dq, NTL_BITS_PER_LONG-1); i++) {
       WordVector& st = stab[((_ntl_ulong)(posb+i))%NTL_BITS_PER_LONG].xrep;
-      st.release();
+      st.KillBig();
    }
 }
 
@@ -156,12 +168,14 @@ void PlainDiv(GF2X& q, const GF2X& a, const GF2X& b)
 
 void PlainRem(GF2X& r, const GF2X& a, const GF2X& b)
 {
+   NTL_TLS_GLOBAL_ACCESS(stab);
+
    long da, sa, posa, db, sb, posb;
 
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("GF2X: division by zero");
+   if (db < 0) ArithmeticError("GF2X: division by zero");
 
    if (da < db) {
       r = a;
@@ -173,6 +187,10 @@ void PlainRem(GF2X& r, const GF2X& a, const GF2X& b)
    sb = b.xrep.length();
    posb = db - NTL_BITS_PER_LONG*(sb-1);
 
+
+   NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+   WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
    _ntl_ulong *ap;
    if (&r == &a)
       ap = r.xrep.elts();
@@ -184,6 +202,13 @@ void PlainRem(GF2X& r, const GF2X& a, const GF2X& b)
    stab.SetLength(NTL_BITS_PER_LONG);
    long i;
 
+   NTL_SCOPE(guard) {
+      for (i = 0; i <= min(da-db, NTL_BITS_PER_LONG-1); i++) {
+         WordVector& st = stab[((_ntl_ulong)(posb+i))%NTL_BITS_PER_LONG].xrep;
+         st.KillBig();
+      }
+   };
+
    stab[posb] = b;
    for (i = 1; i <= min(da-db, NTL_BITS_PER_LONG-1); i++) 
       MulByX(stab[((_ntl_ulong)(posb+i))%NTL_BITS_PER_LONG], 
@@ -230,10 +255,10 @@ void PlainRem(GF2X& r, const GF2X& a, const GF2X& b)
    }
    r.normalize();
 
-   GF2X_rembuf.release();
+   guard.relax();
    for (i = 0; i <= min(da-db, NTL_BITS_PER_LONG-1); i++) {
       WordVector& st = stab[((_ntl_ulong)(posb+i))%NTL_BITS_PER_LONG].xrep;
-      st.release();
+      st.KillBig();
    }
 }
 
@@ -263,7 +288,7 @@ void NewtonInvTrunc(GF2X& c, const GF2X& a, long e)
       return;
    }
 
-   NTL_THREAD_LOCAL static vec_long E;
+   NTL_TLS_LOCAL(vec_long, E);
    E.SetLength(0);
    append(E, e);
    while (e > 8) {
@@ -313,10 +338,10 @@ void NewtonInvTrunc(GF2X& c, const GF2X& a, long e)
 void InvTrunc(GF2X& c, const GF2X& a, long e)
 {
    if (ConstTerm(a) == 0 || e < 0)
-      Error("inv: bad args");
+      LogicError("inv: bad args");
 
    if (NTL_OVERFLOW(e, 1, 0))
-      Error("overflow in InvTrunc");
+      ResourceError("overflow in InvTrunc");
 
    if (e == 0) {
       clear(c);
@@ -408,9 +433,9 @@ void build(GF2XModulus& F, const GF2X& f)
    long n = deg(f);
    long i;
 
-   if (n <= 0) Error("build(GF2XModulus,GF2X): deg(f) <= 0");
+   if (n <= 0) LogicError("build(GF2XModulus,GF2X): deg(f) <= 0");
 
-   F.tracevec.SetLength(0);
+   F.tracevec.make();
 
    F.f = f;
    F.n = n;
@@ -444,35 +469,28 @@ void build(GF2XModulus& F, const GF2X& f)
 
    GF2X f0;
    trunc(f0, f, n);
-   long deg_f0 = deg(f0);
 
-   if (F.sn > 1 && deg_f0 < NTL_BITS_PER_LONG 
-       && deg_f0 >= NTL_BITS_PER_LONG/2) {
-      if (F.size >= 6)
-         F.method = GF2X_MOD_MUL;
-      else
-         F.method = GF2X_MOD_SPECIAL;
-   }
-   else if (F.sn > 1 && deg_f0 < NTL_BITS_PER_LONG/2) {
-      if (F.size >= 4)
-         F.method = GF2X_MOD_MUL;
-      else
-         F.method = GF2X_MOD_SPECIAL;
-   }
-   else if (F.size >= 8)
+   if (F.n >= (NTL_BITS_PER_LONG/2)*XOVER_SCALE)
       F.method = GF2X_MOD_MUL;
    else 
       F.method = GF2X_MOD_PLAIN;
+
+
+   // NOTE: I've run some tests which indicate that the GF2X_MOD_SPECIAL
+   // method is not worth it.
+   // FIXME: in a future version, I should eliminate all code
+   // and data associated with GF2X_MOD_SPECIAL
+
+   // NOTE: I've runs some tests which indicate that the crossover
+   // for GF2X_MOD_MUL is extremely low, even without PCLMUL support.
       
 
    if (F.method == GF2X_MOD_SPECIAL) {
-      if (!F.stab_cnt) F.stab_cnt = NTL_NEW_OP long[NTL_BITS_PER_LONG];
-      long *stab_cnt = F.stab_cnt;
-      if (!stab_cnt) Error("out of memory");
+      if (!F.stab_cnt) F.stab_cnt.SetLength(NTL_BITS_PER_LONG);
+      long *stab_cnt = F.stab_cnt.get();
 
-      if (!F.stab1) F.stab1 = NTL_NEW_OP _ntl_ulong[2*NTL_BITS_PER_LONG];
-      _ntl_ulong *stab1 = F.stab1;
-      if (!stab1) Error("out of memory");
+      if (!F.stab1) F.stab1.SetLength(2*NTL_BITS_PER_LONG);
+      _ntl_ulong *stab1 = F.stab1.get();
 
       stab1[posb<<1] = f.xrep[0];
       stab1[(posb<<1)+1] = 0;
@@ -498,13 +516,11 @@ void build(GF2XModulus& F, const GF2X& f)
       stab.SetLength(NTL_BITS_PER_LONG);
 
 
-      if (!F.stab_ptr) F.stab_ptr = NTL_NEW_OP _ntl_ulong_ptr[NTL_BITS_PER_LONG];
-      _ntl_ulong **stab_ptr = F.stab_ptr;
-      if (!stab_ptr) Error("out of memory");
+      if (!F.stab_ptr) F.stab_ptr.SetLength(NTL_BITS_PER_LONG);
+      _ntl_ulong **stab_ptr = F.stab_ptr.get();
 
-      if (!F.stab_cnt) F.stab_cnt = NTL_NEW_OP long[NTL_BITS_PER_LONG];
-      long *stab_cnt = F.stab_cnt;
-      if (!stab_cnt) Error("out of memory");
+      if (!F.stab_cnt) F.stab_cnt.SetLength(NTL_BITS_PER_LONG);
+      long *stab_cnt = F.stab_cnt.get();
       
    
       stab[posb] = f;
@@ -535,9 +551,6 @@ GF2XModulus::GF2XModulus()
 {
    n = -1;
    method = GF2X_MOD_PLAIN;
-   stab_ptr = 0;
-   stab_cnt = 0;
-   stab1 = 0;
 }
 
 
@@ -549,16 +562,14 @@ GF2XModulus::GF2XModulus(const GF2XModulus& F) :
    f(F.f), n(F.n), sn(F.sn), posn(F.posn), k3(F.k3), k2(F.k2), k1(F.k1),
    size(F.size), 
    msk(F.msk), method(F.method), stab(F.stab), h0(F.h0), f0(F.f0),
-   stab_cnt(0), stab_ptr(0), stab1(0), tracevec(F.tracevec)
+   tracevec(F.tracevec)
 {
    if (method == GF2X_MOD_SPECIAL) {
       long i;
-      stab1 = NTL_NEW_OP _ntl_ulong[2*NTL_BITS_PER_LONG];
-      if (!stab1) Error("GF2XModulus: out of memory");
+      stab1.SetLength(2*NTL_BITS_PER_LONG);
       for (i = 0; i < 2*NTL_BITS_PER_LONG; i++)
          stab1[i] = F.stab1[i];
-      stab_cnt = NTL_NEW_OP long[NTL_BITS_PER_LONG];
-      if (!stab_cnt) Error("GF2XModulus: out of memory");
+      stab_cnt.SetLength(NTL_BITS_PER_LONG);
       for (i = 0; i < NTL_BITS_PER_LONG; i++)
          stab_cnt[i] = F.stab_cnt[i];
    }
@@ -566,16 +577,13 @@ GF2XModulus::GF2XModulus(const GF2XModulus& F) :
       long i;
 
       if (F.stab_cnt) {
-         stab_cnt = NTL_NEW_OP long[NTL_BITS_PER_LONG];
-         if (!stab_cnt) Error("GF2XModulus: out of memory");
+         stab_cnt.SetLength(NTL_BITS_PER_LONG);
          for (i = 0; i < NTL_BITS_PER_LONG; i++)
             stab_cnt[i] = F.stab_cnt[i];
       }
 
       if (F.stab_ptr) {
-         stab_ptr = NTL_NEW_OP _ntl_ulong_ptr[NTL_BITS_PER_LONG];
-         if (!stab_ptr) Error("GF2XModulus: out of memory");
-      
+         stab_ptr.SetLength(NTL_BITS_PER_LONG);
          for (i = 0; i < NTL_BITS_PER_LONG; i++) {
             WordVector& st = stab[((_ntl_ulong)(posn+i))%NTL_BITS_PER_LONG].xrep;
             long k = st.length();
@@ -598,12 +606,10 @@ GF2XModulus& GF2XModulus::operator=(const GF2XModulus& F)
 
    if (method == GF2X_MOD_SPECIAL) {
       long i;
-      if (!stab1) stab1 = NTL_NEW_OP _ntl_ulong[2*NTL_BITS_PER_LONG];
-      if (!stab1) Error("GF2XModulus: out of memory");
+      if (!stab1) stab1.SetLength(2*NTL_BITS_PER_LONG);
       for (i = 0; i < 2*NTL_BITS_PER_LONG; i++)
          stab1[i] = F.stab1[i];
-      if (!stab_cnt) stab_cnt = NTL_NEW_OP long[NTL_BITS_PER_LONG];
-      if (!stab_cnt) Error("GF2XModulus: out of memory");
+      if (!stab_cnt) stab_cnt.SetLength(NTL_BITS_PER_LONG);
       for (i = 0; i < NTL_BITS_PER_LONG; i++)
          stab_cnt[i] = F.stab_cnt[i];
    }
@@ -611,15 +617,13 @@ GF2XModulus& GF2XModulus::operator=(const GF2XModulus& F)
       long i;
 
       if (F.stab_cnt) {
-         if (!stab_cnt) stab_cnt = NTL_NEW_OP long[NTL_BITS_PER_LONG];
-         if (!stab_cnt) Error("GF2XModulus: out of memory");
+         if (!stab_cnt) stab_cnt.SetLength(NTL_BITS_PER_LONG);
          for (i = 0; i < NTL_BITS_PER_LONG; i++)
             stab_cnt[i] = F.stab_cnt[i];
       }
 
       if (F.stab_ptr) {
-         if (!stab_ptr) stab_ptr = NTL_NEW_OP _ntl_ulong_ptr[NTL_BITS_PER_LONG];
-         if (!stab_ptr) Error("GF2XModulus: out of memory");
+         if (!stab_ptr) stab_ptr.SetLength(NTL_BITS_PER_LONG);
       
          for (i = 0; i < NTL_BITS_PER_LONG; i++) {
             WordVector& st = stab[((_ntl_ulong)(posn+i))%NTL_BITS_PER_LONG].xrep;
@@ -635,22 +639,11 @@ GF2XModulus& GF2XModulus::operator=(const GF2XModulus& F)
    
 
 
-GF2XModulus::~GF2XModulus() 
-{ 
-   delete [] stab_ptr; 
-   delete [] stab_cnt; 
-   delete [] stab1; 
-}
-
-
 
 GF2XModulus::GF2XModulus(const GF2X& ff)
 {
    n = -1;
    method = GF2X_MOD_PLAIN;
-   stab_ptr = 0;
-   stab_cnt = 0;
-   stab1 = 0;
 
    build(*this, ff);
 }
@@ -1021,7 +1014,7 @@ static
 void RightShiftAdd(GF2X& c, const GF2X& a, long n)
 {
    if (n < 0) {
-      Error("RightShiftAdd: negative shamt");
+      LogicError("RightShiftAdd: negative shamt");
    }
 
    if (n == 0) {
@@ -1264,7 +1257,7 @@ void rem(GF2X& r, const GF2X& a, const GF2XModulus& F)
 {
    long n = F.n;
 
-   if (n < 0) Error("rem: uninitialized modulus");
+   if (n < 0) LogicError("rem: uninitialized modulus");
 
    if (F.method == GF2X_MOD_TRI) {
       TrinomReduce(r, a, n, F.k3);
@@ -1289,6 +1282,9 @@ void rem(GF2X& r, const GF2X& a, const GF2XModulus& F)
          UseMulRemX1(r, a, F);
    }
    else if (F.method == GF2X_MOD_SPECIAL) {
+      NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+      WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
       long sa = a.xrep.length();
       long posa = da - NTL_BITS_PER_LONG*(sa-1);
    
@@ -1334,10 +1330,12 @@ void rem(GF2X& r, const GF2X& a, const GF2XModulus& F)
       r.normalize();
    }
    else {
+      NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+      WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
       long sa = a.xrep.length();
       long posa = da - NTL_BITS_PER_LONG*(sa-1);
    
-   
       _ntl_ulong *ap;
       if (&r == &a)
          ap = r.xrep.elts();
@@ -1378,7 +1376,6 @@ void rem(GF2X& r, const GF2X& a, const GF2XModulus& F)
       r.normalize();
    }
 
-   GF2X_rembuf.release();
 }
 
 void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2XModulus& F)
@@ -1386,7 +1383,7 @@ void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2XModulus& F)
    long da = deg(a);
    long n = F.n;
 
-   if (n < 0) Error("DivRem: uninitialized modulus");
+   if (n < 0) LogicError("DivRem: uninitialized modulus");
 
    if (da < n) {
       r = a;
@@ -1411,6 +1408,9 @@ void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2XModulus& F)
          UseMulDivRemX1(q, r, a, F);
    }
    else if (F.method == GF2X_MOD_SPECIAL) {
+      NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+      WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
       long sa = a.xrep.length();
       long posa = da - NTL_BITS_PER_LONG*(sa-1);
    
@@ -1475,6 +1475,9 @@ void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2XModulus& F)
       r.normalize();
    }
    else {
+      NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+      WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
       long sa = a.xrep.length();
       long posa = da - NTL_BITS_PER_LONG*(sa-1);
    
@@ -1535,8 +1538,6 @@ void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2XModulus& F)
       }
       r.normalize();
    }
-
-   GF2X_rembuf.release();
 }
 
 
@@ -1546,7 +1547,7 @@ void div(GF2X& q, const GF2X& a, const GF2XModulus& F)
    long da = deg(a);
    long n = F.n;
 
-   if (n < 0) Error("div: uninitialized modulus");
+   if (n < 0) LogicError("div: uninitialized modulus");
 
 
    if (da < n) {
@@ -1571,6 +1572,9 @@ void div(GF2X& q, const GF2X& a, const GF2XModulus& F)
          UseMulDivX1(q, a, F);
    }
    else if (F.method == GF2X_MOD_SPECIAL) {
+      NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+      WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
       long sa = a.xrep.length();
       long posa = da - NTL_BITS_PER_LONG*(sa-1);
    
@@ -1621,6 +1625,9 @@ void div(GF2X& q, const GF2X& a, const GF2XModulus& F)
       }
    }
    else {
+      NTL_TLS_GLOBAL_ACCESS(GF2X_rembuf);
+      WordVectorWatcher watch_GF2X_rembuf(GF2X_rembuf);
+
       long sa = a.xrep.length();
       long posa = da - NTL_BITS_PER_LONG*(sa-1);
    
@@ -1668,14 +1675,12 @@ void div(GF2X& q, const GF2X& a, const GF2XModulus& F)
          }
       }
    }
-
-   GF2X_rembuf.release();
 }
 
 
 void MulMod(GF2X& c, const GF2X& a, const GF2X& b, const GF2XModulus& F)
 {
-   if (F.n < 0) Error("MulMod: uninitialized modulus");
+   if (F.n < 0) LogicError("MulMod: uninitialized modulus");
 
    NTL_GF2XRegister(t);
    mul(t, a, b);
@@ -1685,7 +1690,7 @@ void MulMod(GF2X& c, const GF2X& a, const GF2X& b, const GF2XModulus& F)
 
 void SqrMod(GF2X& c, const GF2X& a, const GF2XModulus& F)
 {
-   if (F.n < 0) Error("SqrMod: uninitialized modulus");
+   if (F.n < 0) LogicError("SqrMod: uninitialized modulus");
 
    NTL_GF2XRegister(t);
    sqr(t, a);
@@ -1739,7 +1744,7 @@ long OptWinSize(long n)
 void PowerMod(GF2X& h, const GF2X& g, const ZZ& e, const GF2XModulus& F)
 // h = g^e mod f using "sliding window" algorithm
 {
-   if (deg(g) >= F.n) Error("PowerMod: bad args");
+   if (deg(g) >= F.n) LogicError("PowerMod: bad args");
 
    if (e == 0) {
       set(h);
@@ -1853,7 +1858,7 @@ void PowerMod(GF2X& h, const GF2X& g, const ZZ& e, const GF2XModulus& F)
 
 void PowerXMod(GF2X& hh, const ZZ& e, const GF2XModulus& F)
 {
-   if (F.n < 0) Error("PowerXMod: uninitialized modulus");
+   if (F.n < 0) LogicError("PowerXMod: uninitialized modulus");
 
    if (IsZero(e)) {
       set(hh);
@@ -1949,16 +1954,16 @@ void UseMulDiv(GF2X& q, const GF2X& a, const GF2X& b)
 }
 
 
-const long GF2X_DIV_CROSS = 100; 
+const long GF2X_DIV_CROSS = (NTL_BITS_PER_LONG/2)*XOVER_SCALE; 
 
 void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2X& b)
 {
-   long sa = a.xrep.length();
-   long sb = b.xrep.length();
+   long da = deg(a);
+   long db = deg(b);
 
-   if (sb < GF2X_DIV_CROSS || sa-sb < GF2X_DIV_CROSS)
+   if (db < GF2X_DIV_CROSS || da-db < GF2X_DIV_CROSS)
       PlainDivRem(q, r, a, b);
-   else if (sa < 4*sb)
+   else if (da < 4*db)
       UseMulDivRem(q, r, a, b);
    else {
       GF2XModulus B;
@@ -1969,12 +1974,12 @@ void DivRem(GF2X& q, GF2X& r, const GF2X& a, const GF2X& b)
 
 void div(GF2X& q, const GF2X& a, const GF2X& b)
 {
-   long sa = a.xrep.length();
-   long sb = b.xrep.length();
+   long da = deg(a);
+   long db = deg(b);
 
-   if (sb < GF2X_DIV_CROSS || sa-sb < GF2X_DIV_CROSS)
+   if (db < GF2X_DIV_CROSS || da-db < GF2X_DIV_CROSS)
       PlainDiv(q, a, b);
-   else if (sa < 4*sb)
+   else if (da < 4*db)
       UseMulDiv(q, a, b);
    else {
       GF2XModulus B;
@@ -1985,12 +1990,12 @@ void div(GF2X& q, const GF2X& a, const GF2X& b)
 
 void rem(GF2X& r, const GF2X& a, const GF2X& b)
 {
-   long sa = a.xrep.length();
-   long sb = b.xrep.length();
+   long da = deg(a);
+   long db = deg(b);
 
-   if (sb < GF2X_DIV_CROSS || sa-sb < GF2X_DIV_CROSS)
+   if (db < GF2X_DIV_CROSS || da-db < GF2X_DIV_CROSS)
       PlainRem(r, a, b);
-   else if (sa < 4*sb)
+   else if (da < 4*db)
       UseMulRem(r, a, b);
    else {
       GF2XModulus B;
@@ -2363,7 +2368,7 @@ void OldXGCD(GF2X& d, GF2X& s, GF2X& t, const GF2X& a, const GF2X& b)
 static
 void BaseInvMod(GF2X& d, GF2X& s, const GF2X& a, const GF2X& f)
 {
-   if (deg(a) >= deg(f) || deg(f) == 0) Error("InvMod: bad args");
+   if (deg(a) >= deg(f) || deg(f) == 0) LogicError("InvMod: bad args");
 
    long sa = a.xrep.length();
    long sf = f.xrep.length();
@@ -2388,7 +2393,7 @@ void InvMod(GF2X& c, const GF2X& a, const GF2X& f)
    NTL_GF2XRegister(s);
    BaseInvMod(d, s, a, f);
 
-   if (!IsOne(d)) Error("InvMod: inverse undefined");
+   if (!IsOne(d)) InvModError("InvMod: inverse undefined");
 
    c = s;
 }
@@ -2485,10 +2490,10 @@ void conv(ZZX& x, const GF2X& a)
 
 void VectorCopy(vec_GF2& x, const GF2X& a, long n)
 {
-   if (n < 0) Error("VectorCopy: negative length"); 
+   if (n < 0) LogicError("VectorCopy: negative length"); 
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("overflow in VectorCopy");
+      ResourceError("overflow in VectorCopy");
 
    long wa = a.xrep.length();
    long wx = (n + NTL_BITS_PER_LONG - 1)/NTL_BITS_PER_LONG;
@@ -2660,7 +2665,7 @@ void CompMod(GF2X& x, const GF2X& g, const GF2XArgument& A, const GF2XModulus& F
 
 void build(GF2XArgument& A, const GF2X& h, const GF2XModulus& F, long m)
 {
-   if (m <= 0 || deg(h) >= F.n) Error("build GF2XArgument: bad args");
+   if (m <= 0 || deg(h) >= F.n) LogicError("build GF2XArgument: bad args");
 
    if (m > F.n) m = F.n;
 
@@ -2755,7 +2760,7 @@ void build(GF2XTransMultiplier& B, const GF2X& b, const GF2XModulus& F)
 {
    long db = deg(b);
 
-   if (db >= F.n) Error("build TransMultiplier: bad args");
+   if (db >= F.n) LogicError("build TransMultiplier: bad args");
 
    GF2X t;
 
@@ -2801,7 +2806,7 @@ void build(GF2XTransMultiplier& B, const GF2X& b, const GF2XModulus& F)
 void TransMulMod(GF2X& x, const GF2X& a, const GF2XTransMultiplier& B,
                const GF2XModulus& F)
 {
-   if (deg(a) >= F.n) Error("TransMulMod: bad args");
+   if (deg(a) >= F.n) LogicError("TransMulMod: bad args");
 
    NTL_GF2XRegister(t1);
    NTL_GF2XRegister(t2);
@@ -2853,8 +2858,11 @@ void ProjectPowers(GF2X& x, const GF2X& a, long k, const GF2XArgument& H,
 {
    long n = F.n;
 
-   if (deg(a) >= n || k < 0 || NTL_OVERFLOW(k, 1, 0)) 
-      Error("ProjectPowers: bad args");
+   if (deg(a) >= n || k < 0) 
+      LogicError("ProjectPowers: bad args");
+
+   if (NTL_OVERFLOW(k, 1, 0)) 
+      ResourceError("ProjectPowers: excessive parameter");
 
    long m = H.H.length()-1;
    long l = (k+m-1)/m - 1;
@@ -2892,7 +2900,7 @@ void ProjectPowers(vec_GF2& x, const vec_GF2& a, long k,
 void ProjectPowers(GF2X& x, const GF2X& a, long k, const GF2X& h, 
                    const GF2XModulus& F)
 {
-   if (deg(a) >= F.n || k < 0) Error("ProjectPowers: bad args");
+   if (deg(a) >= F.n || k < 0) LogicError("ProjectPowers: bad args");
 
    if (k == 0) {
       clear(x);
@@ -3033,8 +3041,8 @@ void DoMinPolyMod(GF2X& h, const GF2X& g, const GF2XModulus& F, long m,
 
 void MinPolySeq(GF2X& h, const vec_GF2& a, long m)
 {
-   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) Error("MinPoly: bad args");
-   if (a.length() < 2*m) Error("MinPoly: sequence too short");
+   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) LogicError("MinPoly: bad args");
+   if (a.length() < 2*m) LogicError("MinPoly: sequence too short");
    GF2X x;
    x.xrep = a.rep;
    x.normalize();
@@ -3044,7 +3052,7 @@ void MinPolySeq(GF2X& h, const vec_GF2& a, long m)
 void ProbMinPolyMod(GF2X& h, const GF2X& g, const GF2XModulus& F, long m)
 {
    long n = F.n;
-   if (m < 1 || m > n) Error("ProbMinPoly: bad args");
+   if (m < 1 || m > n) LogicError("ProbMinPoly: bad args");
 
    GF2X R;
    random(R, n);
@@ -3061,7 +3069,7 @@ void MinPolyMod(GF2X& hh, const GF2X& g, const GF2XModulus& F, long m)
 {
    GF2X h, h1;
    long n = F.n;
-   if (m < 1 || m > n) Error("MinPoly: bad args");
+   if (m < 1 || m > n) LogicError("MinPoly: bad args");
 
    /* probabilistically compute min-poly */
 
@@ -3094,7 +3102,7 @@ void MinPolyMod(GF2X& hh, const GF2X& g, const GF2XModulus& F, long m)
 
 void IrredPolyMod(GF2X& h, const GF2X& g, const GF2XModulus& F, long m)
 {
-   if (m < 1 || m > F.n) Error("IrredPoly: bad args");
+   if (m < 1 || m > F.n) LogicError("IrredPoly: bad args");
 
    GF2X R;
    set(R);
@@ -3122,7 +3130,7 @@ void MulByXMod(GF2X& c, const GF2X& a, const GF2XModulus& F)
 {
    long da = deg(a);
    long df = deg(F);
-   if (da >= df) Error("MulByXMod: bad args"); 
+   if (da >= df) LogicError("MulByXMod: bad args"); 
 
    MulByX(c, a);
 
@@ -3135,7 +3143,7 @@ void MulByXModAux(GF2X& c, const GF2X& a, const GF2X& f)
 {
    long da = deg(a);
    long df = deg(f);
-   if (da >= df) Error("MulByXMod: bad args"); 
+   if (da >= df) LogicError("MulByXMod: bad args"); 
 
    MulByX(c, a);
 
@@ -3160,7 +3168,7 @@ void MulByXMod(GF2X& h, const GF2X& a, const GF2X& f)
 void power(GF2X& x, const GF2X& a, long e)
 {
    if (e < 0) {
-      Error("power: negative exponent");
+      ArithmeticError("power: negative exponent");
    }
 
    if (e == 0) {
@@ -3176,7 +3184,7 @@ void power(GF2X& x, const GF2X& a, long e)
    long da = deg(a);
 
    if (da > (NTL_MAX_LONG-1)/e)
-      Error("overflow in power");
+      ResourceError("overflow in power");
 
    GF2X res;
    res.SetMaxLength(da*e + 1);
@@ -3200,7 +3208,7 @@ void FastTraceVec(vec_GF2& S, const GF2XModulus& f)
 {
    long n = deg(f);
 
-   if (n <= 0) Error("TraceVec: bad args");
+   if (n <= 0) LogicError("TraceVec: bad args");
 
    GF2X x = reverse(-LeftShift(reverse(diff(reverse(f)), n-1), n-1)/f, n-1);
 
@@ -3214,7 +3222,7 @@ void PlainTraceVec(vec_GF2& S, const GF2X& f)
    long n = deg(f);
 
    if (n <= 0) 
-      Error("TraceVec: bad args");
+      LogicError("TraceVec: bad args");
 
    if (n == 0) {
       S.SetLength(0);
@@ -3234,13 +3242,8 @@ void TraceVec(vec_GF2& S, const GF2X& f)
 }
 
 static
-void ComputeTraceVec(const GF2XModulus& F)
+void ComputeTraceVec(vec_GF2& S, const GF2XModulus& F)
 {
-   vec_GF2& S = *((vec_GF2 *) &F.tracevec);
-
-   if (S.length() > 0)
-      return;
-
    if (F.method == GF2X_MOD_PLAIN) {
       PlainTraceVec(S, F.f);
    }
@@ -3254,20 +3257,24 @@ void TraceMod(ref_GF2 x, const GF2X& a, const GF2XModulus& F)
    long n = F.n;
 
    if (deg(a) >= n)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
-   // FIXME: in a thread-safe impl, this needs a mutex
+   do { // NOTE: thread safe lazy init
+      Lazy<vec_GF2>::Builder builder(F.tracevec.val()); 
+      if (!builder()) break;
+      UniquePtr<vec_GF2> p;
+      p.make();
+      ComputeTraceVec(*p, F);
+      builder.move(p);
+   } while (0);
 
-   if (F.tracevec.length() == 0) 
-      ComputeTraceVec(F);
-
-   project(x, F.tracevec, a);
+   project(x, *F.tracevec.val(), a);
 }
 
 void TraceMod(ref_GF2 x, const GF2X& a, const GF2X& f)
 {
    if (deg(a) >= deg(f) || deg(f) <= 0)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
    project(x, TraceVec(f), a);
 }
