@@ -1,8 +1,10 @@
 #include <NTL/config.h>
 
-#include <time.h>
+#include <ctime>
 #include <NTL/ctools.h>
 
+using namespace std;
+
 // FIXME: this is the GetTime that ends up getting used
 // on Windows. However, it returns the wall time, not CPU time.
 // We could perhaps switch to using GetProcessTimes.
@@ -14,8 +16,8 @@
 
 double _ntl_GetTime()
 {
-   NTL_THREAD_LOCAL static clock_t last_clock = 0;
-   NTL_THREAD_LOCAL static double acc = 0;
+   static NTL_CHEAP_THREAD_LOCAL clock_t last_clock = 0;
+   static NTL_CHEAP_THREAD_LOCAL double acc = 0;
 
    clock_t this_clock;
    double delta;
