@@ -1,45 +1,93 @@
 
-#include <NTL/config.h>
+#include <NTL/ZZ.h>
 
-#if (defined(NTL_STD_CXX) || defined(NTL_PSTD_NHF))
-#include <iostream>
-using namespace std;
-#else
-#include <iostream.h>
-#endif
+NTL_CLIENT
 
 #define make_string_aux(x) #x
 #define make_string(x) make_string_aux(x)
 
 
-
 int main()
 {
 
-#ifdef NTL_STD_CXX
-   cout << "NTL_STD_CXX=1\n";
+#ifdef NTL_LEGACY_NO_NAMESPACE
+   cout << "NTL_LEGACY_NO_NAMESPACE=1\n";
 #else
-   cout << "NTL_STD_CXX=0\n";
+   cout << "NTL_LEGACY_NO_NAMESPACE=0\n";
 #endif
 
-#ifdef NTL_PSTD_NNS
-   cout << "NTL_PSTD_NNS=1\n";
+#ifdef NTL_LEGACY_INPUT_ERROR
+   cout << "NTL_LEGACY_INPUT_ERROR=1\n";
 #else
-   cout << "NTL_PSTD_NNS=0\n";
+   cout << "NTL_LEGACY_INPUT_ERROR=0\n";
 #endif
 
-#ifdef NTL_PSTD_NHF
-   cout << "NTL_PSTD_NHF=1\n";
+#ifdef NTL_THREADS
+   cout << "NTL_THREADS=1\n";
 #else
-   cout << "NTL_PSTD_NHF=0\n";
+   cout << "NTL_THREADS=0\n";
 #endif
 
-#ifdef NTL_PSTD_NTN
-   cout << "NTL_PSTD_NTN=1\n";
+#ifdef NTL_DISABLE_TLS_HACK
+   cout << "NTL_DISABLE_TLS_HACK=1\n";
 #else
-   cout << "NTL_PSTD_NTN=0\n";
+   cout << "NTL_DISABLE_TLS_HACK=0\n";
 #endif
 
+#ifdef NTL_ENABLE_TLS_HACK
+   cout << "NTL_ENABLE_TLS_HACK=1\n";
+#else
+   cout << "NTL_ENABLE_TLS_HACK=0\n";
+#endif
+
+#ifdef NTL_EXCEPTIONS
+   cout << "NTL_EXCEPTIONS=1\n";
+#else
+   cout << "NTL_EXCEPTIONS=0\n";
+#endif
+
+#ifdef NTL_THREAD_BOOST
+   cout << "NTL_THREAD_BOOST=1\n";
+#else
+   cout << "NTL_THREAD_BOOST=0\n";
+#endif
+
+
+#ifdef NTL_LEGACY_SP_MULMOD
+   cout << "NTL_LEGACY_SP_MULMOD=1\n";
+#else
+   cout << "NTL_LEGACY_SP_MULMOD=0\n";
+#endif
+
+
+#ifdef NTL_DISABLE_LONGDOUBLE
+   cout << "NTL_DISABLE_LONGDOUBLE=1\n";
+#else
+   cout << "NTL_DISABLE_LONGDOUBLE=0\n";
+#endif
+
+
+#ifdef NTL_DISABLE_LONGLONG
+   cout << "NTL_DISABLE_LONGLONG=1\n";
+#else
+   cout << "NTL_DISABLE_LONGLONG=0\n";
+#endif
+
+
+#ifdef NTL_DISABLE_LL_ASM
+   cout << "NTL_DISABLE_LL_ASM=1\n";
+#else
+   cout << "NTL_DISABLE_LL_ASM=0\n";
+#endif
+
+#ifdef NTL_MAXIMIZE_SP_NBITS
+   cout << "NTL_MAXIMIZE_SP_NBITS=1\n";
+#else
+   cout << "NTL_MAXIMIZE_SP_NBITS=0\n";
+#endif
+
+
+
 #ifdef NTL_GMP_LIP
    cout << "NTL_GMP_LIP=1\n";
 #else
@@ -109,5 +157,13 @@ int main()
 #endif
 
 
+// the following is synthetically defined
+#ifdef NTL_LONGLONG_SP_MULMOD
+   cout << "NTL_LONGLONG_SP_MULMOD=1\n";
+#else
+   cout << "NTL_LONGLONG_SP_MULMOD=0\n";
+#endif
+
+
    return 0;
 }
