@@ -10,14 +10,14 @@ NTL_START_IMPL
 static inline
 void CheckFinite(double *p)
 {
-   if (!IsFinite(p)) Error("LLL_QP: numbers too big...use LLL_XD");
+   if (!IsFinite(p)) ResourceError("LLL_QP: numbers too big...use LLL_XD");
 }
 
 
 static inline
 void CheckFinite(quad_float *p)
 {
-   if (!IsFinite(p)) Error("LLL_QP: numbers too big...use LLL_XD");
+   if (!IsFinite(p)) ResourceError("LLL_QP: numbers too big...use LLL_XD");
 }
 
 
@@ -430,12 +430,13 @@ void ComputeGS(mat_ZZ& B, quad_float **B1, quad_float **mu, quad_float *b,
    c[k] = b[k] - s;
 }
 
-NTL_THREAD_LOCAL static quad_float red_fudge = to_quad_float(0);
-NTL_THREAD_LOCAL static long log_red = 0;
-NTL_THREAD_LOCAL static long verbose = 0;
-NTL_THREAD_LOCAL static unsigned long NumSwaps = 0;
-NTL_THREAD_LOCAL static double StartTime = 0;
-NTL_THREAD_LOCAL static double LastTime = 0;
+NTL_TLS_GLOBAL_DECL_INIT(quad_float, red_fudge, (to_quad_float(0)))
+
+static NTL_CHEAP_THREAD_LOCAL long log_red = 0;
+static NTL_CHEAP_THREAD_LOCAL long verbose = 0;
+static NTL_CHEAP_THREAD_LOCAL unsigned long NumSwaps = 0;
+static NTL_CHEAP_THREAD_LOCAL double StartTime = 0;
+static NTL_CHEAP_THREAD_LOCAL double LastTime = 0;
 
 
 static void LLLStatus(long max_k, double t, long m, const mat_ZZ& B)
@@ -483,6 +484,8 @@ static void LLLStatus(long max_k, double t, long m, const mat_ZZ& B)
 
 static void init_red_fudge()
 {
+   NTL_TLS_GLOBAL_ACCESS(red_fudge);
+
    long i;
 
    // initial log_red should be <= NTL_DOUBLE_PRECISION-2,
@@ -498,6 +501,8 @@ static void init_red_fudge()
 
 static void inc_red_fudge()
 {
+   NTL_TLS_GLOBAL_ACCESS(red_fudge);
+
 
    red_fudge = red_fudge * 2;
    log_red--;
@@ -505,7 +510,7 @@ static void inc_red_fudge()
    cerr << "LLL_QP: warning--relaxing reduction (" << log_red << ")\n";
 
    if (log_red < 4)
-      Error("LLL_QP: too much loss of precision...stop!");
+      ResourceError("LLL_QP: too much loss of precision...stop!");
 }
 
 
@@ -515,6 +520,8 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
            quad_float *b, quad_float *c,
            long m, long init_k, long &quit)
 {
+   NTL_TLS_GLOBAL_ACCESS(red_fudge);
+
    long n = B.NumCols();
 
    long i, j, k, Fc1;
@@ -527,7 +534,7 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    quad_float *tp;
 
 
-   NTL_THREAD_LOCAL static double bound = 0;
+   static NTL_CHEAP_THREAD_LOCAL double bound = 0;
 
 
    if (bound == 0) {
@@ -557,17 +564,17 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
    for (i = k; i <= m+1; i++)
       st[i] = 1;
 
-   quad_float *buf;
-   buf = NTL_NEW_OP quad_float [m+1];
-   if (!buf) Error("out of memory in lll_LLL_QP");
+   UniqueArray<quad_float> buf_store;
+   buf_store.SetLength(m+1);
+   quad_float *buf = buf_store.get();
 
    vec_long in_vec_mem;
    in_vec_mem.SetLength(n+1);
    long *in_vec = in_vec_mem.elts();
 
-   double *max_b;
-   max_b = NTL_NEW_OP double [m+1];
-   if (!max_b) Error("out of memory in lll_LLL_QP");
+   UniqueArray<double> max_b_store;
+   max_b_store.SetLength(m+1);
+   double *max_b = max_b_store.get();
 
    for (i = 1; i <= m; i++)
       max_b[i] = max_abs(B1[i], n);
@@ -779,10 +786,6 @@ long ll_LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
       LLLStatus(m+1, GetTime(), m, B);
    }
 
-
-   delete [] buf;
-   delete [] max_b;
-
    return m;
 }
 
@@ -806,38 +809,22 @@ long LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
 
    if (U) ident(*U, m);
 
-   quad_float **B1;  // approximates B
-
-   typedef quad_float *quad_floatptr;
-
-   B1 = NTL_NEW_OP quad_floatptr[m+1];
-   if (!B1) Error("LLL_QP: out of memory");
-
-   for (i = 1; i <= m; i++) {
-      B1[i] = NTL_NEW_OP quad_float[n+1];
-      if (!B1[i]) Error("LLL_QP: out of memory");
-   }
-
-   quad_float **mu;
-   mu = NTL_NEW_OP quad_floatptr[m+1];
-   if (!mu) Error("LLL_QP: out of memory");
-
-   for (i = 1; i <= m; i++) {
-      mu[i] = NTL_NEW_OP quad_float[m+1];
-      if (!mu[i]) Error("LLL_QP: out of memory");
-   }
-
-   quad_float *c; // squared lengths of Gramm-Schmidt basis vectors
-
-   c = NTL_NEW_OP quad_float[m+1];
-   if (!c) Error("LLL_QP: out of memory");
-
-   quad_float *b; // squared lengths of basis vectors
-
-   b = NTL_NEW_OP quad_float[m+1];
-   if (!b) Error("LLL_QP: out of memory");
+   Unique2DArray<quad_float> B1_store;
+   B1_store.SetDimsFrom1(m+1, n+1);
+   quad_float **B1 = B1_store.get();  // approximates B
 
 
+   Unique2DArray<quad_float> mu_store;
+   mu_store.SetDimsFrom1(m+1, m+1);
+   quad_float **mu = mu_store.get();
+
+   UniqueArray<quad_float> c_store;
+   c_store.SetLength(m+1);
+   quad_float *c = c_store.get(); // squared lengths of Gramm-Schmidt basis vectors
+
+   UniqueArray<quad_float> b_store;
+   b_store.SetLength(m+1);
+   quad_float *b = b_store.get(); // squared lengths of basis vectors
 
    for (i = 1; i <=m; i++)
       for (j = 1; j <= n; j++) {
@@ -866,25 +853,6 @@ long LLL_QP(mat_ZZ& B, mat_ZZ* U, quad_float delta, long deep,
       }
    }
 
-
-   // clean-up
-
-   for (i = 1; i <= m+dep; i++) {
-      delete [] B1[i];
-   }
-
-   delete [] B1;
-
-   for (i = 1; i <= m+dep; i++) {
-      delete [] mu[i];
-   }
-
-   delete [] mu;
-
-   delete [] c;
-
-   delete [] b;
-
    return m;
 }
 
@@ -900,8 +868,8 @@ long LLL_QP(mat_ZZ& B, double delta, long deep, LLLCheckFct check,
       LastTime = StartTime;
    }
 
-   if (delta < 0.50 || delta >= 1) Error("LLL_QP: bad delta");
-   if (deep < 0) Error("LLL_QP: bad deep");
+   if (delta < 0.50 || delta >= 1) LogicError("LLL_QP: bad delta");
+   if (deep < 0) LogicError("LLL_QP: bad deep");
    return LLL_QP(B, 0, to_quad_float(delta), deep, check);
 }
 
@@ -916,18 +884,21 @@ long LLL_QP(mat_ZZ& B, mat_ZZ& U, double delta, long deep,
    }
 
 
-   if (delta < 0.50 || delta >= 1) Error("LLL_QP: bad delta");
-   if (deep < 0) Error("LLL_QP: bad deep");
+   if (delta < 0.50 || delta >= 1) LogicError("LLL_QP: bad delta");
+   if (deep < 0) LogicError("LLL_QP: bad deep");
    return LLL_QP(B, &U, to_quad_float(delta), deep, check);
 }
 
 
 
-NTL_THREAD_LOCAL static vec_quad_float BKZConstant;
+NTL_TLS_GLOBAL_DECL(vec_quad_float, BKZConstant)
 
 static
 void ComputeBKZConstant(long beta, long p)
 {
+   NTL_TLS_GLOBAL_ACCESS(BKZConstant);
+
+
    const quad_float c_PI = 
       to_quad_float("3.141592653589793238462643383279502884197");
    const quad_float LogPI = 
@@ -981,11 +952,14 @@ void ComputeBKZConstant(long beta, long p)
 }
 
 
-NTL_THREAD_LOCAL static vec_quad_float BKZThresh;
+NTL_TLS_GLOBAL_DECL(vec_quad_float, BKZThresh)
 
 static 
 void ComputeBKZThresh(quad_float *c, long beta)
 {
+   NTL_TLS_GLOBAL_ACCESS(BKZConstant);
+   NTL_TLS_GLOBAL_ACCESS(BKZThresh);
+
    BKZThresh.SetLength(beta-1);
 
    long i;
@@ -1060,6 +1034,10 @@ static
 long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta, 
          long beta, long prune, LLLCheckFct check)
 {
+   NTL_TLS_GLOBAL_ACCESS(red_fudge);
+   NTL_TLS_GLOBAL_ACCESS(BKZThresh);
+
+
 
    long m = BB.NumRows();
    long n = BB.NumCols();
@@ -1079,68 +1057,55 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
    B.SetDims(m+1, n);
 
+   Unique2DArray<quad_float> B1_store;
+   B1_store.SetDimsFrom1(m+2, n+1);
+   quad_float **B1 = B1_store.get();  // approximates B
 
-   quad_float **B1;  // approximates B
 
-   typedef quad_float *quad_floatptr;
+   Unique2DArray<quad_float> mu_store;
+   mu_store.SetDimsFrom1(m+2, m+1);
+   quad_float **mu = mu_store.get();
 
-   B1 = NTL_NEW_OP quad_floatptr[m+2];
-   if (!B1) Error("BKZ_QP: out of memory");
+   UniqueArray<quad_float> c_store;
+   c_store.SetLength(m+2);
+   quad_float *c = c_store.get(); // squared lengths of Gramm-Schmidt basis vectors
 
-   for (i = 1; i <= m+1; i++) {
-      B1[i] = NTL_NEW_OP quad_float[n+1];
-      if (!B1[i]) Error("BKZ_QP: out of memory");
-   }
-
-   quad_float **mu;
-   mu = NTL_NEW_OP quad_floatptr[m+2];
-   if (!mu) Error("BKZ_QP: out of memory");
-
-   for (i = 1; i <= m+1; i++) {
-      mu[i] = NTL_NEW_OP quad_float[m+1];
-      if (!mu[i]) Error("BKZ_QP: out of memory");
-   }
-
-   quad_float *c; // squared lengths of Gramm-Schmidt basis vectors
-
-   c = NTL_NEW_OP quad_float[m+2];
-   if (!c) Error("BKZ_QP: out of memory");
-
-   quad_float *b; // squared lengths of basis vectors
-
-   b = NTL_NEW_OP quad_float[m+2];
-   if (!b) Error("BKZ_QP: out of memory");
+   UniqueArray<quad_float> b_store;
+   b_store.SetLength(m+2);
+   quad_float *b = b_store.get(); // squared lengths of basis vectors
 
    quad_float cbar;
 
-   quad_float *ctilda;
-   ctilda = NTL_NEW_OP quad_float[m+2];
-   if (!ctilda) Error("BKZ_QP: out of memory");
 
-   quad_float *vvec;
-   vvec = NTL_NEW_OP quad_float[m+2];
-   if (!vvec) Error("BKZ_QP: out of memory");
-
-   quad_float *yvec;
-   yvec = NTL_NEW_OP quad_float[m+2];
-   if (!yvec) Error("BKZ_QP: out of memory");
-
-   quad_float *uvec;
-   uvec = NTL_NEW_OP quad_float[m+2];
-   if (!uvec) Error("BKZ_QP: out of memory");
-
-   quad_float *utildavec;
-   utildavec = NTL_NEW_OP quad_float[m+2];
-   if (!utildavec) Error("BKZ_QP: out of memory");
+   UniqueArray<quad_float> ctilda_store;
+   ctilda_store.SetLength(m+2);
+   quad_float *ctilda = ctilda_store.get();
 
 
-   long *Deltavec;
-   Deltavec = NTL_NEW_OP long[m+2];
-   if (!Deltavec) Error("BKZ_QP: out of memory");
+   UniqueArray<quad_float> vvec_store;
+   vvec_store.SetLength(m+2);
+   quad_float *vvec = vvec_store.get();
+
+   UniqueArray<quad_float> yvec_store;
+   yvec_store.SetLength(m+2);
+   quad_float *yvec = yvec_store.get();
+
+   UniqueArray<quad_float> uvec_store;
+   uvec_store.SetLength(m+2);
+   quad_float *uvec = uvec_store.get();
+
+   UniqueArray<quad_float> utildavec_store;
+   utildavec_store.SetLength(m+2);
+   quad_float *utildavec = utildavec_store.get();
+
+   UniqueArray<long> Deltavec_store;
+   Deltavec_store.SetLength(m+2);
+   long *Deltavec = Deltavec_store.get();
+
+   UniqueArray<long> deltavec_store;
+   deltavec_store.SetLength(m+2);
+   long *deltavec = deltavec_store.get();;
 
-   long *deltavec;
-   deltavec = NTL_NEW_OP long[m+2];
-   if (!deltavec) Error("BKZ_QP: out of memory");
 
    mat_ZZ Ulocal;
    mat_ZZ *U;
@@ -1354,7 +1319,7 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                }
             }
    
-            if (s == 0) Error("BKZ_QP: internal error");
+            if (s == 0) LogicError("BKZ_QP: internal error");
    
             if (s > 0) {
                // special case
@@ -1372,7 +1337,7 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                // cerr << "special case\n";
                new_m = ll_LLL_QP(B, U, delta, 0, check, 
                                 B1, mu, b, c, h, jj, quit);
-               if (new_m != h) Error("BKZ_QP: internal error");
+               if (new_m != h) LogicError("BKZ_QP: internal error");
                if (quit) break;
             }
             else {
@@ -1411,14 +1376,14 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                b[jj] = InnerProduct(B1[jj], B1[jj], n);
                CheckFinite(&b[jj]);
       
-               if (b[jj] == 0) Error("BKZ_QP: internal error"); 
+               if (b[jj] == 0) LogicError("BKZ_QP: internal error"); 
       
                // remove linear dependencies
    
                // cerr << "general case\n";
                new_m = ll_LLL_QP(B, U, delta, 0, 0, B1, mu, b, c, kk+1, jj, quit);
               
-               if (new_m != kk) Error("BKZ_QP: internal error"); 
+               if (new_m != kk) LogicError("BKZ_QP: internal error"); 
 
                // remove zero vector
       
@@ -1447,7 +1412,7 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                   new_m = ll_LLL_QP(B, U, delta, 0, check, 
                                    B1, mu, b, c, h, h, quit);
    
-                  if (new_m != h) Error("BKZ_QP: internal error");
+                  if (new_m != h) LogicError("BKZ_QP: internal error");
                   if (quit) break;
                }
             }
@@ -1464,7 +1429,7 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
             if (!clean) {
                new_m = 
                   ll_LLL_QP(B, U, delta, 0, check, B1, mu, b, c, h, h, quit);
-               if (new_m != h) Error("BKZ_QP: internal error");
+               if (new_m != h) LogicError("BKZ_QP: internal error");
                if (quit) break;
             }
    
@@ -1505,29 +1470,6 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
       *UU = *U;
    }
 
-   for (i = 1; i <= m_orig+1; i++) {
-      delete [] B1[i];
-   }
-
-   delete [] B1;
-
-   for (i = 1; i <= m_orig+1; i++) {
-      delete [] mu[i];
-   }
-
-   delete [] mu;
-
-
-   delete [] c;
-   delete [] b;
-   delete [] ctilda;
-   delete [] vvec;
-   delete [] yvec;
-   delete [] uvec;
-   delete [] utildavec;
-   delete [] Deltavec;
-   delete [] deltavec;
-
    return m;
 }
 
@@ -1542,8 +1484,8 @@ long BKZ_QP(mat_ZZ& BB, mat_ZZ& UU, double delta,
    }
 
 
-   if (delta < 0.50 || delta >= 1) Error("BKZ_QP: bad delta");
-   if (beta < 2) Error("BKZ_QP: bad block size");
+   if (delta < 0.50 || delta >= 1) LogicError("BKZ_QP: bad delta");
+   if (beta < 2) LogicError("BKZ_QP: bad block size");
 
    return BKZ_QP(BB, &UU, to_quad_float(delta), beta, prune, check);
 }
@@ -1560,8 +1502,8 @@ long BKZ_QP(mat_ZZ& BB, double delta,
 
 
 
-   if (delta < 0.50 || delta >= 1) Error("BKZ_QP: bad delta");
-   if (beta < 2) Error("BKZ_QP: bad block size");
+   if (delta < 0.50 || delta >= 1) LogicError("BKZ_QP: bad delta");
+   if (beta < 2) LogicError("BKZ_QP: bad block size");
 
    return BKZ_QP(BB, 0, to_quad_float(delta), beta, prune, check);
 }
@@ -1570,6 +1512,10 @@ static
 long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta, 
          long beta, long prune, LLLCheckFct check)
 {
+   NTL_TLS_GLOBAL_ACCESS(red_fudge);
+   NTL_TLS_GLOBAL_ACCESS(BKZThresh);
+
+
 
    long m = BB.NumRows();
    long n = BB.NumCols();
@@ -1588,68 +1534,54 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
 
    B.SetDims(m+1, n);
 
+   Unique2DArray<quad_float> B1_store;
+   B1_store.SetDimsFrom1(m+2, n+1);
+   quad_float **B1 = B1_store.get();  // approximates B
 
-   quad_float **B1;  // approximates B
 
-   typedef quad_float *quad_floatptr;
+   Unique2DArray<quad_float> mu_store;
+   mu_store.SetDimsFrom1(m+2, m+1);
+   quad_float **mu = mu_store.get();
 
-   B1 = NTL_NEW_OP quad_floatptr[m+2];
-   if (!B1) Error("BKZ_QP: out of memory");
+   UniqueArray<quad_float> c_store;
+   c_store.SetLength(m+2);
+   quad_float *c = c_store.get(); // squared lengths of Gramm-Schmidt basis vectors
 
-   for (i = 1; i <= m+1; i++) {
-      B1[i] = NTL_NEW_OP quad_float[n+1];
-      if (!B1[i]) Error("BKZ_QP: out of memory");
-   }
-
-   quad_float **mu;
-   mu = NTL_NEW_OP quad_floatptr[m+2];
-   if (!mu) Error("BKZ_QP: out of memory");
-
-   for (i = 1; i <= m+1; i++) {
-      mu[i] = NTL_NEW_OP quad_float[m+1];
-      if (!mu[i]) Error("BKZ_QP: out of memory");
-   }
-
-   quad_float *c; // squared lengths of Gramm-Schmidt basis vectors
-
-   c = NTL_NEW_OP quad_float[m+2];
-   if (!c) Error("BKZ_QP: out of memory");
-
-   quad_float *b; // squared lengths of basis vectors
-
-   b = NTL_NEW_OP quad_float[m+2];
-   if (!b) Error("BKZ_QP: out of memory");
+   UniqueArray<quad_float> b_store;
+   b_store.SetLength(m+2);
+   quad_float *b = b_store.get(); // squared lengths of basis vectors
 
    double cbar;
 
-   double *ctilda;
-   ctilda = NTL_NEW_OP double[m+2];
-   if (!ctilda) Error("BKZ_QP: out of memory");
 
-   double *vvec;
-   vvec = NTL_NEW_OP double[m+2];
-   if (!vvec) Error("BKZ_QP: out of memory");
-
-   double *yvec;
-   yvec = NTL_NEW_OP double[m+2];
-   if (!yvec) Error("BKZ_QP: out of memory");
-
-   double *uvec;
-   uvec = NTL_NEW_OP double[m+2];
-   if (!uvec) Error("BKZ_QP: out of memory");
-
-   double *utildavec;
-   utildavec = NTL_NEW_OP double[m+2];
-   if (!utildavec) Error("BKZ_QP: out of memory");
+   UniqueArray<double> ctilda_store;
+   ctilda_store.SetLength(m+2);
+   double *ctilda = ctilda_store.get();
 
 
-   long *Deltavec;
-   Deltavec = NTL_NEW_OP long[m+2];
-   if (!Deltavec) Error("BKZ_QP: out of memory");
+   UniqueArray<double> vvec_store;
+   vvec_store.SetLength(m+2);
+   double *vvec = vvec_store.get();
 
-   long *deltavec;
-   deltavec = NTL_NEW_OP long[m+2];
-   if (!deltavec) Error("BKZ_QP: out of memory");
+   UniqueArray<double> yvec_store;
+   yvec_store.SetLength(m+2);
+   double *yvec = yvec_store.get();
+
+   UniqueArray<double> uvec_store;
+   uvec_store.SetLength(m+2);
+   double *uvec = uvec_store.get();
+
+   UniqueArray<double> utildavec_store;
+   utildavec_store.SetLength(m+2);
+   double *utildavec = utildavec_store.get();
+
+   UniqueArray<long> Deltavec_store;
+   Deltavec_store.SetLength(m+2);
+   long *Deltavec = Deltavec_store.get();
+
+   UniqueArray<long> deltavec_store;
+   deltavec_store.SetLength(m+2);
+   long *deltavec = deltavec_store.get();;
 
    mat_ZZ Ulocal;
    mat_ZZ *U;
@@ -1870,7 +1802,7 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                }
             }
    
-            if (s == 0) Error("BKZ_QP: internal error");
+            if (s == 0) LogicError("BKZ_QP: internal error");
    
             if (s > 0) {
                // special case
@@ -1888,7 +1820,7 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                // cerr << "special case\n";
                new_m = ll_LLL_QP(B, U, delta, 0, check, 
                                 B1, mu, b, c, h, jj, quit);
-               if (new_m != h) Error("BKZ_QP: internal error");
+               if (new_m != h) LogicError("BKZ_QP: internal error");
                if (quit) break;
             }
             else {
@@ -1927,14 +1859,14 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                b[jj] = InnerProduct(B1[jj], B1[jj], n);
                CheckFinite(&b[jj]);
       
-               if (b[jj] == 0) Error("BKZ_QP: internal error"); 
+               if (b[jj] == 0) LogicError("BKZ_QP: internal error"); 
       
                // remove linear dependencies
    
                // cerr << "general case\n";
                new_m = ll_LLL_QP(B, U, delta, 0, 0, B1, mu, b, c, kk+1, jj, quit);
               
-               if (new_m != kk) Error("BKZ_QP: internal error"); 
+               if (new_m != kk) LogicError("BKZ_QP: internal error"); 
 
                // remove zero vector
       
@@ -1963,7 +1895,7 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
                   new_m = ll_LLL_QP(B, U, delta, 0, check, 
                                    B1, mu, b, c, h, h, quit);
    
-                  if (new_m != h) Error("BKZ_QP: internal error");
+                  if (new_m != h) LogicError("BKZ_QP: internal error");
                   if (quit) break;
                }
             }
@@ -1980,7 +1912,7 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
             if (!clean) {
                new_m = 
                   ll_LLL_QP(B, U, delta, 0, check, B1, mu, b, c, h, h, quit);
-               if (new_m != h) Error("BKZ_QP: internal error");
+               if (new_m != h) LogicError("BKZ_QP: internal error");
                if (quit) break;
             }
    
@@ -2021,29 +1953,6 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ* UU, quad_float delta,
       *UU = *U;
    }
 
-   for (i = 1; i <= m_orig+1; i++) {
-      delete [] B1[i];
-   }
-
-   delete [] B1;
-
-   for (i = 1; i <= m_orig+1; i++) {
-      delete [] mu[i];
-   }
-
-   delete [] mu;
-
-
-   delete [] c;
-   delete [] b;
-   delete [] ctilda;
-   delete [] vvec;
-   delete [] yvec;
-   delete [] uvec;
-   delete [] utildavec;
-   delete [] Deltavec;
-   delete [] deltavec;
-
    return m;
 }
 
@@ -2058,8 +1967,8 @@ long BKZ_QP1(mat_ZZ& BB, mat_ZZ& UU, double delta,
    }
 
 
-   if (delta < 0.50 || delta >= 1) Error("BKZ_QP: bad delta");
-   if (beta < 2) Error("BKZ_QP: bad block size");
+   if (delta < 0.50 || delta >= 1) LogicError("BKZ_QP: bad delta");
+   if (beta < 2) LogicError("BKZ_QP: bad block size");
 
    return BKZ_QP1(BB, &UU, to_quad_float(delta), beta, prune, check);
 }
@@ -2076,8 +1985,8 @@ long BKZ_QP1(mat_ZZ& BB, double delta,
 
 
 
-   if (delta < 0.50 || delta >= 1) Error("BKZ_QP: bad delta");
-   if (beta < 2) Error("BKZ_QP: bad block size");
+   if (delta < 0.50 || delta >= 1) LogicError("BKZ_QP: bad delta");
+   if (beta < 2) LogicError("BKZ_QP: bad block size");
 
    return BKZ_QP1(BB, 0, to_quad_float(delta), beta, prune, check);
 }
