@@ -1,23 +1,26 @@
 
 
-#include <stdio.h>
-#include <limits.h>
-#include <float.h>
-#include <stdlib.h>
-#include <math.h>
+#include <cstdio>
+#include <climits>
+#include <cfloat>
+#include <cstdlib>
+#include <cmath>
+#include <limits>
 
 
 #include <NTL/version.h>
 
-
-#if (defined(__GNUC__) && (defined(__i386__) || defined(__i486__) || defined(__i586__)))
+using namespace std;
 
 
-#define AutoFix (1)
+#if (defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__)))
+
+
+#define GNUC_INTEL (1)
 
 #else
 
-#define AutoFix (0)
+#define GNUC_INTEL (0)
 
 #endif
 
@@ -31,6 +34,7 @@ unsigned long val_ulong(unsigned long x);
 size_t val_size_t(size_t x);
 
 double val_double(double x);
+long double val_ldouble(double x);
 
 void touch_int(int* x);
 void touch_uint(unsigned int* x);
@@ -41,34 +45,47 @@ void touch_ulong(unsigned long* x);
 void touch_size_t(size_t* x);
 
 void touch_double(double* x);
+void touch_ldouble(long double* x);
+
+double sum_double(double *x, long n);
+
+double fma_test(double a, double b, double c);
 
 
+double power2(long k);
 
 
-double power2(long k)
+long FMADetected(long dp)
 {
-   long i;
-   double res;
+   double x = power2(0) + power2(dp-1);
+   double y = power2(0) + power2(dp-1);
 
-   res = 1;
+   touch_double(&x);
+   touch_double(&y);
 
-   for (i = 1; i <= k; i++)
-      res = res * 2;
+   double z = x*y;
+   touch_double(&z);
+   z = -z;
+   touch_double(&z);
 
-   return res;
+   double lo = fma_test(x, y, z);
+   return lo != 0;
 }
 
-
 long DoubleRounding(long dp)
 {
    double a = power2(dp-1) + 1;
    double b = (power2(dp)-1)/power2(dp+1);
-   register double x = a + b;
-   double y = x;
 
-   touch_double(&y);
+   double vec[2];
+   vec[0] = a;
+   vec[1] = b;
 
-   if (y != power2(dp-1) + 1)
+   double sum = sum_double(vec, 2);
+
+   touch_double(&sum);
+
+   if (sum != a)
       return 1;
    else 
       return 0; 
@@ -76,7 +93,7 @@ long DoubleRounding(long dp)
 
 
 
-long DoublePrecision(void)
+long DoublePrecision()
 {
    double eps, one, res;
    long k;
@@ -99,53 +116,34 @@ long DoublePrecision(void)
    return k;
 }
 
-long DoublePrecision1(void)
+long LongDoublePrecision()
 {
-   double eps, one, res;
+   long double eps, one, res;
    long k;
 
-   one = val_double(1.0);
-   eps = val_double(1.0);
+   one = val_ldouble(1.0);
+   eps = val_ldouble(1.0);
 
    k = 0;
 
    do {
-      register double tmp;
+      long double tmp;
 
       k++;
       eps *= 1.0/2.0;
       tmp = 1.0 + eps;
+      touch_ldouble(&tmp);
       res = tmp - one;
-   } while (res == eps);
+   } while (res == eps && k < 500);
 
+   // if k >= 500, then most likely this is some
+   // weird double/double implementation.
+   // We also check what numeric_limits says about long doubles.
+
+   if (k >= 500 || !numeric_limits<long double>::is_iec559) k = 0;
    return k;
 }
 
-
-union d_or_rep {
-   double d;
-   unsigned long rep[2];
-};
-
-long RepTest(void)
-{
-   union d_or_rep v;
-
-   if (sizeof(double) != 2*sizeof(long))
-      return 0;
-
-   v.rep[0] = v.rep[1] = 0;
-
-   v.d = 565656565656.0;
-
-   if (v.rep[0] == 0x42607678 && v.rep[1] == 0x46f30000)
-      return 1;
-   else if (v.rep[1] == 0x42607678 && v.rep[0] == 0x46f30000)
-      return -1;
-   else
-      return 0;
-}
-
 void print2k(FILE *f, long k, long bpl)
 {
    long m, l;
@@ -181,6 +179,43 @@ void print2k(FILE *f, long k, long bpl)
 }
 
 
+void print2k_WD(FILE *f, long k, long bpl)
+{
+   long m, l;
+   long first;
+
+   if (k <= 0) {
+      fprintf(f, "(wide_double(1L))");
+      return;
+   }
+
+   m = bpl - 2;
+   first = 1;
+
+   fprintf(f, "(");
+
+   while (k > 0) {
+      if (k > m)
+         l = m;
+      else
+         l = k;
+
+      k = k - l;
+
+      if (first)
+         first = 0;
+      else 
+         fprintf(f, "*");
+
+      fprintf(f, "(wide_double(1L<<%ld))", l);
+   }
+
+   fprintf(f, ")");
+}
+
+
+
+
 
 
 
@@ -712,14 +747,16 @@ void print_BB_rev_code(FILE *f, long n)
 
 
 
-char *yn_vec[2] = { "no", "yes" }; 
+const char *yn_vec[2] = { "no", "yes" }; 
 
 
 
 int main()
 {
-   long bpl, bpi, bpt, rs_arith, nbits, single_mul_ok;
-   long dp, dp1, dr;
+   long bpl, bpi, bpt, rs_arith, nbits, wnbits;
+   long dp, dr;
+   long fma_detected;
+   long ldp;
    FILE *f;
    long warnings = 0;
 
@@ -794,28 +831,35 @@ int main()
    }
 
    /*
-    * check that ints are bigger than chars.
+    * check that there are 8 bits in a char. This is a POSIX requirement.
     */
 
-   if (bpi <= CHAR_BIT) {
-      fprintf(stderr, "BAD NEWS: int type must be longer than char type.\n");
+   if (CHAR_BIT != 8) {
+      fprintf(stderr, "BAD NEWS: char type must have 8 bits.\n");
       return 1;
    }
 
 
+   /*
+    * check that bpi is a multiple of 8.
+    */
+
+   if (bpi % 8 != 0) {
+      fprintf(stderr, "BAD NEWS: int type must be multiple of 8 bits.\n");
+      return 1;
+   }
+
 
    /*
     * check that bpl is a multiple of 8.
     */
 
    if (bpl % 8 != 0) {
-      fprintf(stderr, "BAD NEWS: word size must be multiple of 8 bits.\n");
+      fprintf(stderr, "BAD NEWS: long type must be multiple of 8 bits.\n");
       return 1;
    }
 
 
-
-
    /*
     * check if width of signed versions of int and long agree with that of
     * the unsigned versions, and that negative numbers are represented
@@ -869,6 +913,9 @@ int main()
       return 1;
    }
 
+ 
+
+
 
 
    /*
@@ -947,20 +994,39 @@ int main()
     * This test almost always yields the correct result --- if not,
     * you will have to set the NTL_EXT_DOUBLE in "mach_desc.h"
     * by hand.
-    * 
-    * The test effectively proves that in-register doubles are wide
-    * if dp1 > dp || dr.
     */
 
 
-   dp1 = DoublePrecision1();
    dr = DoubleRounding(dp);
 
 
+   /* 
+    * Next, we check if the platform uses FMA (fused multiply add),
+    * even across statement boundaries.
+    */ 
+
+   fma_detected = FMADetected(dp);
+
+
+
+   /* 
+    * Next, we test the precision of long doubles.
+    * If long doubles don't look good or useful, ldp == 0.
+    * Right now, we ony enable long double usage on Intel/gcc
+    * platforms.
+    */
+
+   ldp = LongDoublePrecision(); 
+   if (ldp <= dp || !GNUC_INTEL) ldp = 0;
+
+   // Disable if it looks like rounding doesn't work right
+   if (((long) val_ldouble(1.75)) != 1L) ldp = 0;
+   if (((long) val_ldouble(-1.75)) != -1L) ldp = 0;
+
+
    /*
-    * Set nbits --- the default radix size for NTL's "built in"
-    * long integer arithmetic.
-    *
+    * Set nbits = min(bpl-2, dp-3) [and even]
+
     *  Given the minimum size of blp and dp, the smallest possible
     *  value of nbits is 30.
     */
@@ -973,23 +1039,25 @@ int main()
 
    if (nbits % 2 != 0) nbits--;
 
-
-   /* 
-    * We next test if the NTL_SINGLE_MUL option is valid.  This test is
-    * inherently DIRTY (i.e., the behavior of the test itself is not well
-    * defined according to the standard), but in practice should not cause any
-    * bad behavior, especially if the NTL_SINGLE_MUL option is never used.
-    * This option is anyway considered fairly obsolete, and if desired, one may
-    * change the following "if 1" to "if 0" and avoid performing this test
-    * altogether.
+   /*
+    * Set wnbits = min(bpl-2, ldp-2) [and even]
     */
 
-#if 1
-   single_mul_ok = RepTest();
-#else
-   single_mul_ok = 0;
-#endif
+   if (ldp) {
+      if (bpl-2 < ldp-2)
+         wnbits = bpl-2;
+      else
+         wnbits = ldp-2;
 
+      if (wnbits % 2 != 0) wnbits--;
+   }
+   else {
+      wnbits = nbits;
+   }
+
+   if (wnbits <= nbits) ldp = 0;
+   // disable long doubles if it doesn't increase nbits...
+   // (for example, on 32-bit machines)
 
    /*
     * That's it!  All tests have passed.
@@ -1002,12 +1070,13 @@ int main()
    fprintf(stderr, "bits per size_t = %ld\n", bpt);
    fprintf(stderr, "arith right shift = %s\n", yn_vec[rs_arith]);
    fprintf(stderr, "double precision = %ld\n", dp);
+   fprintf(stderr, "long double precision = %ld\n", ldp);
    fprintf(stderr, "NBITS (maximum) = %ld\n", nbits);
-   fprintf(stderr, "single mul ok = %s\n", yn_vec[single_mul_ok != 0]);
-   fprintf(stderr, "register double precision = %ld\n", dp1);
+   fprintf(stderr, "WNBITS (maximum) = %ld\n", wnbits);
    fprintf(stderr, "double rounding detected = %s\n", yn_vec[dr]);  
+   fprintf(stderr, "FMA detected = %s\n", yn_vec[fma_detected]);  
 
-   if (((dp1 > dp) || dr) && AutoFix)
+   if (dr && GNUC_INTEL)
       fprintf(stderr, "-- auto x86 fix\n");
 
    if (dp != 53) {
@@ -1031,7 +1100,7 @@ int main()
 
 #endif
 
-   if (((dp1 > dp) || dr) && !AutoFix) {
+   if (dr && !GNUC_INTEL) {
       warnings = 1;
       fprintf(stderr, "\n\nWARNING:\n\n");
       fprintf(stderr, "This platform has extended double precision registers.\n");
@@ -1078,16 +1147,40 @@ int main()
    fprintf(f, "#define NTL_BITS_PER_SIZE_T (%ld)\n", bpt);
    fprintf(f, "#define NTL_ARITH_RIGHT_SHIFT (%ld)\n", rs_arith);
    fprintf(f, "#define NTL_NBITS_MAX (%ld)\n", nbits);
+   fprintf(f, "#define NTL_WNBITS_MAX (%ld)\n", wnbits);
    fprintf(f, "#define NTL_DOUBLE_PRECISION (%ld)\n", dp);
    fprintf(f, "#define NTL_FDOUBLE_PRECISION ");
    print2k(f, dp-1, bpl);
    fprintf(f, "\n");
+
+
+   if (ldp) {
+      fprintf(f, "#define NTL_LONGDOUBLE_OK (1)\n"); 
+      fprintf(f, "#define NTL_LONGDOUBLE_PRECISION (%ld)\n", ldp); 
+      fprintf(f, "#define NTL_WIDE_DOUBLE_LDP ");
+      print2k_WD(f, ldp-1, bpl);
+      fprintf(f, "\n");
+      fprintf(f, "#define NTL_WIDE_DOUBLE_DP ");
+      print2k_WD(f, dp-1, bpl);
+      fprintf(f, "\n");
+   }
+   else {
+      fprintf(f, "#define NTL_LONGDOUBLE_OK (0)\n"); 
+      fprintf(f, "#define NTL_WIDE_DOUBLE_DP ");
+      print2k_WD(f, dp-1, bpl);
+      fprintf(f, "\n");
+   }
+
+
    fprintf(f, "#define NTL_QUAD_FLOAT_SPLIT (");
    print2k(f, dp - (dp/2), bpl);
    fprintf(f, "+1.0)\n");
-   fprintf(f, "#define NTL_EXT_DOUBLE (%d)\n", ((dp1 > dp) || dr));
-   fprintf(f, "#define NTL_SINGLE_MUL_OK (%d)\n", single_mul_ok != 0);
-   fprintf(f, "#define NTL_DOUBLES_LOW_HIGH (%d)\n\n\n", single_mul_ok < 0);
+   fprintf(f, "#define NTL_EXT_DOUBLE (%ld)\n", dr);
+
+   fprintf(f, "#define NTL_FMA_DETECTED (%ld)\n", fma_detected);
+
+
+
    print_BB_mul_code(f, bpl);
    print_BB_sqr_code(f, bpl);
    print_BB_rev_code(f, bpl);
