@@ -2,7 +2,7 @@
 #include <NTL/ZZ_pX.h>
 #include <NTL/FFT.h>
 
-#include <stdio.h>
+#include <cstdio>
 
 NTL_CLIENT
 
@@ -40,15 +40,14 @@ double clean_data(double *t)
 void print_flag()
 {
 
+#if defined(NTL_FFT_LAZYMUL)
+printf("FFT_LAZYMUL ");
+#endif
 
-#if defined(NTL_SPMM_UL)
-printf("SPMM_UL ");
-#elif defined(NTL_SPMM_ULL)
+#if defined(NTL_SPMM_ULL)
 printf("SPMM_ULL ");
 #elif defined(NTL_SPMM_ASM)
 printf("SPMM_ASM ");
-#else
-printf("DEFAULT ");
 #endif
 
 #if defined(NTL_AVOID_BRANCHING)
@@ -59,11 +58,6 @@ printf("AVOID_BRANCHING ");
 printf("FFT_BIGTAB ");
 #endif
 
-#if defined(NTL_FFT_LAZYMUL)
-printf("FFT_LAZYMUL ");
-#endif
-
-
 printf("\n");
 
 }
@@ -82,6 +76,8 @@ int main()
 
 #endif
 
+   SetSeed(ZZ(0));
+
 
    long n, k;
 
@@ -91,6 +87,7 @@ int main()
    ZZ p;
 
    RandomLen(p, k);
+   if (!IsOdd(p)) p++;
 
 
    ZZ_p::init(p);         // initialization
@@ -131,34 +128,64 @@ int main()
    }
 
    double t;
-   long i;
+   long i, j;
    long iter;
 
-   const int nprimes = 10;
+   const int nprimes = 30;
+   const long L = 12; 
+   const long N = 1L << L;
    long r;
    
 
    for (r = 0; r < nprimes; r++) UseFFTPrime(r);
 
-   vec_long aa[nprimes], AA[nprimes];
+   vec_long A1[nprimes], A2[nprimes];
+   vec_long B1[nprimes], B2[nprimes];
 
    for (r = 0; r < nprimes; r++) {
-      aa[r].SetLength(4096);
-      AA[r].SetLength(4096);
+      A1[r].SetLength(N);
+      A2[r].SetLength(N);
+      B1[r].SetLength(N);
+      B2[r].SetLength(N);
 
-      for (i = 0; i < 4096; i++)
-         aa[r][i] = RandomBnd(FFTPrime[r]);
+      for (i = 0; i < N; i++) {
+         A1[r][i] = RandomBnd(GetFFTPrime(r));
+         A2[r][i] = RandomBnd(GetFFTPrime(r));
+      }
+   }
 
-
-      FFTFwd(AA[r].elts(), aa[r].elts(), 12, r);
+   for (r = 0; r < nprimes; r++) {
+      long *A1p = A1[r].elts();
+      long *A2p = A2[r].elts();
+      long *B1p = B1[r].elts();
+      long *B2p = B2[r].elts();
+      long q = GetFFTPrime(r);
+      mulmod_t qinv = GetFFTPrimeInv(r);
+ 
+      FFTFwd(B1p, A1p, L, r);
+      FFTFwd(B2p, A2p, L, r);
+      for (i = 0; i < N; i++) B1p[i] = NormalizedMulMod(B1p[i], B2p[i], q, qinv);
+      FFTRev1(B1p, B1p, L, r);
    }
 
    iter = 1;
 
    do {
      t = GetTime();
-     for (i = 0; i < iter; i++) {
-        for (r = 0; r < nprimes; r++) FFTFwd(AA[r].elts(), aa[r].elts(), 12, r);
+     for (j = 0; j < iter; j++) {
+        for (r = 0; r < nprimes; r++) {
+           long *A1p = A1[r].elts();
+           long *A2p = A2[r].elts();
+           long *B1p = B1[r].elts();
+           long *B2p = B2[r].elts();
+           long q = GetFFTPrime(r);
+           mulmod_t qinv = GetFFTPrimeInv(r);
+
+           FFTFwd(B1p, A1p, L, r);
+           FFTFwd(B2p, A2p, L, r);
+           for (i = 0; i < N; i++) B1p[i] = NormalizedMulMod(B1p[i], B2p[i], q, qinv);
+           FFTRev1(B1p, B1p, L, r);
+        }
      }
      t = GetTime() - t;
      iter = 2*iter;
@@ -166,7 +193,7 @@ int main()
 
    iter = iter/2;
 
-   iter = long((2/t)*iter) + 1;
+   iter = long((3/t)*iter) + 1;
 
 
    double tvec[5];
@@ -174,8 +201,20 @@ int main()
 
    for (w = 0; w < 5; w++) {
      t = GetTime();
-     for (i = 0; i < iter; i++) {
-        for (r = 0; r < nprimes; r++) FFTFwd(AA[r].elts(), aa[r].elts(), 12, r);
+     for (j = 0; j < iter; j++) {
+        for (r = 0; r < nprimes; r++) {
+           long *A1p = A1[r].elts();
+           long *A2p = A2[r].elts();
+           long *B1p = B1[r].elts();
+           long *B2p = B2[r].elts();
+           long q = GetFFTPrime(r);
+           mulmod_t qinv = GetFFTPrimeInv(r);
+
+           FFTFwd(B1p, A1p, L, r);
+           FFTFwd(B2p, A2p, L, r);
+           for (i = 0; i < N; i++) B1p[i] = NormalizedMulMod(B1p[i], B2p[i], q, qinv);
+           FFTRev1(B1p, B1p, L, r);
+        }
      }
      t = GetTime() - t;
      tvec[w] = t;
