@@ -1,4 +1,5 @@
 
+#include <NTL/ZZX.h>
 #include <NTL/ZZ_pX.h>
 #include <NTL/lzz_pX.h>
 #include <NTL/GF2X.h>
@@ -100,31 +101,120 @@ void GF2X_time()
 }
 
 
+
+ZZX KarMul(const ZZX& a, const ZZX& b)
+{
+   ZZX res;
+   KarMul(res, a, b);
+   return res;
+}
+
+
+
 int main()
 {
+   SetSeed(ZZ(0));
 
 
    cerr << "This is NTL version " << NTL_VERSION << "\n"; 
 
+   cerr << "Hardware charactersitics:\n";
+   cerr << "NTL_BITS_PER_LONG = " << NTL_BITS_PER_LONG << "\n";
+   cerr << "NTL_ZZ_NBITS = " << NTL_ZZ_NBITS << "\n";
+   cerr << "NTL_SP_NBITS = " << NTL_SP_NBITS << "\n";
+
+#ifdef NTL_HAVE_LL_TYPE
+   cerr << "NTL_HAVE_LL_TYPE\n";
+#endif
+
+#ifdef NTL_HAVE_BUILTIN_CLZL
+   cerr << "NTL_HAVE_BUILTIN_CLZL\n";
+#endif
+
+#ifdef NTL_HAVE_AVX
+   cerr << "NTL_HAVE_AVX\n";
+#endif
+
+#ifdef NTL_HAVE_FMA
+   cerr << "NTL_HAVE_FMA\n";
+#endif
+
+
+
+#ifdef NTL_LONGDOUBLE_SP_MULMOD
+   cerr << "NTL_LONGDOUBLE_SP_MULMOD\n";
+#endif
+
+#ifdef NTL_LONGLONG_SP_MULMOD
+   cerr << "NTL_LONGLONG_SP_MULMOD\n";
+#endif
+
+   cerr << "\n";
+
+   
+
+
    cerr << "Basic Configuration Options:\n";
 
 
-#ifdef NTL_STD_CXX
-   cout << "NTL_STD_CXX\n";
+
+#ifdef NTL_LEGACY_NO_NAMESPACE
+   cerr << "NTL_LEGACY_NO_NAMESPACE\n";
 #endif
 
-#ifdef NTL_PSTD_NNS
-   cerr << "NTL_PSTD_NNS\n";
+
+#ifdef NTL_LEGACY_INPUT_ERROR
+   cerr << "NTL_LEGACY_INPUT_ERROR\n";
 #endif
 
-#ifdef NTL_PSTD_NHF
-   cerr << "NTL_PSTD_NHF\n";
+
+#ifdef NTL_THREADS
+   cerr << "NTL_THREADS\n";
 #endif
 
-#ifdef NTL_PSTD_NTN
-   cerr << "NTL_PSTD_NTN\n";
+#ifdef NTL_DISABLE_TLS_HACK
+   cerr << "NTL_DISABLE_TLS_HACK\n";
 #endif
 
+#ifdef NTL_ENABLE_TLS_HACK
+   cerr << "NTL_ENABLE_TLS_HACK\n";
+#endif
+
+
+#ifdef NTL_EXCEPTIONS
+   cerr << "NTL_EXCEPTIONS\n";
+#endif
+
+#ifdef NTL_THREAD_BOOST
+   cerr << "NTL_THREAD_BOOST\n";
+#endif
+
+
+#ifdef NTL_LEGACY_SP_MULMOD
+   cerr << "NTL_LEGACY_SP_MULMOD\n";
+#endif
+
+
+#ifdef NTL_DISABLE_LONGDOUBLE
+   cerr << "NTL_DISABLE_LONGDOUBLE\n";
+#endif
+
+
+#ifdef NTL_DISABLE_LONGLONG
+   cerr << "NTL_DISABLE_LONGLONG\n";
+#endif
+
+#ifdef NTL_DISABLE_LL_ASM
+   cerr << "NTL_DISABLE_LL_ASM\n";
+#endif
+
+#ifdef NTL_MAXIMIZE_SP_NBITS
+   cerr << "NTL_MAXIMIZE_SP_NBITS\n";
+#endif
+
+
+
+
 #ifdef NTL_GMP_LIP
    cerr << "NTL_GMP_LIP\n";
 #endif
@@ -135,6 +225,7 @@ int main()
 #endif
 
 
+
 #ifdef NTL_LONG_LONG_TYPE
    cerr << "NTL_LONG_LONG_TYPE: ";
    cerr << make_string(NTL_LONG_LONG_TYPE) << "\n";
@@ -188,10 +279,6 @@ cerr << "Performance Options:\n";
    cerr << "NTL_AVOID_FLOAT\n";
 #endif
 
-#ifdef NTL_SPMM_UL
-   cerr << "NTL_SPMM_UL\n";
-#endif
-
 
 #ifdef NTL_SPMM_ULL
    cerr << "NTL_SPMM_ULL\n";
@@ -211,22 +298,31 @@ cerr << "Performance Options:\n";
 
 
 #ifdef NTL_FFT_BIGTAB
-   cout << "NTL_FFT_BIGTAB\n";
+   cerr << "NTL_FFT_BIGTAB\n";
 #endif
 
 #ifdef NTL_FFT_LAZYMUL
-   cout << "NTL_FFT_LAZYMUL\n";
+   cerr << "NTL_FFT_LAZYMUL\n";
 #endif
 
 
-
-
-
 #ifdef NTL_TBL_REM
    cerr << "NTL_TBL_REM\n";
 #endif
 
 
+#ifdef NTL_TBL_REM_LL
+   cerr << "NTL_TBL_REM_LL\n";
+#endif
+
+#ifdef NTL_CRT_ALTCODE
+   cerr << "NTL_CRT_ALTCODE\n";
+#endif
+
+#ifdef NTL_CRT_ALTCODE_SMALL
+   cerr << "NTL_CRT_ALTCODE_SMALL\n";
+#endif
+
 #ifdef NTL_GF2X_ALTCODE
    cerr << "NTL_GF2X_ALTCODE\n";
 #endif
@@ -235,66 +331,52 @@ cerr << "Performance Options:\n";
    cerr << "NTL_GF2X_ALTCODE1\n";
 #endif
 
-
 #ifdef NTL_GF2X_NOINLINE
    cerr << "NTL_GF2X_NOINLINE\n";
 #endif
 
+#ifdef NTL_PCLMUL
+   cerr << "NTL_PCLMUL\n";
+#endif
+
+
    cerr << "\n\n";
 
-   cerr << "running tests...";
+   cerr << "running tests";
 
-   long n, k;
+   long n, k, i;
 
-   n = 200;
-   k = 10*NTL_ZZ_NBITS;
+   n = 250;
+   k = 16000;
 
    ZZ p;
 
-   GenPrime(p, k);
+
+   for (i = 0; i < 15; i++) {
+      // cerr << n << "/" << k; 
+      cerr << ".";
+      RandomLen(p, k);
+      ZZ_p::init(p);  
+    
+
+      ZZ_pX a, b, c, c1;
 
 
-   ZZ_p::init(p);         // initialization
+      random(a, n);
+      random(b, n);
 
-   ZZ_pX f, g, h, r1, r2, r3;
+      FFTMul(c, a, b);
+      //cerr << ZZ_pInfo->FFTInfo->NumPrimes;
 
-   random(g, n);    // g = random polynomial of degree < n
-   random(h, n);    // h =             "   "
-   random(f, n);    // f =             "   "
+      c1 = conv<ZZ_pX>( KarMul( conv<ZZX>(a), conv<ZZX>(b) ) );
 
-   // SetCoeff(f, n);  // Sets coefficient of X^n to 1
-   
-   ZZ_p lc;
+      if (c1 != c) {
+         cerr << "ZZ_pX mul failed!\n";
+         return 1;
+      }
 
-   do {
-      random(lc);
-   } while (IsZero(lc));
-
-   SetCoeff(f, n, lc);
-
-
-   // For doing arithmetic mod f quickly, one must pre-compute
-   // some information.
-
-   ZZ_pXModulus F;
-   build(F, f);
-
-   PlainMul(r1, g, h);  // this uses classical arithmetic
-   PlainRem(r1, r1, f);
-
-   MulMod(r2, g, h, F);  // this uses the FFT
-
-   MulMod(r3, g, h, f);  // uses FFT, but slower
-
-   // compare the results...
-
-   if (r1 != r2) {
-      cerr << "r1 != r2!!\n";
-      return 1;
-   }
-   else if (r1 != r3) {
-      cerr << "r1 != r3!!\n";
-      return 1;
+      n = long(n * 1.35);
+      k = long(k / 1.414);
    }
 
 
@@ -324,7 +406,6 @@ cerr << "Performance Options:\n";
 
    ZZ x1, x2, x3, x4;
    double t;
-   long i;
 
    RandomLen(x1, 1024);
    RandomBnd(x2, x1);
@@ -374,19 +455,21 @@ cerr << "Performance Options:\n";
    RandomLen(p, k);
 
    ZZ_p::init(p);
-   ZZ_pInfo->check();
+   if (!IsOdd(p)) p++;
 
    ZZ_pX j1, j2, j3;
 
    random(j1, n);
    random(j2, n);
 
+   mul(j3, j1, j2);
+
    t = GetTime();
-   for (i = 0; i < 20; i++) mul(j3, j1, j2);
+   for (i = 0; i < 500; i++) mul(j3, j1, j2);
    t = GetTime()-t;
 
    cerr << "time to multiply degree 1023 polynomials\n   modulo a 1024-bit number: ";
-   cerr << (t/20) << "s";
+   cerr << (t/500) << "s";
    cerr << "\n";
 
    GF2X_time();
