@@ -6,7 +6,230 @@
 
 NTL_START_IMPL
 
+/*****************************************************************/
 
+// Fast CRT routines
+//   - we could perhaps move these out of here to make them
+//   - more widely accessible
+
+
+struct FastCRTHelper {
+
+   ZZ prod;
+   ZZ prod_half;
+   
+   long nprimes;
+
+   long nlevels;
+   long veclen;
+   long nblocks;   // number of nodes in the last level
+   long start_last_level;   // index of first item in last level
+
+   Vec<long> nprimes_vec;  // length == veclen
+   Vec<long> first_vec;    // length == nblocks
+   Vec<ZZ> prod_vec;       // length == veclen
+   Vec<long> coeff_vec;    // length == nprimes, coeff_vec[i] = (prod/p_i)^{-1} mod p_i
+
+
+   Vec<ZZ> tmp_vec;        // length == nlevels
+
+
+   ZZ tmp1, tmp2, tmp3;
+
+   FastCRTHelper(long bound, long thresh); 
+
+   void fill_nprimes_vec(long index); 
+   void fill_prod_vec(long index);
+
+   void reduce_aux(const ZZ& value, ZZ **remainders, long index, long level);
+   void reconstruct_aux(ZZ& value, ZZ **remainders, long index, long level);
+
+   void reduce(const ZZ& value, ZZ **remainders);
+   void reconstruct(ZZ& value, ZZ **remainders);
+
+};
+
+void FastCRTHelper::fill_nprimes_vec(long index) 
+{
+   long left, right;
+   left = 2*index + 1;
+   right = 2*index + 2;
+   if (left >= veclen) return;
+
+   nprimes_vec[left] = nprimes_vec[index]/2;
+   nprimes_vec[right] = nprimes_vec[index] - nprimes_vec[left];
+   fill_nprimes_vec(left);
+   fill_nprimes_vec(right);
+}
+
+void FastCRTHelper::fill_prod_vec(long index)
+{
+   long left, right;
+   left = 2*index + 1;
+   right = 2*index + 2;
+   if (left >= veclen) return;
+
+   fill_prod_vec(left);
+   fill_prod_vec(right);
+   mul(prod_vec[index], prod_vec[left], prod_vec[right]);
+}
+
+FastCRTHelper::FastCRTHelper(long bound, long thresh) 
+{
+   // assumes bound >= 1, thresh >= 1
+
+   prod = 1;
+   for (nprimes = 0; NumBits(prod) <= bound; nprimes++) {
+      UseFFTPrime(nprimes);
+      prod *= GetFFTPrime(nprimes);
+   }
+
+   RightShift(prod_half, prod, 1);
+
+   long sz = nprimes;
+   nlevels = 1;
+   while (sz > thresh) {
+      sz = sz/2;
+      nlevels++;
+   }
+
+   veclen = (1L << nlevels) - 1;
+   nblocks = 1L << (nlevels-1);
+   start_last_level = (1L << (nlevels-1)) - 1;
+
+   nprimes_vec.SetLength(veclen);
+   nprimes_vec[0] = nprimes;
+
+   fill_nprimes_vec(0);
+
+   long k, i;
+   first_vec.SetLength(nblocks+1);
+
+   first_vec[0] = 0;
+   for (k = 1; k <= nblocks; k++)
+      first_vec[k] = first_vec[k-1] + nprimes_vec[start_last_level + k-1];
+
+   prod_vec.SetLength(veclen);
+
+   // fill product leaves
+   for (k = 0; k < nblocks; k++) {
+      prod_vec[start_last_level + k] = 1;
+      for (i = first_vec[k]; i < first_vec[k+1]; i++) {
+         prod_vec[start_last_level + k] *= GetFFTPrime(i);
+      }
+   }
+
+   // fill rest of product trees
+   fill_prod_vec(0);
+
+   ZZ t1;
+   long tt;
+
+   // fill coeff_vec: simple, quadratic time for now
+   // not really essential to speed this up
+   coeff_vec.SetLength(nprimes);
+   for (long i = 0; i < nprimes; i++) {
+      long p = GetFFTPrime(i);
+      div(t1, prod, p);
+      tt = rem(t1, p);
+      tt = InvMod(tt, p);
+      coeff_vec[i] = tt;
+   }
+
+   tmp_vec.SetLength(nlevels);
+}
+
+void FastCRTHelper::reduce_aux(const ZZ& value, ZZ **remainders, long index, long level)
+{
+   long left, right;
+   left = 2*index + 1;
+   right = 2*index + 2;
+
+   ZZ *result = 0;
+
+   if (left >= veclen) 
+      result = remainders[index - start_last_level];
+   else
+      result = &tmp_vec[level];
+
+   if (NumBits(value) <= NumBits(prod_vec[index]))
+      *result = value;
+   else {
+      rem(tmp1, value, prod_vec[index]);
+      sub(tmp2, tmp1, prod_vec[index]);
+      if (NumBits(tmp2) < NumBits(tmp1))
+         *result = tmp2;
+      else
+         *result = tmp1;
+   }
+
+   if (left < veclen) {
+      reduce_aux(*result, remainders, left, level+1);
+      reduce_aux(*result, remainders, right, level+1);
+   }
+
+}
+
+void FastCRTHelper::reduce(const ZZ& value, ZZ **remainders)
+{
+   reduce_aux(value, remainders, 0, 0);
+}
+
+void FastCRTHelper::reconstruct_aux(ZZ& value, ZZ **remainders, long index, long level)
+{
+   long left, right;
+   left = 2*index + 1;
+   right = 2*index + 2;
+
+   if (left >= veclen) {
+      value = *remainders[index - start_last_level];
+      return;
+   }
+
+   reconstruct_aux(tmp_vec[level], remainders, left, level+1);
+   reconstruct_aux(tmp1, remainders, right, level+1);
+
+   mul(tmp2, tmp_vec[level], prod_vec[right]);
+   mul(tmp3, tmp1, prod_vec[left]);
+   add(value, tmp2, tmp3);
+}
+
+void FastCRTHelper::reconstruct(ZZ& value, ZZ **remainders)
+{
+   reconstruct_aux(tmp1, remainders, 0, 0);
+   rem(tmp1, tmp1, prod);
+   if (tmp1 > prod_half)
+      sub(tmp1, tmp1, prod);
+
+   value = tmp1;
+}
+
+
+
+
+
+
+
+
+
+/*****************************************************************/
+
+
+static
+long MaxSize(const ZZX& a)
+{
+   long res = 0;
+   long n = a.rep.length();
+
+   long i;
+   for (i = 0; i < n; i++) {
+      long t = a.rep[i].size();
+      if (t > res)
+         res = t;
+   }
+
+   return res;
+}
 
 
 
@@ -199,7 +422,7 @@ long CRT(ZZX& gg, ZZ& a, const ZZ_pX& G)
 
 /* Compute a = b * 2^l mod p, where p = 2^n+1. 0<=l<=n and 0<b<p are
    assumed. */
-static void LeftRotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n)
+static void LeftRotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n, ZZ& scratch)
 {
   if (l == 0) {
     if (&a != &b) {
@@ -208,14 +431,13 @@ static void LeftRotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n)
     return;
   }
 
-  /* tmp := upper l bits of b */
-  NTL_ZZRegister(tmp);
-  RightShift(tmp, b, n - l);
+  /* scratch := upper l bits of b */
+  RightShift(scratch, b, n - l);
   /* a := 2^l * lower n - l bits of b */
   trunc(a, b, n - l);
   LeftShift(a, a, l);
-  /* a -= tmp */
-  sub(a, a, tmp);
+  /* a -= scratch */
+  sub(a, a, scratch);
   if (sign(a) < 0) {
     add(a, a, p);
   }
@@ -223,7 +445,7 @@ static void LeftRotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n)
 
 
 /* Compute a = b * 2^l mod p, where p = 2^n+1. 0<=p<b is assumed. */
-static void Rotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n)
+static void Rotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n, ZZ& scratch)
 {
   if (IsZero(b)) {
     clear(a);
@@ -239,9 +461,9 @@ static void Rotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n)
 
   /* a = b * 2^l mod p */
   if (l < n) {
-    LeftRotate(a, b, l, p, n);
+    LeftRotate(a, b, l, p, n, scratch);
   } else {
-    LeftRotate(a, b, l - n, p, n);
+    LeftRotate(a, b, l - n, p, n, scratch);
     SubPos(a, p, a);
   }
 }
@@ -252,12 +474,13 @@ static void Rotate(ZZ& a, const ZZ& b, long l, const ZZ& p, long n)
    p = 2^n+1, w = 2^r mod p is a primitive (2^l)th root of
    unity. Returns a(1),a(w),...,a(w^{2^l-1}) mod p in bit-reverse
    order. */
-static void fft(vec_ZZ& a, long r, long l, const ZZ& p, long n)
+static void fft(ZZVec& a, long r, long l, const ZZ& p, long n)
 {
   long round;
   long off, i, j, e;
   long halfsize;
   ZZ tmp, tmp1;
+  ZZ scratch;
 
   for (round = 0; round < l; round++, r <<= 1) {
     halfsize =  1L << (l - 1 - round);
@@ -278,7 +501,7 @@ static void fft(vec_ZZ& a, long r, long l, const ZZ& p, long n)
           a[off] = tmp1;
         }
         /* a[off + halfsize] = tmp * w^{j2^round} mod p */
-        Rotate(a[off + halfsize], tmp, e, p, n);
+        Rotate(a[off + halfsize], tmp, e, p, n, scratch);
       }
     }
   }
@@ -286,12 +509,13 @@ static void fft(vec_ZZ& a, long r, long l, const ZZ& p, long n)
 
 /* Inverse FFT. r must be the same as in the call to FFT. Result is
    by 2^l too large. */
-static void ifft(vec_ZZ& a, long r, long l, const ZZ& p, long n)
+static void ifft(ZZVec& a, long r, long l, const ZZ& p, long n)
 {
   long round;
   long off, i, j, e;
   long halfsize;
   ZZ tmp, tmp1;
+  ZZ scratch;
 
   for (round = l - 1, r <<= l - 1; round >= 0; round--, r >>= 1) {
     halfsize = 1L << (l - 1 - round);
@@ -301,7 +525,7 @@ static void ifft(vec_ZZ& a, long r, long l, const ZZ& p, long n)
          ( a[off], a[off+halfsize] ) *= ( 1               1             )
                                         ( w^{-j2^round}  -w^{-j2^round} ) */
         /* a[off + halfsize] *= w^{-j2^round} mod p */
-        Rotate(a[off + halfsize], a[off + halfsize], -e, p, n);
+        Rotate(a[off + halfsize], a[off + halfsize], -e, p, n, scratch);
         /* tmp = a[off] - a[off + halfsize] */
         sub(tmp, a[off], a[off + halfsize]);
 
@@ -334,6 +558,11 @@ static void ifft(vec_ZZ& a, long r, long l, const ZZ& p, long n)
 
 void SSMul(ZZX& c, const ZZX& a, const ZZX& b)
 {
+  if (&a == &b) {
+    SSSqr(c, a);
+    return;
+  }
+
   long na = deg(a);
   long nb = deg(b);
 
@@ -364,9 +593,9 @@ void SSMul(ZZX& c, const ZZX& a, const ZZX& b)
   add(p, p, 1);
 
   /* Make coefficients of a and b positive */
-  vec_ZZ aa, bb;
-  aa.SetLength(m2);
-  bb.SetLength(m2);
+  ZZVec aa, bb;
+  aa.SetSize(m2, p.size());
+  bb.SetSize(m2, p.size());
 
   long i;
   for (i = 0; i <= deg(a); i++) {
@@ -385,6 +614,7 @@ void SSMul(ZZX& c, const ZZX& a, const ZZX& b)
     }
   }
 
+
   /* 2m-point FFT's mod p */
   fft(aa, r, l + 1, p, mr);
   fft(bb, r, l + 1, p, mr);
@@ -405,6 +635,7 @@ void SSMul(ZZX& c, const ZZX& a, const ZZX& b)
   }
   
   ifft(aa, r, l + 1, p, mr);
+  ZZ scratch;
 
   /* Retrieve c, dividing by 2m, and subtracting p where necessary */
   c.rep.SetLength(n + 1);
@@ -413,7 +644,7 @@ void SSMul(ZZX& c, const ZZX& a, const ZZX& b)
     ZZ& ci = c.rep[i];
     if (!IsZero(ai)) {
       /* ci = -ai * 2^{mr-l-1} = ai * 2^{-l-1} = ai / 2m mod p */
-      LeftRotate(ai, ai, mr - l - 1, p, mr);
+      LeftRotate(ai, ai, mr - l - 1, p, mr, scratch);
       sub(tmp, p, ai);
       if (NumBits(tmp) >= mr) { /* ci >= (p-1)/2 */
         negate(ci, ai); /* ci = -ai = ci - p */
@@ -446,6 +677,26 @@ double SSRatio(long na, long maxa, long nb, long maxb)
   return double(mr + 1)/double(bound);
 }
 
+
+
+void conv(vec_zz_p& x, const ZZVec& a)
+{
+   long i, n;
+
+   n = a.length();
+   x.SetLength(n);
+
+   zz_p* xp = x.elts();
+   const ZZ* ap = a.elts();
+
+   long p = zz_p::modulus();
+
+   for (i = 0; i < n; i++)
+      xp[i].LoopHole() = rem(a[i], p);
+}
+
+
+
 void HomMul(ZZX& x, const ZZX& a, const ZZX& b)
 {
    if (&a == &b) {
@@ -453,6 +704,7 @@ void HomMul(ZZX& x, const ZZX& a, const ZZX& b)
       return;
    }
 
+
    long da = deg(a);
    long db = deg(b);
 
@@ -461,93 +713,78 @@ void HomMul(ZZX& x, const ZZX& a, const ZZX& b)
       return;
    }
 
-   long bound = 2 + NumBits(min(da, db)+1) + MaxBits(a) + MaxBits(b);
-
-
-   ZZ prod;
-   set(prod);
-
-   long i, nprimes;
-
    zz_pBak bak;
    bak.save();
 
-   for (nprimes = 0; NumBits(prod) <= bound; nprimes++) {
-      if (nprimes >= NumFFTPrimes)
-         zz_p::FFTInit(nprimes);
-      mul(prod, prod, FFTPrime[nprimes]);
+   long bound = 2 + NumBits(min(da, db)+1) + MaxBits(a) + MaxBits(b);
+
+   FastCRTHelper H(bound, 48);      
+
+   long i, j, k;
+
+   Vec<ZZVec> c, aa, bb;
+
+   c.SetLength(H.nblocks);
+   aa.SetLength(H.nblocks);
+   bb.SetLength(H.nblocks);
+
+   long sz_a = max(1, MaxSize(a));
+   long sz_b = max(1, MaxSize(b));
+
+   for (k = 0; k < H.nblocks; k++) {
+      ZZ *prod_vec = &H.prod_vec[H.start_last_level];
+      c[k].SetSize(da+db+1, prod_vec[k].size()+1);
+      aa[k].SetSize(da+1, min(sz_a, prod_vec[k].size()));
+      bb[k].SetSize(db+1, min(sz_b, prod_vec[k].size()));
    }
 
+   Vec<ZZ*> ptr_vec;
+   ptr_vec.SetLength(H.nblocks);
+
+   for (j = 0; j <= da; j++) {
+      for (k = 0; k < H.nblocks; k++) ptr_vec[k] = &aa[k][j];
+      H.reduce(a.rep[j], ptr_vec.elts());
+   }
+
+   for (j = 0; j <= db; j++) {
+      for (k = 0; k < H.nblocks; k++) ptr_vec[k] = &bb[k][j];
+      H.reduce(b.rep[j], ptr_vec.elts());
+   }
 
-   ZZ coeff;
    ZZ t1;
-   long tt;
+     
+   for (k = 0; k < H.nblocks; k++) {
+      for (i = H.first_vec[k]; i < H.first_vec[k+1]; i++) {
+         zz_p::FFTInit(i);
 
-   vec_ZZ c;
+         zz_pX A, B, C;
+         conv(A.rep, aa[k]); A.normalize();
+         conv(B.rep, bb[k]); B.normalize();
+         mul(C, A, B);
 
-   c.SetLength(da+db+1);
-
-   long j;
-
-   for (i = 0; i < nprimes; i++) {
-      zz_p::FFTInit(i);
-      long p = zz_p::modulus();
-
-      div(t1, prod, p);
-      tt = rem(t1, p);
-      tt = InvMod(tt, p);
-      mul(coeff, t1, tt);
-
-      zz_pX A, B, C;
-
-      conv(A, a);
-      conv(B, b);
-      mul(C, A, B);
-
-      long m = deg(C);
-
-      for (j = 0; j <= m; j++) {
-         /* c[j] += coeff*rep(C.rep[j]) */
-         MulAddTo(c[j], coeff, rep(C.rep[j]));
-         // mul(t1, coeff, rep(C.rep[j]));
-         // add(c[j], c[j], t1); 
+         long m = deg(C);
+         long p = zz_p::modulus();
+         long tt = H.coeff_vec[i];
+         mulmod_precon_t ttpinv = PrepMulModPrecon(tt, p);
+         div(t1, H.prod_vec[H.start_last_level+k], p);
+         for (j = 0; j <= m; j++) {
+            long tt1 = MulModPrecon(rep(C.rep[j]), tt, p, ttpinv);
+            MulAddTo(c[k][j], t1, tt1);
+         }
       }
    }
 
    x.rep.SetLength(da+db+1);
-
-   ZZ prod2;
-   RightShift(prod2, prod, 1);
-
    for (j = 0; j <= da+db; j++) {
-      rem(t1, c[j], prod);
-
-      if (t1 > prod2)
-         sub(x.rep[j], t1, prod);
-      else
-         x.rep[j] = t1;
+      for (k = 0; k < H.nblocks; k++) ptr_vec[k] = &c[k][j];
+      H.reconstruct(x.rep[j], ptr_vec.elts());
    }
 
    x.normalize();
-
    bak.restore();
 }
 
-static
-long MaxSize(const ZZX& a)
-{
-   long res = 0;
-   long n = a.rep.length();
 
-   long i;
-   for (i = 0; i < n; i++) {
-      long t = a.rep[i].size();
-      if (t > res)
-         res = t;
-   }
-
-   return res;
-}
 
 
 void mul(ZZX& c, const ZZX& a, const ZZX& b)
@@ -568,6 +805,9 @@ void mul(ZZX& c, const ZZX& a, const ZZX& b)
    long k = min(maxa, maxb);
    long s = min(deg(a), deg(b)) + 1;
 
+   // FIXME: I should have a way of setting all these crossovers
+   // automatically
+
    if (s == 1 || (k == 1 && s < 40) || (k == 2 && s < 20) || 
                  (k == 3 && s < 10)) {
 
@@ -581,11 +821,22 @@ void mul(ZZX& c, const ZZX& a, const ZZX& b)
    }
 
 
-   if (maxa + maxb >= 40 && 
-       SSRatio(deg(a), MaxBits(a), deg(b), MaxBits(b)) < 1.75) 
+   double rat = SSRatio(deg(a), MaxBits(a), deg(b), MaxBits(b));
+   long k1 = (maxa + maxb)/2;
+
+   if ( 
+
+      (k1 >= 26  && rat < 1.40) || 
+      (k1 >= 53  && rat < 1.60) || 
+      (k1 >= 106 && rat < 1.80) || 
+      (k1 >= 212 && rat < 2.00) 
+
+   ) {
       SSMul(c, a, b);
-   else
+   }
+   else {
       HomMul(c, a, b);
+   }
 }
 
 
@@ -612,8 +863,8 @@ void SSSqr(ZZX& c, const ZZX& a)
   LeftShift(p, p, mr);
   add(p, p, 1);
 
-  vec_ZZ aa;
-  aa.SetLength(m2);
+  ZZVec aa;
+  aa.SetSize(m2, p.size());
 
   long i;
   for (i = 0; i <= deg(a); i++) {
@@ -650,12 +901,14 @@ void SSSqr(ZZX& c, const ZZX& a)
   /* Retrieve c, dividing by 2m, and subtracting p where necessary */
   c.rep.SetLength(n + 1);
 
+  ZZ scratch;
+
   for (i = 0; i <= n; i++) {
     ai = aa[i];
     ZZ& ci = c.rep[i];
     if (!IsZero(ai)) {
       /* ci = -ai * 2^{mr-l-1} = ai * 2^{-l-1} = ai / 2m mod p */
-      LeftRotate(ai, ai, mr - l - 1, p, mr);
+      LeftRotate(ai, ai, mr - l - 1, p, mr, scratch);
       sub(tmp, p, ai);
       if (NumBits(tmp) >= mr) { /* ci >= (p-1)/2 */
         negate(ci, ai); /* ci = -ai = ci - p */
@@ -670,7 +923,6 @@ void SSSqr(ZZX& c, const ZZX& a)
 
 void HomSqr(ZZX& x, const ZZX& a)
 {
-
    long da = deg(a);
 
    if (da < 0) {
@@ -678,78 +930,70 @@ void HomSqr(ZZX& x, const ZZX& a)
       return;
    }
 
-   long bound = 2 + NumBits(da+1) + 2*MaxBits(a);
-
-
-   ZZ prod;
-   set(prod);
-
-   long i, nprimes;
-
    zz_pBak bak;
    bak.save();
 
-   for (nprimes = 0; NumBits(prod) <= bound; nprimes++) {
-      if (nprimes >= NumFFTPrimes)
-         zz_p::FFTInit(nprimes);
-      mul(prod, prod, FFTPrime[nprimes]);
+   long bound = 2 + NumBits(da+1) + 2*MaxBits(a);
+
+   FastCRTHelper H(bound, 48);      
+
+   long i, j, k;
+
+   Vec<ZZVec> c, aa;
+
+   c.SetLength(H.nblocks);
+   aa.SetLength(H.nblocks);
+
+   long sz_a = max(1, MaxSize(a));
+
+   for (k = 0; k < H.nblocks; k++) {
+      ZZ *prod_vec = &H.prod_vec[H.start_last_level];
+      c[k].SetSize(da+da+1, prod_vec[k].size()+1);
+      aa[k].SetSize(da+1, min(sz_a, prod_vec[k].size()));
    }
 
+   Vec<ZZ*> ptr_vec;
+   ptr_vec.SetLength(H.nblocks);
+
+   for (j = 0; j <= da; j++) {
+      for (k = 0; k < H.nblocks; k++) ptr_vec[k] = &aa[k][j];
+      H.reduce(a.rep[j], ptr_vec.elts());
+   }
 
-   ZZ coeff;
    ZZ t1;
-   long tt;
+     
+   for (k = 0; k < H.nblocks; k++) {
+      for (i = H.first_vec[k]; i < H.first_vec[k+1]; i++) {
+         zz_p::FFTInit(i);
 
-   vec_ZZ c;
+         zz_pX A, C;
+         conv(A.rep, aa[k]); A.normalize();
+         sqr(C, A);
 
-   c.SetLength(da+da+1);
-
-   long j;
-
-   for (i = 0; i < nprimes; i++) {
-      zz_p::FFTInit(i);
-      long p = zz_p::modulus();
-
-      div(t1, prod, p);
-      tt = rem(t1, p);
-      tt = InvMod(tt, p);
-      mul(coeff, t1, tt);
-
-      zz_pX A, C;
-
-      conv(A, a);
-      sqr(C, A);
-
-      long m = deg(C);
-
-      for (j = 0; j <= m; j++) {
-         /* c[j] += coeff*rep(C.rep[j]) */
-         MulAddTo(c[j], coeff, rep(C.rep[j]));
-         // mul(t1, coeff, rep(C.rep[j]));
-         // add(c[j], c[j], t1); 
+         long m = deg(C);
+         long p = zz_p::modulus();
+         long tt = H.coeff_vec[i];
+         mulmod_precon_t ttpinv = PrepMulModPrecon(tt, p);
+         div(t1, H.prod_vec[H.start_last_level+k], p);
+         for (j = 0; j <= m; j++) {
+            long tt1 = MulModPrecon(rep(C.rep[j]), tt, p, ttpinv);
+            MulAddTo(c[k][j], t1, tt1);
+         }
       }
    }
 
    x.rep.SetLength(da+da+1);
-
-   ZZ prod2;
-   RightShift(prod2, prod, 1);
-
    for (j = 0; j <= da+da; j++) {
-      rem(t1, c[j], prod);
-
-      if (t1 > prod2)
-         sub(x.rep[j], t1, prod);
-      else
-         x.rep[j] = t1;
+      for (k = 0; k < H.nblocks; k++) ptr_vec[k] = &c[k][j];
+      H.reconstruct(x.rep[j], ptr_vec.elts());
    }
 
    x.normalize();
-
    bak.restore();
 }
 
 
+
 void sqr(ZZX& c, const ZZX& a)
 {
    if (IsZero(a)) {
@@ -775,12 +1019,22 @@ void sqr(ZZX& c, const ZZX& a)
    }
 
    long mba = MaxBits(a);
-   
-   if (2*maxa >= 40 && 
-       SSRatio(deg(a), mba, deg(a), mba) < 1.75) 
+   double rat = SSRatio(deg(a), mba, deg(a), mba);
+   long k1 = maxa;
+
+   if ( 
+
+      (k1 >= 26  && rat < 1.40) || 
+      (k1 >= 53  && rat < 1.60) || 
+      (k1 >= 106 && rat < 1.80) || 
+      (k1 >= 212 && rat < 2.00) 
+
+   ) {
       SSSqr(c, a);
-   else
+   }
+   else {
       HomSqr(c, a);
+   }
 }
 
 
@@ -856,7 +1110,7 @@ void diff(ZZX& x, const ZZX& a)
 
 void HomPseudoDivRem(ZZX& q, ZZX& r, const ZZX& a, const ZZX& b)
 {
-   if (IsZero(b)) Error("division by zero");
+   if (IsZero(b)) ArithmeticError("division by zero");
 
    long da = deg(a);
    long db = deg(b);
@@ -985,7 +1239,7 @@ void PlainPseudoDivRem(ZZX& q, ZZX& r, const ZZX& a, const ZZX& b)
    da = deg(a);
    db = deg(b);
 
-   if (db < 0) Error("ZZX: division by zero");
+   if (db < 0) ArithmeticError("ZZX: division by zero");
 
    if (da < db) {
       r = a;
@@ -1064,24 +1318,24 @@ void PlainPseudoRem(ZZX& r, const ZZX& a, const ZZX& b)
 
 void div(ZZX& q, const ZZX& a, long b)
 {
-   if (b == 0) Error("div: division by zero");
+   if (b == 0) ArithmeticError("div: division by zero");
 
-   if (!divide(q, a, b)) Error("DivRem: quotient undefined over ZZ");
+   if (!divide(q, a, b)) ArithmeticError("DivRem: quotient undefined over ZZ");
 }
 
 void div(ZZX& q, const ZZX& a, const ZZ& b)
 {
-   if (b == 0) Error("div: division by zero");
+   if (b == 0) ArithmeticError("div: division by zero");
 
-   if (!divide(q, a, b)) Error("DivRem: quotient undefined over ZZ");
+   if (!divide(q, a, b)) ArithmeticError("DivRem: quotient undefined over ZZ");
 }
 
 static
 void ConstDivRem(ZZX& q, ZZX& r, const ZZX& a, const ZZ& b)
 {
-   if (b == 0) Error("DivRem: division by zero");
+   if (b == 0) ArithmeticError("DivRem: division by zero");
 
-   if (!divide(q, a, b)) Error("DivRem: quotient undefined over ZZ");
+   if (!divide(q, a, b)) ArithmeticError("DivRem: quotient undefined over ZZ");
 
    r = 0;
 }
@@ -1089,7 +1343,7 @@ void ConstDivRem(ZZX& q, ZZX& r, const ZZX& a, const ZZ& b)
 static
 void ConstRem(ZZX& r, const ZZX& a, const ZZ& b)
 {
-   if (b == 0) Error("rem: division by zero");
+   if (b == 0) ArithmeticError("rem: division by zero");
 
    r = 0;
 }
@@ -1101,7 +1355,7 @@ void DivRem(ZZX& q, ZZX& r, const ZZX& a, const ZZX& b)
    long da = deg(a);
    long db = deg(b);
 
-   if (db < 0) Error("DivRem: division by zero");
+   if (db < 0) ArithmeticError("DivRem: division by zero");
 
    if (da < db) {
       r = a;
@@ -1127,8 +1381,8 @@ void DivRem(ZZX& q, ZZX& r, const ZZX& a, const ZZX& b)
       ZZ m;
       PseudoDivRem(q1, r1, a, b);
       power(m, LeadCoeff(b), da-db+1);
-      if (!divide(q, q1, m)) Error("DivRem: quotient not defined over ZZ");
-      if (!divide(r, r1, m)) Error("DivRem: remainder not defined over ZZ");
+      if (!divide(q, q1, m)) ArithmeticError("DivRem: quotient not defined over ZZ");
+      if (!divide(r, r1, m)) ArithmeticError("DivRem: remainder not defined over ZZ");
    }
 }
 
@@ -1137,7 +1391,7 @@ void div(ZZX& q, const ZZX& a, const ZZX& b)
    long da = deg(a);
    long db = deg(b);
 
-   if (db < 0) Error("div: division by zero");
+   if (db < 0) ArithmeticError("div: division by zero");
 
    if (da < db) {
       q = 0;
@@ -1164,7 +1418,7 @@ void div(ZZX& q, const ZZX& a, const ZZX& b)
       ZZ m;
       PseudoDiv(q1, a, b);
       power(m, LeadCoeff(b), da-db+1);
-      if (!divide(q, q1, m)) Error("div: quotient not defined over ZZ");
+      if (!divide(q, q1, m)) ArithmeticError("div: quotient not defined over ZZ");
    }
 }
 
@@ -1173,7 +1427,7 @@ void rem(ZZX& r, const ZZX& a, const ZZX& b)
    long da = deg(a);
    long db = deg(b);
 
-   if (db < 0) Error("rem: division by zero");
+   if (db < 0) ArithmeticError("rem: division by zero");
 
    if (da < db) {
       r = a;
@@ -1197,7 +1451,7 @@ void rem(ZZX& r, const ZZX& a, const ZZX& b)
       ZZ m;
       PseudoRem(r1, a, b);
       power(m, LeadCoeff(b), da-db+1);
-      if (!divide(r, r1, m)) Error("rem: remainder not defined over ZZ");
+      if (!divide(r, r1, m)) ArithmeticError("rem: remainder not defined over ZZ");
    }
 }
 
@@ -1703,7 +1957,7 @@ void trunc(ZZX& x, const ZZX& a, long m)
 // x = a % X^m, output may alias input
 
 {
-   if (m < 0) Error("trunc: bad args");
+   if (m < 0) LogicError("trunc: bad args");
 
    if (&x == &a) {
       if (x.rep.length() > m) {
@@ -1747,7 +2001,7 @@ void LeftShift(ZZX& x, const ZZX& a, long n)
    }
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("overflow in LeftShift");
+      ResourceError("overflow in LeftShift");
 
    long m = a.rep.length();
 
@@ -1770,7 +2024,7 @@ void RightShift(ZZX& x, const ZZX& a, long n)
    }
 
    if (n < 0) {
-      if (n < -NTL_MAX_LONG) Error("overflow in RightShift");
+      if (n < -NTL_MAX_LONG) ResourceError("overflow in RightShift");
       LeftShift(x, a, -n);
       return;
    }
@@ -1799,7 +2053,7 @@ void RightShift(ZZX& x, const ZZX& a, long n)
 void TraceVec(vec_ZZ& S, const ZZX& ff)
 {
    if (!IsOne(LeadCoeff(ff)))
-      Error("TraceVec: bad args");
+      LogicError("TraceVec: bad args");
 
    ZZX f;
    f = ff;
@@ -1957,7 +2211,7 @@ void MinPolyMod(ZZX& gg, const ZZX& a, const ZZX& f)
 
 {
    if (!IsOne(LeadCoeff(f)) || deg(f) < 1 || deg(a) >= deg(f))
-      Error("MinPolyMod: bad args");
+      LogicError("MinPolyMod: bad args");
 
    if (IsZero(a)) {
       SetX(gg);
@@ -2144,7 +2398,7 @@ void XGCD(ZZ& rr, ZZX& ss, ZZX& tt, const ZZX& a, const ZZX& b,
 void NormMod(ZZ& x, const ZZX& a, const ZZX& f, long deterministic)
 {
    if (!IsOne(LeadCoeff(f)) || deg(a) >= deg(f) || deg(f) <= 0)
-      Error("norm: bad args");
+      LogicError("norm: bad args");
 
    if (IsZero(a)) {
       clear(x);
@@ -2157,7 +2411,7 @@ void NormMod(ZZ& x, const ZZX& a, const ZZX& f, long deterministic)
 void TraceMod(ZZ& res, const ZZX& a, const ZZX& f)
 {
    if (!IsOne(LeadCoeff(f)) || deg(a) >= deg(f) || deg(f) <= 0)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
    vec_ZZ S;
 
@@ -2182,7 +2436,7 @@ void discriminant(ZZ& d, const ZZX& a, long deterministic)
    diff(a1, a);
    resultant(res, a, a1, deterministic);
    if (!divide(res, res, LeadCoeff(a)))
-      Error("discriminant: inexact division");
+      LogicError("discriminant: inexact division");
 
    m = m & 3;
    if (m >= 2)
@@ -2196,7 +2450,7 @@ void MulMod(ZZX& x, const ZZX& a, const ZZX& b, const ZZX& f)
 {
    if (deg(a) >= deg(f) || deg(b) >= deg(f) || deg(f) == 0 || 
        !IsOne(LeadCoeff(f)))
-      Error("MulMod: bad args");
+      LogicError("MulMod: bad args");
 
    ZZX t;
    mul(t, a, b);
@@ -2206,7 +2460,7 @@ void MulMod(ZZX& x, const ZZX& a, const ZZX& b, const ZZX& f)
 void SqrMod(ZZX& x, const ZZX& a, const ZZX& f)
 {
    if (deg(a) >= deg(f) || deg(f) == 0 || !IsOne(LeadCoeff(f)))
-      Error("MulMod: bad args");
+      LogicError("MulMod: bad args");
 
    ZZX t;
    sqr(t, a);
@@ -2229,7 +2483,7 @@ void MulByXModAux(ZZX& h, const ZZX& a, const ZZX& f)
    m = deg(a);
 
    if (m >= n || n == 0 || !IsOne(LeadCoeff(f)))
-      Error("MulByXMod: bad args");
+      LogicError("MulByXMod: bad args");
 
    if (m < 0) {
       clear(h);
@@ -2307,7 +2561,7 @@ long CharPolyBound(const ZZX& a, const ZZX& f)
 
 {
    if (IsZero(a) || IsZero(f))
-      Error("CharPolyBound: bad args");
+      LogicError("CharPolyBound: bad args");
 
    ZZ t1, t2, t;
    EuclLength1(t1, a);
@@ -2362,7 +2616,7 @@ void reverse(ZZX& x, const ZZX& a, long hi)
 {
    if (hi < 0) { clear(x); return; }
    if (NTL_OVERFLOW(hi, 1, 0))
-      Error("overflow in reverse");
+      ResourceError("overflow in reverse");
 
    if (&x == &a) {
       ZZX tmp;
@@ -2397,7 +2651,7 @@ void NewtonInvTrunc(ZZX& c, const ZZX& a, long e)
    else if (ConstTerm(a) == -1)
       x = -1;
    else
-      Error("InvTrunc: non-invertible constant term");
+      ArithmeticError("InvTrunc: non-invertible constant term");
 
    if (e == 1) {
       conv(c, x);
@@ -2451,7 +2705,7 @@ void NewtonInvTrunc(ZZX& c, const ZZX& a, long e)
 
 void InvTrunc(ZZX& c, const ZZX& a, long e)
 {
-   if (e < 0) Error("InvTrunc: bad args");
+   if (e < 0) LogicError("InvTrunc: bad args");
 
    if (e == 0) {
       clear(c);
@@ -2459,7 +2713,7 @@ void InvTrunc(ZZX& c, const ZZX& a, long e)
    }
 
    if (NTL_OVERFLOW(e, 1, 0))
-      Error("overflow in InvTrunc");
+      ResourceError("overflow in InvTrunc");
 
    NewtonInvTrunc(c, a, e);
 }
