@@ -2,11 +2,9 @@
 #include <NTL/ZZ_pEXFactoring.h>
 #include <NTL/FacVec.h>
 #include <NTL/fileio.h>
-
-#include <stdio.h>
-
 #include <NTL/new.h>
 
+
 NTL_START_IMPL
 
 
@@ -33,7 +31,7 @@ void SquareFreeDecomp(vec_pair_ZZ_pEX_long& u, const ZZ_pEX& ff)
    ZZ_pEX f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("SquareFreeDecomp: bad args");
+      LogicError("SquareFreeDecomp: bad args");
 
    ZZ_pEX r, t, v, tmp1;
    long m, j, finished, done;
@@ -155,7 +153,7 @@ void FindRoots(vec_ZZ_pE& x, const ZZ_pEX& ff)
    ZZ_pEX f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("FindRoots: bad args");
+      LogicError("FindRoots: bad args");
 
    x.SetMaxLength(deg(f));
    x.SetLength(0);
@@ -245,7 +243,7 @@ void TraceMap(ZZ_pEX& w, const ZZ_pEX& a, long d, const ZZ_pEXModulus& F,
               const ZZ_pEX& b)
 
 {
-   if (d < 0) Error("TraceMap: bad args");
+   if (d < 0) LogicError("TraceMap: bad args");
 
    ZZ_pEX y, z, t;
 
@@ -284,7 +282,7 @@ void TraceMap(ZZ_pEX& w, const ZZ_pEX& a, long d, const ZZ_pEXModulus& F,
 
 void PowerCompose(ZZ_pEX& y, const ZZ_pEX& h, long q, const ZZ_pEXModulus& F)
 {
-   if (q < 0) Error("PowerCompose: bad args");
+   if (q < 0) LogicError("PowerCompose: bad args");
 
    ZZ_pEX z(INIT_SIZE, F.n);
    long sw;
@@ -360,7 +358,7 @@ long ProbIrredTest(const ZZ_pEX& f, long iter)
 }
 
 
-NTL_THREAD_LOCAL long ZZ_pEX_BlockingFactor = 10;
+NTL_CHEAP_THREAD_LOCAL long ZZ_pEX_BlockingFactor = 10;
 
 
 
@@ -429,7 +427,7 @@ void EDF(vec_ZZ_pEX& factors, const ZZ_pEX& ff, const ZZ_pEX& bb,
    ZZ_pEX b = bb;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("EDF: bad args");
+      LogicError("EDF: bad args");
 
    long n = deg(f);
    long r = n/d;
@@ -470,7 +468,7 @@ void SFCanZass(vec_ZZ_pEX& factors, const ZZ_pEX& ff, long verbose)
    ZZ_pEX f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("SFCanZass: bad args");
+      LogicError("SFCanZass: bad args");
 
    if (deg(f) == 0) {
       factors.SetLength(0);
@@ -540,7 +538,7 @@ void SFCanZass(vec_ZZ_pEX& factors, const ZZ_pEX& ff, long verbose)
 void CanZass(vec_pair_ZZ_pEX_long& factors, const ZZ_pEX& f, long verbose)
 {
    if (!IsOne(LeadCoeff(f)))
-      Error("CanZass: bad args");
+      LogicError("CanZass: bad args");
 
    double t;
    vec_pair_ZZ_pEX_long sfd;
@@ -727,10 +725,10 @@ void FindRoot(ZZ_pE& root, const ZZ_pEX& ff)
    f = ff;
    
    if (!IsOne(LeadCoeff(f)))
-      Error("FindRoot: bad args");
+      LogicError("FindRoot: bad args");
 
    if (deg(f) == 0)
-      Error("FindRoot: bad args");
+      LogicError("FindRoot: bad args");
 
 
    while (deg(f) > 1) {
@@ -1004,9 +1002,9 @@ void RecBuildIrred(ZZ_pEX& f, long u, const FacVec& fvec)
 void BuildIrred(ZZ_pEX& f, long n)
 {
    if (n <= 0)
-      Error("BuildIrred: n must be positive");
+      LogicError("BuildIrred: n must be positive");
 
-   if (NTL_OVERFLOW(n, 1, 0)) Error("overflow in BuildIrred");
+   if (NTL_OVERFLOW(n, 1, 0)) ResourceError("overflow in BuildIrred");
 
    if (n == 1) {
       SetX(f);
@@ -1026,7 +1024,7 @@ void BuildIrred(ZZ_pEX& f, long n)
 void BuildIrred(ZZ_pEX& f, long n)
 {
    if (n <= 0)
-      Error("BuildIrred: n must be positive");
+      LogicError("BuildIrred: n must be positive");
 
    if (n == 1) {
       SetX(f);
@@ -1064,18 +1062,11 @@ void BuildRandomIrred(ZZ_pEX& f, const ZZ_pEX& g)
 
 /************* NEW DDF ****************/
 
-NTL_THREAD_LOCAL long ZZ_pEX_GCDTableSize = 4;
-NTL_THREAD_LOCAL char ZZ_pEX_stem[256] = "";
-
-NTL_THREAD_LOCAL double ZZ_pEXFileThresh = NTL_FILE_THRESH;
-
-NTL_THREAD_LOCAL static vec_ZZ_pEX BabyStepFile;
-NTL_THREAD_LOCAL static vec_ZZ_pEX GiantStepFile;
-
-NTL_THREAD_LOCAL static long use_files;
-
-// FIXME: in a thread-safe impl, we should be careful
-// with external file names.
+NTL_CHEAP_THREAD_LOCAL long ZZ_pEX_GCDTableSize = 4;
+NTL_CHEAP_THREAD_LOCAL double ZZ_pEXFileThresh = NTL_FILE_THRESH;
+static NTL_CHEAP_THREAD_LOCAL vec_ZZ_pEX *BabyStepFile=0;
+static NTL_CHEAP_THREAD_LOCAL vec_ZZ_pEX *GiantStepFile=0;
+static NTL_CHEAP_THREAD_LOCAL long use_files;
 
 
 static
@@ -1094,7 +1085,7 @@ double CalcTableSize(long n, long k)
 
 static
 void GenerateBabySteps(ZZ_pEX& h1, const ZZ_pEX& f, const ZZ_pEX& h, long k,
-                       long verbose)
+                       FileList& flist, long verbose)
 
 {
    double t;
@@ -1123,19 +1114,18 @@ void GenerateBabySteps(ZZ_pEX& h1, const ZZ_pEX& f, const ZZ_pEX& h, long k,
    long i;
 
    if (!use_files) {
-      BabyStepFile.kill();
-      BabyStepFile.SetLength(k-1);
+      (*BabyStepFile).SetLength(k-1);
    }
 
    for (i = 1; i <= k-1; i++) {
       if (use_files) {
          ofstream s;
-         OpenWrite(s, FileName(ZZ_pEX_stem, "baby", i));
+         OpenWrite(s, FileName("baby", i), flist);
          s << h1 << "\n";
-         s.close();
+         CloseWrite(s);
       }
       else
-         BabyStepFile(i) = h1;
+         (*BabyStepFile)(i) = h1;
 
       CompMod(h1, h1, H, F);
       if (verbose) cerr << "+";
@@ -1148,7 +1138,8 @@ void GenerateBabySteps(ZZ_pEX& h1, const ZZ_pEX& f, const ZZ_pEX& h, long k,
 
 
 static
-void GenerateGiantSteps(const ZZ_pEX& f, const ZZ_pEX& h, long l, long verbose)
+void GenerateGiantSteps(const ZZ_pEX& f, const ZZ_pEX& h, long l, 
+                        FileList& flist, long verbose)
 {
 
    double t;
@@ -1178,19 +1169,18 @@ void GenerateGiantSteps(const ZZ_pEX& f, const ZZ_pEX& h, long l, long verbose)
    long i;
 
    if (!use_files) {
-      GiantStepFile.kill();
-      GiantStepFile.SetLength(l);
+      (*GiantStepFile).SetLength(l);
    }
 
    for (i = 1; i <= l-1; i++) {
       if (use_files) {
          ofstream s;
-         OpenWrite(s, FileName(ZZ_pEX_stem, "giant", i));
+         OpenWrite(s, FileName("giant", i), flist);
          s << h1 << "\n";
-         s.close();
+         CloseWrite(s);
       }
       else
-        GiantStepFile(i) = h1;
+        (*GiantStepFile)(i) = h1;
 
       CompMod(h1, h1, H, F);
       if (verbose) cerr << "+";
@@ -1198,36 +1188,18 @@ void GenerateGiantSteps(const ZZ_pEX& f, const ZZ_pEX& h, long l, long verbose)
 
    if (use_files) {
       ofstream s;
-      OpenWrite(s, FileName(ZZ_pEX_stem, "giant", i));
+      OpenWrite(s, FileName("giant", i), flist);
       s << h1 << "\n";
-      s.close();
+      CloseWrite(s);
    }
    else
-      GiantStepFile(i) = h1;
+      (*GiantStepFile)(i) = h1;
 
    if (verbose)
       cerr << (GetTime()-t) << "\n";
 
 }
 
-static
-void FileCleanup(long k, long l)
-{
-   if (use_files) {
-      long i;
-   
-      for (i = 1; i <= k-1; i++)
-         remove(FileName(ZZ_pEX_stem, "baby", i));
-   
-      for (i = 1; i <= l; i++)
-         remove(FileName(ZZ_pEX_stem, "giant", i));
-   }
-   else {
-      BabyStepFile.kill();
-      GiantStepFile.kill();
-   }
-}
-
 
 static
 void NewAddFactor(vec_pair_ZZ_pEX_long& u, const ZZ_pEX& g, long m, long verbose)
@@ -1297,14 +1269,11 @@ void FetchGiantStep(ZZ_pEX& g, long gs, const ZZ_pEXModulus& F)
 {
    if (use_files) {
       ifstream s;
-   
-      OpenRead(s, FileName(ZZ_pEX_stem, "giant", gs));
-   
-      s >> g;
-      s.close();
+      OpenRead(s, FileName("giant", gs));
+      NTL_INPUT_CHECK_ERR(s >> g);
    }
    else
-      g = GiantStepFile(gs);
+      g = (*GiantStepFile)(gs);
 
 
    rem(g, g, F);
@@ -1322,12 +1291,11 @@ void FetchBabySteps(vec_ZZ_pEX& v, long k)
    for (i = 1; i <= k-1; i++) {
       if (use_files) {
          ifstream s;
-         OpenRead(s, FileName(ZZ_pEX_stem, "baby", i));
-         s >> v[i];
-         s.close();
+         OpenRead(s, FileName("baby", i));
+         NTL_INPUT_CHECK_ERR(s >> v[i]);
       }
       else
-         v[i] = BabyStepFile(i);
+         v[i] = (*BabyStepFile)(i);
    }
 }
       
@@ -1524,7 +1492,7 @@ void NewDDF(vec_pair_ZZ_pEX_long& factors,
 
 {
    if (!IsOne(LeadCoeff(f)))
-      Error("NewDDF: bad args");
+      LogicError("NewDDF: bad args");
 
    if (deg(f) == 0) {
       factors.SetLength(0);
@@ -1537,9 +1505,6 @@ void NewDDF(vec_pair_ZZ_pEX_long& factors,
       return;
    }
 
-   if (!ZZ_pEX_stem[0])
-      sprintf(ZZ_pEX_stem, "ddf-%ld", RandomBnd(10000));
-      
    long B = deg(f)/2;
    long k = SqrRoot(B);
    long l = (B+k-1)/k;
@@ -1552,15 +1517,22 @@ void NewDDF(vec_pair_ZZ_pEX_long& factors,
       use_files = 0;
 
 
-   GenerateBabySteps(h1, f, h, k, verbose);
+   FileList flist;
 
-   GenerateGiantSteps(f, h1, l, verbose);
+   vec_ZZ_pEX local_BabyStepFile;
+   vec_ZZ_pEX local_GiantStepFile;
+
+   BabyStepFile = &local_BabyStepFile;
+   GiantStepFile = &local_GiantStepFile;
+
+
+   GenerateBabySteps(h1, f, h, k, flist, verbose);
+
+   GenerateGiantSteps(f, h1, l, flist, verbose);
 
    vec_pair_ZZ_pEX_long u;
    GiantRefine(u, f, k, l, verbose);
    BabyRefine(factors, u, k, l, verbose);
-
-   FileCleanup(k, l);
 }
 
 long IterComputeDegree(const ZZ_pEX& h, const ZZ_pEXModulus& F)
