@@ -1,10 +1,12 @@
 
-#include <stdlib.h>
-#include <math.h>
-#include <stdio.h>
+#include <cstdlib>
+#include <cmath>
+#include <cstdio>
 
 #include <NTL/config.h>
 
+using namespace std;
+
 
 #ifndef NTL_GMP_LIP
 
@@ -59,6 +61,8 @@ void print2k(FILE *f, long k, long bpl)
    fprintf(f, ")");
 }
 
+
+
 void Error(const char *s)
 {
    fprintf(stderr, "%s\n", s);
@@ -69,7 +73,7 @@ void Error(const char *s)
 int main()
 {
    long bpl;
-   long ntl_zz_nbits, ntl_wsp_nbits, ntl_sp_nbits;
+   long ntl_zz_nbits;
 
    fprintf(stderr, "NTL_GMP_LIP flag set\n");
 
@@ -97,9 +101,6 @@ int main()
 
    Error("sorry...this is a funny gmp");
 
-   ntl_wsp_nbits = bpl - 2;
-
-   ntl_sp_nbits = NTL_NBITS_MAX;
 
    if (sizeof(mp_size_t) < sizeof(long)) {
       printf("#define NTL_SMALL_MP_SIZE_T\n");
@@ -107,8 +108,6 @@ int main()
    }
 
    fprintf(stderr, "NTL_ZZ_NBITS = %ld\n", ntl_zz_nbits);
-   fprintf(stderr, "NTL_WSP_NBITS = %ld\n", ntl_wsp_nbits);
-   fprintf(stderr, "NTL_SP_NBITS = %ld\n", ntl_sp_nbits);
 
    printf("#define NTL_ZZ_NBITS (%ld)\n",  ntl_zz_nbits);
 
