@@ -1,9 +1,12 @@
 
-
 #include <NTL/lzz_pX.h>
-
 #include <NTL/new.h>
 
+#ifdef NTL_HAVE_AVX
+#include <immintrin.h>
+#endif
+
+
 NTL_START_IMPL
 
 
@@ -54,7 +57,7 @@ void RightShift(zz_pX& x, const zz_pX& a, long n)
    }
 
    if (n < 0) {
-      if (n < -NTL_MAX_LONG) Error("overflow in RightShift");
+      if (n < -NTL_MAX_LONG) ResourceError("overflow in RightShift");
       LeftShift(x, a, -n);
       return;
    }
@@ -95,7 +98,7 @@ void LeftShift(zz_pX& x, const zz_pX& a, long n)
    }
 
    if (NTL_OVERFLOW(n, 1, 0))
-      Error("overflow in LeftShift");
+      ResourceError("overflow in LeftShift");
 
    long m = a.rep.length();
 
@@ -824,7 +827,7 @@ void eval(vec_zz_p& b, const zz_pX& f, const vec_zz_p& a)
 void interpolate(zz_pX& f, const vec_zz_p& a, const vec_zz_p& b)
 {
    long m = a.length();
-   if (b.length() != m) Error("interpolate: vector length mismatch");
+   if (b.length() != m) LogicError("interpolate: vector length mismatch");
 
    if (m == 0) {
       clear(f);
@@ -904,7 +907,7 @@ void InnerProduct(zz_pX& x, const vec_zz_p& v, long low, long high,
 
 
    long p = zz_p::modulus();
-   double pinv = zz_p::ModulusInverse();
+   mulmod_t pinv = zz_p::ModulusInverse();
 
    high = min(high, v.length()-1);
    for (i = low; i <= high; i++) {
@@ -913,7 +916,7 @@ void InnerProduct(zz_pX& x, const vec_zz_p& v, long low, long high,
       zz_p w = (v[i]);
 
       long W = rep(w);
-      mulmod_precon_t Wpinv = PrepMulModPrecon(W, p, pinv); // ((double) W)*pinv;
+      mulmod_precon_t Wpinv = PrepMulModPrecon(W, p, pinv); 
       const zz_p *hp = h.elts();
 
       for (j = 0; j < m; j++) {
@@ -959,7 +962,7 @@ void CompMod(zz_pX& x, const zz_pX& g, const zz_pXArgument& A,
 
 void build(zz_pXArgument& A, const zz_pX& h, const zz_pXModulus& F, long m)
 {
-   if (m <= 0 || deg(h) >= F.n) Error("build: bad args");
+   if (m <= 0 || deg(h) >= F.n) LogicError("build: bad args");
 
    if (m > F.n) m = F.n;
 
@@ -990,7 +993,7 @@ void build(zz_pXArgument& A, const zz_pX& h, const zz_pXModulus& F, long m)
 
 
 
-NTL_THREAD_LOCAL long zz_pXArgBound = 0;
+NTL_CHEAP_THREAD_LOCAL long zz_pXArgBound = 0;
 
 
 void CompMod(zz_pX& x, const zz_pX& g, const zz_pX& h, const zz_pXModulus& F)
@@ -1067,6 +1070,521 @@ void Comp3Mod(zz_pX& x1, zz_pX& x2, zz_pX& x3,
    x3 = xx3;
 }
 
+
+// BEGIN zz_pXAltArgument variation 
+
+
+
+
+void build(zz_pXAltArgument& altH, const zz_pXArgument& H, const zz_pXModulus& F)
+{
+   altH.orig = &H;
+
+
+#ifdef NTL_HAVE_LL_TYPE
+   altH.mem.kill();
+   altH.row.kill();
+
+#ifdef NTL_HAVE_AVX
+   altH.dmem.kill();
+   altH.drow.kill();
+#endif
+
+   if (H.H.length() < 10 || F.n < 50) { altH.strategy = 0; return; }
+
+   altH.n = F.n;
+   altH.m = H.H.length()-1;
+
+   long p = zz_p::modulus();
+   long n = altH.n;
+   long m = altH.m;
+
+
+#ifdef NTL_HAVE_AVX
+   if (n >= 128 && m <= ((1L << NTL_DOUBLE_PRECISION)-1)/(p-1) &&
+      m*(p-1) <= ((1L << NTL_DOUBLE_PRECISION)-1)/(p-1)) {
+         altH.strategy = 3;
+         altH.pinv_L = sp_PrepRem(p);
+   }
+   else
+#endif
+   if (cast_unsigned(m) <= (~(0UL))/cast_unsigned(p-1) &&
+       cast_unsigned(m)*cast_unsigned(p-1) <= (~(0UL))/cast_unsigned(p-1)) {
+      altH.strategy = 1;
+      altH.pinv_L = sp_PrepRem(p);
+   }
+   else {
+      altH.strategy = 2;
+      altH.pinv_LL = make_sp_ll_reduce_struct(p);
+   }
+
+
+   if (altH.strategy == 1 || altH.strategy == 2) {
+
+      altH.row.SetLength(n);
+      long **row = altH.row.elts();
+      
+      const long  AllocAmt = 1L << 18;
+   
+      long BlockSize = (AllocAmt + m - 1)/m;
+      long NumBlocks = (n + BlockSize - 1)/BlockSize;
+   
+      altH.mem.SetLength(NumBlocks);
+   
+      for (long i = 0; i < NumBlocks; i++) {
+         long first = i*BlockSize;
+         long last =  min(n, first + BlockSize);
+         altH.mem[i].SetLength((last-first)*m);
+         for (long j = first; j < last; j++) {
+            row[j] = altH.mem[i].elts() + (j-first)*m;
+         }
+      }
+   
+      for (long i = 0; i < m; i++) {
+         const zz_p* ptr = H.H[i].rep.elts();
+         long len = H.H[i].rep.length();
+         for (long j = 0; j < len; j++) 
+            row[j][i] = rep(ptr[j]);
+         for (long j = len; j < n; j++)
+            row[j][i] = 0;
+      }
+   }
+#ifdef NTL_HAVE_AVX
+   else {
+
+      // sanity check
+      if (m >= (1L << (NTL_BITS_PER_LONG-8))) ResourceError("zz_pXAltArgument: overflow");
+
+      long npanels = (n+15)/16;
+      long panel_size = 16*m;
+
+      const long AllocAmt = 1L << 18;
+
+      long BlockSize = (AllocAmt + panel_size - 1)/panel_size;
+      long NumBlocks = (npanels + BlockSize - 1)/BlockSize;
+
+      altH.dmem.SetLength(NumBlocks);
+      altH.drow.SetLength(npanels);
+      double **drow = altH.drow.elts();
+
+      for (long i = 0; i < NumBlocks; i++) {
+         long first = i*BlockSize;
+         long last = min(npanels, first + BlockSize);
+         altH.dmem[i].SetLength((last-first)*panel_size);
+
+         double *ptr = altH.dmem[i].get();
+
+         for (long j = first; j < last; j++)
+            drow[j] = ptr + (j-first)*panel_size;
+      }
+
+      for (long i = 0; i < m; i++) {
+         const zz_p *ptr = H.H[i].rep.elts();
+         long len = H.H[i].rep.length();
+         for (long j = 0; j < len; j++)
+            drow[j/16][(i*16) + (j%16)] = rep(ptr[j]);
+         for (long j = len; j < npanels*16; j++) 
+            drow[j/16][(i*16) + (j%16)] = 0;
+      }
+   }
+
+#endif
+
+
+#endif
+}
+
+
+#ifdef NTL_HAVE_LL_TYPE
+
+
+#ifdef NTL_HAVE_AVX
+static
+void mul16rowsD(double *x, const double *a, const double *b, long n)
+{
+   __m256d avec0, avec1, avec2, avec3;
+
+   __m256d acc0 = _mm256_setzero_pd();
+   __m256d acc1 = _mm256_setzero_pd();
+   __m256d acc2 = _mm256_setzero_pd();
+   __m256d acc3 = _mm256_setzero_pd();
+
+   __m256d bvec;
+
+   for (long i = 0; i < n; i++) {
+      bvec = _mm256_broadcast_sd(&b[i]); 
+
+      avec0 = _mm256_load_pd(a); a += 4;
+      avec1 = _mm256_load_pd(a); a += 4;
+      avec2 = _mm256_load_pd(a); a += 4;
+      avec3 = _mm256_load_pd(a); a += 4;
+
+#ifdef NTL_HAVE_FMA
+
+      acc0 = _mm256_fmadd_pd(avec0, bvec, acc0);
+      acc1 = _mm256_fmadd_pd(avec1, bvec, acc1);
+      acc2 = _mm256_fmadd_pd(avec2, bvec, acc2);
+      acc3 = _mm256_fmadd_pd(avec3, bvec, acc3);
+
+#else
+
+      acc0 = _mm256_add_pd(_mm256_mul_pd(avec0, bvec), acc0);
+      acc1 = _mm256_add_pd(_mm256_mul_pd(avec1, bvec), acc1);
+      acc2 = _mm256_add_pd(_mm256_mul_pd(avec2, bvec), acc2);
+      acc3 = _mm256_add_pd(_mm256_mul_pd(avec3, bvec), acc3);
+
+#endif
+
+   }
+
+   _mm256_store_pd(x + 0*4, acc0);
+   _mm256_store_pd(x + 1*4, acc1);
+   _mm256_store_pd(x + 2*4, acc2);
+   _mm256_store_pd(x + 3*4, acc3);
+}
+
+static
+void mul16rows2D(double *x, double *x_, const double *a, const double *b, const double *b_, long n)
+{
+   __m256d avec0, avec1, avec2, avec3;
+
+   __m256d acc0 = _mm256_setzero_pd();
+   __m256d acc1 = _mm256_setzero_pd();
+   __m256d acc2 = _mm256_setzero_pd();
+   __m256d acc3 = _mm256_setzero_pd();
+
+   __m256d acc0_ = _mm256_setzero_pd();
+   __m256d acc1_ = _mm256_setzero_pd();
+   __m256d acc2_ = _mm256_setzero_pd();
+   __m256d acc3_ = _mm256_setzero_pd();
+
+
+   __m256d bvec;
+   __m256d bvec_;
+
+   for (long i = 0; i < n; i++) {
+      bvec = _mm256_broadcast_sd(&b[i]); 
+      bvec_ = _mm256_broadcast_sd(&b_[i]); 
+
+      avec0 = _mm256_load_pd(a); a += 4;
+      avec1 = _mm256_load_pd(a); a += 4;
+      avec2 = _mm256_load_pd(a); a += 4;
+      avec3 = _mm256_load_pd(a); a += 4;
+
+#ifdef NTL_HAVE_FMA
+
+      acc0 = _mm256_fmadd_pd(avec0, bvec, acc0);
+      acc1 = _mm256_fmadd_pd(avec1, bvec, acc1);
+      acc2 = _mm256_fmadd_pd(avec2, bvec, acc2);
+      acc3 = _mm256_fmadd_pd(avec3, bvec, acc3);
+
+      acc0_ = _mm256_fmadd_pd(avec0, bvec_, acc0_);
+      acc1_ = _mm256_fmadd_pd(avec1, bvec_, acc1_);
+      acc2_ = _mm256_fmadd_pd(avec2, bvec_, acc2_);
+      acc3_ = _mm256_fmadd_pd(avec3, bvec_, acc3_);
+
+#else
+      acc0 = _mm256_add_pd(_mm256_mul_pd(avec0, bvec), acc0);
+      acc1 = _mm256_add_pd(_mm256_mul_pd(avec1, bvec), acc1);
+      acc2 = _mm256_add_pd(_mm256_mul_pd(avec2, bvec), acc2);
+      acc3 = _mm256_add_pd(_mm256_mul_pd(avec3, bvec), acc3);
+
+      acc0_ = _mm256_add_pd(_mm256_mul_pd(avec0, bvec_), acc0_);
+      acc1_ = _mm256_add_pd(_mm256_mul_pd(avec1, bvec_), acc1_);
+      acc2_ = _mm256_add_pd(_mm256_mul_pd(avec2, bvec_), acc2_);
+      acc3_ = _mm256_add_pd(_mm256_mul_pd(avec3, bvec_), acc3_);
+
+#endif
+
+   }
+
+   _mm256_store_pd(x + 0*4, acc0);
+   _mm256_store_pd(x + 1*4, acc1);
+   _mm256_store_pd(x + 2*4, acc2);
+   _mm256_store_pd(x + 3*4, acc3);
+
+   _mm256_store_pd(x_ + 0*4, acc0_);
+   _mm256_store_pd(x_ + 1*4, acc1_);
+   _mm256_store_pd(x_ + 2*4, acc2_);
+   _mm256_store_pd(x_ + 3*4, acc3_);
+}
+
+
+#endif
+
+
+
+static
+void InnerProduct_LL(zz_pX& x, const vec_zz_p& v, long low, long high, 
+                   const zz_pXAltArgument& H, long n)
+{
+   high = min(high, v.length()-1);
+   long len = high-low+1;
+   if (len <= 0) {
+      clear(x);
+      return;
+   }
+
+   x.rep.SetLength(n);
+   zz_p *xp = x.rep.elts();
+
+   long p = zz_p::modulus();
+   sp_ll_reduce_struct pinv = H.pinv_LL;
+
+   const zz_p *vp = v.elts() + low;
+
+   for (long i = 0; i < n; i++) 
+      xp[i].LoopHole() = InnerProd_LL(H.row[i], vp, len, p, pinv);
+
+   x.normalize();
+}
+
+static
+void CompMod_LL(zz_pX& x, const zz_pX& g, const zz_pXAltArgument& A, 
+             const zz_pXModulus& F)
+{
+   if (deg(g) <= 0) {
+      x = g;
+      return;
+   }
+
+
+   zz_pX s, t;
+
+   long m = A.m;
+   long l = ((g.rep.length()+m-1)/m) - 1;
+
+   zz_pXMultiplier M;
+   build(M, A.orig->H[m], F);
+
+   InnerProduct_LL(t, g.rep, l*m, l*m + m - 1, A, F.n);
+   for (long i = l-1; i >= 0; i--) {
+      InnerProduct_LL(s, g.rep, i*m, i*m + m - 1, A, F.n);
+      MulMod(t, t, M, F);
+      add(t, t, s);
+   }
+
+   x = t;
+}
+
+static
+void InnerProduct_L(zz_pX& x, const vec_zz_p& v, long low, long high, 
+                   const zz_pXAltArgument& H, long n)
+{
+   high = min(high, v.length()-1);
+   long len = high-low+1;
+   if (len <= 0) {
+      clear(x);
+      return;
+   }
+
+   x.rep.SetLength(n);
+   zz_p *xp = x.rep.elts();
+
+   long p = zz_p::modulus();
+   sp_reduce_struct pinv = H.pinv_L;
+
+
+   const zz_p *vp = v.elts() + low;
+
+   for (long i = 0; i < n; i++) 
+      xp[i].LoopHole() = InnerProd_L(H.row[i], vp, len, p, pinv);
+
+   x.normalize();
+}
+
+static
+void CompMod_L(zz_pX& x, const zz_pX& g, const zz_pXAltArgument& A, 
+             const zz_pXModulus& F)
+{
+   if (deg(g) <= 0) {
+      x = g;
+      return;
+   }
+
+
+   zz_pX s, t;
+
+   long m = A.m;
+   long l = ((g.rep.length()+m-1)/m) - 1;
+
+   zz_pXMultiplier M;
+   build(M, A.orig->H[m], F);
+
+   InnerProduct_L(t, g.rep, l*m, l*m + m - 1, A, F.n);
+   for (long i = l-1; i >= 0; i--) {
+      InnerProduct_L(s, g.rep, i*m, i*m + m - 1, A, F.n);
+      MulMod(t, t, M, F);
+      add(t, t, s);
+   }
+
+   x = t;
+}
+
+
+#ifdef NTL_HAVE_AVX
+
+static
+void InnerProduct_AVX(zz_pX& x, const Vec<double>& v, long low, long high, 
+                   const zz_pXAltArgument& H, long n)
+{
+   high = min(high, v.length()-1);
+   long len = high-low+1;
+   if (len <= 0) {
+      clear(x);
+      return;
+   }
+
+   x.rep.SetLength(n);
+   zz_p *xp = x.rep.elts();
+
+   long p = zz_p::modulus();
+   sp_reduce_struct pinv = H.pinv_L;
+
+
+   const double *vp = v.elts() + low;
+
+   NTL_AVX_LOCAL_ARRAY(res, double, 16);
+
+   long npanels = H.drow.length();
+
+   for (long i = 0, first = 0; i < npanels; i++, first += 16)  {
+      mul16rowsD(res, H.drow[i], vp, len);
+      long last = min(n, first + 16);
+      for (long ii = first; ii < last; ii++)
+         xp[ii].LoopHole() = rem((unsigned long) (long) res[ii-first], p, pinv);
+   }
+
+   x.normalize();
+}
+
+static
+void InnerProduct2_AVX(zz_pX& x, zz_pX& x_, const Vec<double>& v, long low, long low_, long len,
+                       const zz_pXAltArgument& H, long n)
+{
+   x.rep.SetLength(n);
+   zz_p *xp = x.rep.elts();
+
+   x_.rep.SetLength(n);
+   zz_p *xp_ = x_.rep.elts();
+
+   long p = zz_p::modulus();
+   sp_reduce_struct pinv = H.pinv_L;
+
+
+   const double *vp = v.elts() + low;
+   const double *vp_ = v.elts() + low_;
+
+   NTL_AVX_LOCAL_ARRAY(res, double, 16);
+   NTL_AVX_LOCAL_ARRAY(res_, double, 16);
+
+   long npanels = H.drow.length();
+
+   for (long i = 0, first = 0; i < npanels; i++, first += 16)  {
+      mul16rows2D(res, res_, H.drow[i], vp, vp_, len);
+      long last = min(n, first + 16);
+      for (long ii = first; ii < last; ii++) {
+         xp[ii].LoopHole() = rem((unsigned long) (long) res[ii-first], p, pinv);
+         xp_[ii].LoopHole() = rem((unsigned long) (long) res_[ii-first], p, pinv);
+      }
+   }
+
+   x.normalize();
+   x_.normalize();
+}
+
+static
+void CompMod_AVX(zz_pX& x, const zz_pX& g, const zz_pXAltArgument& A, 
+             const zz_pXModulus& F)
+{
+   if (deg(g) <= 0) {
+      x = g;
+      return;
+   }
+
+
+   zz_pX s, s_, t;
+
+   long m = A.m;
+   long l = ((g.rep.length()+m-1)/m) - 1;
+
+   zz_pXMultiplier M;
+   build(M, A.orig->H[m], F);
+
+   long len = g.rep.length();
+   Vec<double> gg;
+   gg.SetLength(len);
+   for (long i = 0; i < len; i++) gg[i] = rep(g.rep[i]);
+
+   InnerProduct_AVX(t, gg, l*m, l*m + m - 1, A, F.n);
+   long i = l-1;
+   for (; i >= 1; i -= 2) {
+      InnerProduct2_AVX(s, s_, gg, i*m, (i-1)*m, m, A, F.n);
+      MulMod(t, t, M, F);
+      add(t, t, s);
+      MulMod(t, t, M, F);
+      add(t, t, s_);
+   }
+
+   if (i >= 0) {
+      InnerProduct_AVX(s, gg, i*m, i*m + m - 1, A, F.n);
+      MulMod(t, t, M, F);
+      add(t, t, s);
+   }
+
+   x = t;
+}
+#endif
+
+
+
+#endif
+
+
+
+void CompMod(zz_pX& x, const zz_pX& g, const zz_pXAltArgument& A, 
+             const zz_pXModulus& F)
+{
+   if (!A.orig) LogicError("CompMod: uninitialized arg");
+
+#ifndef NTL_HAVE_LL_TYPE
+   CompMod(x, g, *A.orig, F);
+#else
+
+   switch (A.strategy) {
+   case 0: 
+      CompMod(x, g, *A.orig, F);
+      break;
+
+   case 1:
+      CompMod_L(x, g, A, F);
+      break;
+
+   case 2:
+      CompMod_LL(x, g, A, F);
+      break;
+
+#ifdef NTL_HAVE_AVX
+   case 3:
+      CompMod_AVX(x, g, A, F);
+      break;
+
+#endif
+
+   default:
+      LogicError("CompMod: bad strategy");
+   }
+#endif
+
+}
+
+
+
+// END zz_pXAltArgument variation 
+
+
+
+
 static void StripZeroes(vec_zz_p& x)
 {
    long n = x.length();
@@ -1119,7 +1637,7 @@ void UpdateMap(vec_zz_p& x, const vec_zz_p& aa,
    a = aa;
    StripZeroes(a);
 
-   if (a.length() > n) Error("UpdateMap: bad args");
+   if (a.length() > n) LogicError("UpdateMap: bad args");
    long i;
 
    if (!B.UseFFT) {
@@ -1153,8 +1671,10 @@ void ProjectPowers(vec_zz_p& x, const vec_zz_p& a, long k,
 {
    long n = F.n;
 
-   if (a.length() > n || k < 0 || NTL_OVERFLOW(k, 1, 0))
-      Error("ProjectPowers: bad args");
+   if (a.length() > n || k < 0)
+      LogicError("ProjectPowers: bad args");
+   if (NTL_OVERFLOW(k, 1, 0))
+      ResourceError("ProjectPowers: excessive args");
 
    long m = H.H.length()-1;
    long l = (k+m-1)/m - 1;
@@ -1184,7 +1704,7 @@ void ProjectPowers(vec_zz_p& x, const vec_zz_p& a, long k,
                    const zz_pX& h, const zz_pXModulus& F)
 
 {
-   if (a.length() > F.n || k < 0) Error("ProjectPowers: bad args");
+   if (a.length() > F.n || k < 0) LogicError("ProjectPowers: bad args");
 
    if (k == 0) {
       x.SetLength(0);
@@ -1290,8 +1810,8 @@ void GCDMinPolySeq(zz_pX& h, const vec_zz_p& x, long m)
 
 void MinPolySeq(zz_pX& h, const vec_zz_p& a, long m)
 {
-   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) Error("MinPoly: bad args");
-   if (a.length() < 2*m) Error("MinPoly: sequence too short");
+   if (m < 0 || NTL_OVERFLOW(m, 1, 0)) LogicError("MinPoly: bad args");
+   if (a.length() < 2*m) LogicError("MinPoly: sequence too short");
 
    if (m > NTL_zz_pX_BERMASS_CROSSOVER)
       GCDMinPolySeq(h, a, m);
@@ -1313,7 +1833,7 @@ void DoMinPolyMod(zz_pX& h, const zz_pX& g, const zz_pXModulus& F, long m,
 void ProbMinPolyMod(zz_pX& h, const zz_pX& g, const zz_pXModulus& F, long m)
 {
    long n = F.n;
-   if (m < 1 || m > n) Error("ProbMinPoly: bad args");
+   if (m < 1 || m > n) LogicError("ProbMinPoly: bad args");
 
    long i;
    vec_zz_p R(INIT_SIZE, n);
@@ -1326,7 +1846,7 @@ void MinPolyMod(zz_pX& hh, const zz_pX& g, const zz_pXModulus& F, long m)
 {
    zz_pX h, h1;
    long n = F.n;
-   if (m < 1 || m > n) Error("MinPoly: bad args");
+   if (m < 1 || m > n) LogicError("MinPoly: bad args");
 
    /* probabilistically compute min-poly */
 
@@ -1361,7 +1881,7 @@ void MinPolyMod(zz_pX& hh, const zz_pX& g, const zz_pXModulus& F, long m)
 void IrredPolyMod(zz_pX& h, const zz_pX& g, const zz_pXModulus& F, long m)
 {
    vec_zz_p R(INIT_SIZE, 1);
-   if (m < 1 || m > F.n) Error("IrredPoly: bad args");
+   if (m < 1 || m > F.n) LogicError("IrredPoly: bad args");
 
    set(R[0]);
    DoMinPolyMod(h, g, F, m, R);
@@ -1440,7 +1960,7 @@ void FFTMulTrunc(zz_pX& x, const zz_pX& a, const zz_pX& b, long n)
 
 void MulTrunc(zz_pX& x, const zz_pX& a, const zz_pX& b, long n)
 {
-   if (n < 0) Error("MulTrunc: bad args");
+   if (n < 0) LogicError("MulTrunc: bad args");
 
    if (deg(a) <= NTL_zz_pX_MUL_CROSSOVER || deg(b) <= NTL_zz_pX_MUL_CROSSOVER)
       PlainMulTrunc(x, a, b, n);
@@ -1477,7 +1997,7 @@ void FFTSqrTrunc(zz_pX& x, const zz_pX& a, long n)
 
 void SqrTrunc(zz_pX& x, const zz_pX& a, long n)
 {
-   if (n < 0) Error("SqrTrunc: bad args");
+   if (n < 0) LogicError("SqrTrunc: bad args");
 
    if (deg(a) <= NTL_zz_pX_MUL_CROSSOVER)
       PlainSqrTrunc(x, a, n);
@@ -1492,7 +2012,7 @@ void FastTraceVec(vec_zz_p& S, const zz_pX& f)
    long n = deg(f);
 
    if (n <= 0) 
-      Error("FastTraceVec: bad args");
+      LogicError("FastTraceVec: bad args");
 
    if (n == 0) {
       S.SetLength(0);
@@ -1534,7 +2054,7 @@ void FastTraceVec(vec_zz_p& S, const zz_pX& f)
 void PlainTraceVec(vec_zz_p& S, const zz_pX& ff)
 {
    if (deg(ff) <= 0)
-      Error("TraceVec: bad args");
+      LogicError("TraceVec: bad args");
 
    zz_pX f;
    f = ff;
@@ -1576,13 +2096,8 @@ void TraceVec(vec_zz_p& S, const zz_pX& f)
       FastTraceVec(S, f);
 }
 
-void ComputeTraceVec(const zz_pXModulus& F)
+void ComputeTraceVec(vec_zz_p& S, const zz_pXModulus& F)
 {
-   vec_zz_p& S = *((vec_zz_p *) &F.tracevec);
-
-   if (S.length() > 0)
-      return;
-
    if (!F.UseFFT) {
       PlainTraceVec(S, F.f);
       return;
@@ -1615,21 +2130,25 @@ void TraceMod(zz_p& x, const zz_pX& a, const zz_pXModulus& F)
    long n = F.n;
 
    if (deg(a) >= n)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
-   // FIXME: thread-safe-imple: need mutex here
+   do { // NOTE: thread safe lazy init
+      Lazy<vec_zz_p>::Builder builder(F.tracevec.val());
+      if (!builder()) break;
+      UniquePtr<vec_zz_p> p;
+      p.make();
+      ComputeTraceVec(*p, F);
+      builder.move(p);
+   } while (0);
 
-   if (F.tracevec.length() == 0) 
-      ComputeTraceVec(F);
-
-   InnerProduct(x, a.rep, F.tracevec);
+   InnerProduct(x, a.rep, *F.tracevec.val());
 }
 
 
 void TraceMod(zz_p& x, const zz_pX& a, const zz_pX& f)
 {
    if (deg(a) >= deg(f) || deg(f) <= 0)
-      Error("trace: bad args");
+      LogicError("trace: bad args");
 
    project(x, TraceVec(f), a);
 }
@@ -1943,7 +2462,7 @@ void resultant(zz_p& rres, const zz_pX& u, const zz_pX& v)
 void NormMod(zz_p& x, const zz_pX& a, const zz_pX& f)
 {
    if (deg(f) <= 0 || deg(a) >= deg(f)) 
-      Error("norm: bad args");
+      LogicError("norm: bad args");
 
    if (IsZero(a)) {
       clear(x);
