@@ -1,7 +1,6 @@
 
-
 #include <NTL/lzz_pXFactoring.h>
-#include <NTL/vec_vec_lzz_p.h>
+#include <NTL/mat_lzz_p.h>
 #include <NTL/FacVec.h>
 
 #include <NTL/new.h>
@@ -15,7 +14,7 @@ void SquareFreeDecomp(vec_pair_zz_pX_long& u, const zz_pX& ff)
    zz_pX f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("SquareFreeDecomp: bad args");
+      LogicError("SquareFreeDecomp: bad args");
 
    zz_pX r, t, v, tmp1;
    long m, j, finished, done;
@@ -81,7 +80,7 @@ void NullSpace(long& r, vec_long& D, vec_vec_zz_p& M, long verbose)
    for (j = 0; j < n; j++) D[j] = -1;
 
    long p = zz_p::modulus();
-   double pinv = zz_p::ModulusInverse();
+   mulmod_t pinv = zz_p::ModulusInverse();
    long T1, T2;
    mulmod_precon_t T1pinv;
 
@@ -117,7 +116,7 @@ void NullSpace(long& r, vec_long& D, vec_vec_zz_p& M, long verbose)
             t1 = M[i][k];
 
             T1 = rep(t1);
-            T1pinv = PrepMulModPrecon(T1, p, pinv); // ((double) T1)*pinv;
+            T1pinv = PrepMulModPrecon(T1, p, pinv); 
 
             x = M[i].elts() + (k+1);
             y = M[l].elts() + (k+1);
@@ -144,36 +143,27 @@ void NullSpace(long& r, vec_long& D, vec_vec_zz_p& M, long verbose)
 
 
 static
-void BuildMatrix(vec_vec_zz_p& M, 
+void BuildMatrix(mat_zz_p& M, 
                  long n, const zz_pX& g, const zz_pXModulus& F, long verbose)
 {
-   long i, j, m;
    zz_pXMultiplier G;
    zz_pX h;
 
-   M.SetLength(n);
-   for (i = 0; i < n; i++)
-      M[i].SetLength(n);
+   M.SetDims(n, n);
 
    build(G, g, F);
 
    set(h);
-   for (j = 0; j < n; j++) {
-      if (verbose && j % 10 == 0) cerr << "+";
+   for (long i = 0; i < n; i++) {
+      if (verbose && i % 10 == 0) cerr << "+";
 
-      m = deg(h);
-      for (i = 0; i < n; i++) {
-         if (i <= m)
-            M[i][j] = h.rep[i];
-         else
-            clear(M[i][j]);
-      }
+      VectorCopy(M[i], h, n);
 
-      if (j < n-1)
+      if (i < n-1)
          MulMod(h, h, G, F);
    }
 
-   for (i = 0; i < n; i++)
+   for (long i = 0; i < n; i++)
       add(M[i][i], M[i][i], -1);
 
 }
@@ -226,38 +216,16 @@ void FindRoots(vec_zz_p& x, const zz_pX& ff)
 
 
 static
-void RandomBasisElt(zz_pX& g, const vec_long& D, 
-                    const vec_vec_zz_p& M)
+void RandomBasisElt(zz_pX& g, mat_zz_p& ker)
 {
-   zz_p t1, t2;
+   long r = ker.NumRows();
+   long n = ker.NumCols();
 
-   long n = D.length();
-
-   long i, j, s;
-
-   g.rep.SetLength(n);
-
-   vec_zz_p& v = g.rep;
-
-   for (j = n-1; j >= 0; j--) {
-      if (D[j] == -1)
-         random(v[j]);
-      else {
-         i = D[j];
-
-         // v[j] = sum_{s=j+1}^{n-1} v[s]*M[i,s]
-
-         clear(t1);
-
-         for (s = j+1; s < n; s++) {
-            mul(t2, v[s], M[i][s]);
-            add(t1, t1, t2);
-         }
-
-         v[j] = t1;
-      }
-   }
+   vec_zz_p v;
+   v.SetLength(r);
+   for (long i = 0; i < r; i++) random(v[i]);
 
+   mul(g.rep, v, ker);
    g.normalize();
 }
 
@@ -356,7 +324,7 @@ void SFBerlekamp(vec_zz_pX& factors, const zz_pX& ff, long verbose)
    zz_pX f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("SFBerlekamp: bad args");
+      LogicError("SFBerlekamp: bad args");
 
    if (deg(f) == 0) {
       factors.SetLength(0);
@@ -387,20 +355,21 @@ void SFBerlekamp(vec_zz_pX& factors, const zz_pX& ff, long verbose)
    PowerXMod(g, p, F);
    if (verbose) { cerr << (GetTime()-t) << "\n"; }
 
-   vec_long D;
-   long r;
-
-   vec_vec_zz_p M;
+   mat_zz_p M, ker;
 
    if (verbose) { cerr << "building matrix..."; t = GetTime(); }
    BuildMatrix(M, n, g, F, verbose);
    if (verbose) { cerr << (GetTime()-t) << "\n"; }
 
    if (verbose) { cerr << "diagonalizing..."; t = GetTime(); }
-   NullSpace(r, D, M, verbose);
+   kernel(ker, M);
    if (verbose) { cerr << (GetTime()-t) << "\n"; }
 
 
+   M.kill();
+
+   long r = ker.NumRows();
+
    if (verbose) cerr << "number of factors = " << r << "\n";
 
    if (r == 1) {
@@ -413,9 +382,8 @@ void SFBerlekamp(vec_zz_pX& factors, const zz_pX& ff, long verbose)
 
    vec_zz_p roots;
 
-   RandomBasisElt(g, D, M);
+   RandomBasisElt(g, ker);
    MinPolyMod(h, g, F, r);
-   if (deg(h) == r) M.kill();
    FindRoots(roots, h);
    FindFactors(factors, f, g, roots);
 
@@ -425,7 +393,7 @@ void SFBerlekamp(vec_zz_pX& factors, const zz_pX& ff, long verbose)
 
    while (factors.length() < r) {
       if (verbose) cerr << "+";
-      RandomBasisElt(g, D, M);
+      RandomBasisElt(g, ker);
       S.kill();
       for (i = 0; i < factors.length(); i++) {
          const zz_pX& f = factors[i];
@@ -467,7 +435,7 @@ void berlekamp(vec_pair_zz_pX_long& factors, const zz_pX& f, long verbose)
    vec_zz_pX x;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("berlekamp: bad args");
+      LogicError("berlekamp: bad args");
 
    if (verbose) { cerr << "square-free decomposition..."; t = GetTime(); }
    SquareFreeDecomp(sfd, f);
@@ -559,7 +527,7 @@ void TraceMap(zz_pX& w, const zz_pX& a, long d, const zz_pXModulus& F,
               const zz_pX& b)
 
 {
-   if (d < 0) Error("TraceMap: bad args");
+   if (d < 0) LogicError("TraceMap: bad args");
 
    zz_pX y, z, t;
 
@@ -598,7 +566,7 @@ void TraceMap(zz_pX& w, const zz_pX& a, long d, const zz_pXModulus& F,
 
 void PowerCompose(zz_pX& y, const zz_pX& h, long q, const zz_pXModulus& F)
 {
-   if (q < 0) Error("PowerCompose: bad args");
+   if (q < 0) LogicError("PowerCompose: bad args");
 
    zz_pX z(INIT_SIZE, F.n);
    long sw;
@@ -674,7 +642,7 @@ long ProbIrredTest(const zz_pX& f, long iter)
    return !IsX(s);
 }
 
-NTL_THREAD_LOCAL long zz_pX_BlockingFactor = 10;
+NTL_CHEAP_THREAD_LOCAL long zz_pX_BlockingFactor = 10;
 
 void DDF(vec_pair_zz_pX_long& factors, const zz_pX& ff, const zz_pX& hh, 
          long verbose)
@@ -683,7 +651,7 @@ void DDF(vec_pair_zz_pX_long& factors, const zz_pX& ff, const zz_pX& hh,
    zz_pX h = hh;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("DDF: bad args");
+      LogicError("DDF: bad args");
 
    factors.SetLength(0);
 
@@ -794,7 +762,6 @@ void RecEDF(vec_zz_pX& factors, const zz_pX& f, const zz_pX& b, long d,
 {
    vec_zz_pX v;
    long i;
-   zz_pX bb;
 
    if (verbose) cerr << "+";
 
@@ -820,7 +787,7 @@ void EDF(vec_zz_pX& factors, const zz_pX& ff, const zz_pX& bb,
    zz_pX b = bb;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("EDF: bad args");
+      LogicError("EDF: bad args");
 
    long n = deg(f);
    long r = n/d;
@@ -859,7 +826,7 @@ void EDF(vec_zz_pX& factors, const zz_pX& ff, const zz_pX& bb,
 void SFCanZass1(vec_pair_zz_pX_long& u, zz_pX& h, const zz_pX& f, long verbose)
 {
    if (!IsOne(LeadCoeff(f)) || deg(f) == 0)
-      Error("SFCanZass1: bad args");
+      LogicError("SFCanZass1: bad args");
 
    double t;
 
@@ -925,7 +892,7 @@ void SFCanZass(vec_zz_pX& factors, const zz_pX& ff, long verbose)
    zz_pX f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("SFCanZass: bad args");
+      LogicError("SFCanZass: bad args");
 
    if (deg(f) == 0) {
       factors.SetLength(0);
@@ -997,7 +964,7 @@ void SFCanZass(vec_zz_pX& factors, const zz_pX& ff, long verbose)
 void CanZass(vec_pair_zz_pX_long& factors, const zz_pX& f, long verbose)
 {
    if (!IsOne(LeadCoeff(f)))
-      Error("CanZass: bad args");
+      LogicError("CanZass: bad args");
 
    double t;
    vec_pair_zz_pX_long sfd;
@@ -1211,10 +1178,10 @@ void FindRoot(zz_p& root, const zz_pX& ff)
    f = ff;
 
    if (!IsOne(LeadCoeff(f)))
-      Error("FindRoot: bad args");
+      LogicError("FindRoot: bad args");
 
    if (deg(f) == 0)
-      Error("FindRoot: bad args");
+      LogicError("FindRoot: bad args");
 
    p1 = zz_p::modulus() >> 1;
    h1 = 1;
@@ -1496,9 +1463,9 @@ void RecBuildIrred(zz_pX& f, long u, const FacVec& fvec)
 void BuildIrred(zz_pX& f, long n)
 {
    if (n <= 0)
-      Error("BuildIrred: n must be positive");
+      LogicError("BuildIrred: n must be positive");
 
-   if (NTL_OVERFLOW(n, 1, 0)) Error("overflow in BuildIrred");
+   if (NTL_OVERFLOW(n, 1, 0)) ResourceError("overflow in BuildIrred");
 
    if (n == 1) {
       SetX(f);
@@ -1531,11 +1498,12 @@ void BuildRandomIrred(zz_pX& f, const zz_pX& g)
 
 /************* NEW DDF ****************/
 
-NTL_THREAD_LOCAL long zz_pX_GCDTableSize = 4;
-NTL_THREAD_LOCAL static vec_zz_pX *BabyStepFile = 0;
-NTL_THREAD_LOCAL static vec_zz_pX *GiantStepFile = 0;
-NTL_THREAD_LOCAL static zz_pXArgument *HHH = 0;
-NTL_THREAD_LOCAL static long OldN = 0;
+NTL_CHEAP_THREAD_LOCAL long zz_pX_GCDTableSize = 4;
+static NTL_CHEAP_THREAD_LOCAL vec_zz_pX *BabyStepFile = 0;
+static NTL_CHEAP_THREAD_LOCAL vec_zz_pX *GiantStepFile = 0;
+static NTL_CHEAP_THREAD_LOCAL zz_pXArgument *HHH = 0;
+static NTL_CHEAP_THREAD_LOCAL zz_pXAltArgument *HHH1 = 0;
+static NTL_CHEAP_THREAD_LOCAL long OldN = 0;
 
 
 static
@@ -1550,7 +1518,6 @@ void GenerateBabySteps(zz_pX& h1, const zz_pX& f, const zz_pX& h, long k,
    build(F, f);
 
 
-   BabyStepFile = NTL_NEW_OP vec_zz_pX;
    (*BabyStepFile).SetLength(k-1);
 
    h1 = h;
@@ -1570,13 +1537,16 @@ void GenerateBabySteps(zz_pX& h1, const zz_pX& f, const zz_pX& h, long k,
    else {
       zz_pXArgument H;
       build(H, h, F, 2*rootn);
+
+      zz_pXAltArgument H1;
+      build(H1, H, F);
    
    
       for (i = 1; i <= k-1; i++) {
          (*BabyStepFile)(i) = h1; 
    
-         CompMod(h1, h1, H, F);
-         if (verbose) cerr << "+";
+         CompMod(h1, h1, H1, F);
+         if (verbose) cerr << ".";
       }
    }
    
@@ -1593,26 +1563,16 @@ void GenerateGiantSteps(const zz_pX& f, const zz_pX& h, long l, long verbose)
 
    build(F, f);
 
-   HHH = NTL_NEW_OP zz_pXArgument;
    build(*HHH, h, F, 2*SqrRoot(F.n));
+   build(*HHH1, *HHH, F);
 
    OldN = F.n;
 
-   GiantStepFile = NTL_NEW_OP vec_zz_pX;
    (*GiantStepFile).SetLength(1);
    (*GiantStepFile)(1) = h;
 }
 
 
-static
-void FileCleanup(long k, long l)
-{
-   delete BabyStepFile;
-   delete GiantStepFile;
-   delete HHH;
-}
-
-
 static
 void NewAddFactor(vec_pair_zz_pX_long& u, const zz_pX& g, long m, long verbose)
 {
@@ -1683,7 +1643,7 @@ void FetchGiantStep(zz_pX& g, long gs, const zz_pXModulus& F)
    zz_pX last;
 
    if (gs > l+1)
-      Error("bad arg to FetchGiantStep");
+      LogicError("bad arg to FetchGiantStep");
 
    if (gs == l+1) {
       last = (*GiantStepFile)(l);
@@ -1691,11 +1651,12 @@ void FetchGiantStep(zz_pX& g, long gs, const zz_pXModulus& F)
          rem(last, last, F);
          for (long i = 0; i < (*HHH).H.length(); i++)
             rem((*HHH).H[i], (*HHH).H[i], F);
+         build(*HHH1, *HHH, F);
          OldN = F.n;
       }
 
       (*GiantStepFile).SetLength(l+1);
-      CompMod((*GiantStepFile)(l+1), last, *HHH, F);
+      CompMod((*GiantStepFile)(l+1), last, *HHH1, F);
       g = (*GiantStepFile)(l+1);
    }
    else if (deg((*GiantStepFile)(gs)) >= F.n)
@@ -1910,7 +1871,7 @@ void NewDDF(vec_pair_zz_pX_long& factors,
 
 {
    if (!IsOne(LeadCoeff(f)))
-      Error("NewDDF: bad args");
+      LogicError("NewDDF: bad args");
 
    if (deg(f) == 0) {
       factors.SetLength(0);
@@ -1924,19 +1885,41 @@ void NewDDF(vec_pair_zz_pX_long& factors,
    }
 
    long B = deg(f)/2;
+   
    long k = SqrRoot(B);
+
+   // we double the number of baby steps if it seems like
+   // baby steps are significantly cheaper than giant steps.
+   // The calculations below are closely tied to a test in GenerateBabySteps:
+   // if nbm >= sdf/2, then scale should be 1 (baby steps and giant steps balanced)
+   if (B >= 500) {
+      long sdf = SqrRoot(deg(f));
+      long nbm = NumBits(zz_p::modulus());
+      double scale = 0.25*double(sdf)/double(nbm);
+      if (scale < 1) scale = 1;
+      if (scale > 2) scale = 2;
+      k = long(scale*k);
+   }
+
    long l = (B+k-1)/k;
 
+   vec_zz_pX local_BabyStepFile;
+   vec_zz_pX local_GiantStepFile;
+   zz_pXArgument local_HHH;
+   zz_pXAltArgument local_HHH1;
+
+   BabyStepFile = &local_BabyStepFile;
+   GiantStepFile = &local_GiantStepFile;
+   HHH = &local_HHH;
+   HHH1 = &local_HHH1;
+   
    zz_pX h1;
    GenerateBabySteps(h1, f, h, k, verbose);
-
    GenerateGiantSteps(f, h1, l, verbose);
 
    vec_pair_zz_pX_long u;
    GiantRefine(u, f, k, l, verbose);
    BabyRefine(factors, u, k, l, verbose);
-
-   FileCleanup(k, l);
 }
 
 NTL_END_IMPL
