@@ -12,7 +12,7 @@ void add(mat_ZZ_pE& X, const mat_ZZ_pE& A, const mat_ZZ_pE& B)
    long m = A.NumCols();  
   
    if (B.NumRows() != n || B.NumCols() != m)   
-      Error("matrix add: dimension mismatch");  
+      LogicError("matrix add: dimension mismatch");  
   
    X.SetDims(n, m);  
   
@@ -28,7 +28,7 @@ void sub(mat_ZZ_pE& X, const mat_ZZ_pE& A, const mat_ZZ_pE& B)
    long m = A.NumCols();  
   
    if (B.NumRows() != n || B.NumCols() != m)  
-      Error("matrix sub: dimension mismatch");  
+      LogicError("matrix sub: dimension mismatch");  
   
    X.SetDims(n, m);  
   
@@ -59,7 +59,7 @@ void mul_aux(mat_ZZ_pE& X, const mat_ZZ_pE& A, const mat_ZZ_pE& B)
    long m = B.NumCols();  
   
    if (l != B.NumRows())  
-      Error("matrix mul: dimension mismatch");  
+      LogicError("matrix mul: dimension mismatch");  
   
    X.SetDims(n, m);  
   
@@ -98,7 +98,7 @@ void mul_aux(vec_ZZ_pE& x, const mat_ZZ_pE& A, const vec_ZZ_pE& b)
    long l = A.NumCols();  
   
    if (l != b.length())  
-      Error("matrix mul: dimension mismatch");  
+      LogicError("matrix mul: dimension mismatch");  
   
    x.SetLength(n);  
   
@@ -134,7 +134,7 @@ void mul_aux(vec_ZZ_pE& x, const vec_ZZ_pE& a, const mat_ZZ_pE& B)
    long l = B.NumCols();  
   
    if (n != a.length())  
-      Error("matrix mul: dimension mismatch");  
+      LogicError("matrix mul: dimension mismatch");  
   
    x.SetLength(l);  
   
@@ -192,14 +192,16 @@ void determinant(ZZ_pE& d, const mat_ZZ_pE& M_in)
    n = M_in.NumRows();
 
    if (M_in.NumCols() != n)
-      Error("determinant: nonsquare matrix");
+      LogicError("determinant: nonsquare matrix");
 
    if (n == 0) {
       set(d);
       return;
    }
 
-   vec_ZZ_pX *M = NTL_NEW_OP vec_ZZ_pX[n];
+   UniqueArray<vec_ZZ_pX> M_store;
+   M_store.SetLength(n);
+   vec_ZZ_pX *M = M_store.get();
 
    for (i = 0; i < n; i++) {
       M[i].SetLength(n);
@@ -256,14 +258,11 @@ void determinant(ZZ_pE& d, const mat_ZZ_pE& M_in)
       }
       else {
          clear(d);
-         goto done;
+         return;
       }
    }
 
    conv(d, det);
-
-done:
-   delete[] M;
 }
 
 long IsIdent(const mat_ZZ_pE& A, long n)
@@ -317,16 +316,16 @@ void transpose(mat_ZZ_pE& X, const mat_ZZ_pE& A)
 }
    
 
-void solve(ZZ_pE& d, vec_ZZ_pE& X, 
-           const mat_ZZ_pE& A, const vec_ZZ_pE& b)
+static
+void solve_impl(ZZ_pE& d, vec_ZZ_pE& X, const mat_ZZ_pE& A, const vec_ZZ_pE& b, bool trans)
 
 {
    long n = A.NumRows();
    if (A.NumCols() != n)
-      Error("solve: nonsquare matrix");
+      LogicError("solve: nonsquare matrix");
 
    if (b.length() != n)
-      Error("solve: dimension mismatch");
+      LogicError("solve: dimension mismatch");
 
    if (n == 0) {
       set(d);
@@ -340,13 +339,23 @@ void solve(ZZ_pE& d, vec_ZZ_pE& X,
 
    const ZZ_pXModulus& p = ZZ_pE::modulus();
 
-   vec_ZZ_pX *M = NTL_NEW_OP vec_ZZ_pX[n];
+   UniqueArray<vec_ZZ_pX> M_store;
+   M_store.SetLength(n);
+   vec_ZZ_pX *M = M_store.get();
 
    for (i = 0; i < n; i++) {
       M[i].SetLength(n+1);
-      for (j = 0; j < n; j++) {
-         M[i][j].rep.SetMaxLength(2*deg(p)-1);
-         M[i][j] = rep(A[j][i]);
+      if (trans) {
+	 for (j = 0; j < n; j++) {
+	    M[i][j].rep.SetMaxLength(2*deg(p)-1);
+	    M[i][j] = rep(A[j][i]);
+	 }
+      }
+      else {
+	 for (j = 0; j < n; j++) {
+	    M[i][j].rep.SetMaxLength(2*deg(p)-1);
+	    M[i][j] = rep(A[i][j]);
+	 }
       }
       M[i][n].rep.SetMaxLength(2*deg(p)-1);
       M[i][n] = rep(b[i]);
@@ -400,7 +409,7 @@ void solve(ZZ_pE& d, vec_ZZ_pE& X,
       }
       else {
          clear(d);
-         goto done;
+         return;
       }
    }
 
@@ -416,16 +425,23 @@ void solve(ZZ_pE& d, vec_ZZ_pE& X,
    }
 
    conv(d, det);
+}
 
-done:
-   delete[] M;
+void solve(ZZ_pE& d, vec_ZZ_pE& x, const mat_ZZ_pE& A, const vec_ZZ_pE& b)
+{
+   solve_impl(d, x, A, b, true);
+}
+
+void solve(ZZ_pE& d, const mat_ZZ_pE& A, vec_ZZ_pE& x,  const vec_ZZ_pE& b)
+{
+   solve_impl(d, x, A, b, false);
 }
 
 void inv(ZZ_pE& d, mat_ZZ_pE& X, const mat_ZZ_pE& A)
 {
    long n = A.NumRows();
    if (A.NumCols() != n)
-      Error("inv: nonsquare matrix");
+      LogicError("inv: nonsquare matrix");
 
    if (n == 0) {
       set(d);
@@ -440,7 +456,9 @@ void inv(ZZ_pE& d, mat_ZZ_pE& X, const mat_ZZ_pE& A)
    const ZZ_pXModulus& p = ZZ_pE::modulus();
 
 
-   vec_ZZ_pX *M = NTL_NEW_OP vec_ZZ_pX[n];
+   UniqueArray<vec_ZZ_pX> M_store;
+   M_store.SetLength(n);
+   vec_ZZ_pX *M = M_store.get();
 
    for (i = 0; i < n; i++) {
       M[i].SetLength(2*n);
@@ -501,7 +519,7 @@ void inv(ZZ_pE& d, mat_ZZ_pE& X, const mat_ZZ_pE& A)
       }
       else {
          clear(d);
-         goto done;
+         return;
       }
    }
 
@@ -519,9 +537,6 @@ void inv(ZZ_pE& d, mat_ZZ_pE& X, const mat_ZZ_pE& A)
    }
 
    conv(d, det);
-
-done:
-   delete[] M;
 }
 
 
@@ -538,12 +553,14 @@ long gauss(mat_ZZ_pE& M_in, long w)
    long m = M_in.NumCols();
 
    if (w < 0 || w > m)
-      Error("gauss: bad args");
+      LogicError("gauss: bad args");
 
    const ZZ_pXModulus& p = ZZ_pE::modulus();
 
 
-   vec_ZZ_pX *M = NTL_NEW_OP vec_ZZ_pX[n];
+   UniqueArray<vec_ZZ_pX> M_store;
+   M_store.SetLength(n);
+   vec_ZZ_pX *M = M_store.get();
 
    for (i = 0; i < n; i++) {
       M[i].SetLength(m);
@@ -602,8 +619,6 @@ long gauss(mat_ZZ_pE& M_in, long w)
       for (j = 0; j < m; j++)
          conv(M_in[i][j], M[i][j]);
 
-   delete [] M;
-
    return l;
 }
 
@@ -832,12 +847,12 @@ void inv(mat_ZZ_pE& X, const mat_ZZ_pE& A)
 {
    ZZ_pE d;
    inv(d, X, A);
-   if (d == 0) Error("inv: non-invertible matrix");
+   if (d == 0) ArithmeticError("inv: non-invertible matrix");
 }
 
 void power(mat_ZZ_pE& X, const mat_ZZ_pE& A, const ZZ& e)
 {
-   if (A.NumRows() != A.NumCols()) Error("power: non-square matrix");
+   if (A.NumRows() != A.NumCols()) LogicError("power: non-square matrix");
 
    if (e == 0) {
       ident(X, A.NumRows());
