@@ -44,16 +44,25 @@ along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
+
+
+#ifdef __INTEL_COMPILER
+#pragma float_control(precise,on)
+#endif
+
+// NOTE: the above will force the Intel compiler to adhere to
+// language standards, which it does not do by default
+
 #include <NTL/quad_float.h>
 #include <NTL/RR.h>
 
-#include <float.h>
+#include <cfloat>
 
 #include <NTL/new.h>
 
 NTL_START_IMPL
 
-#if (defined(__GNUC__) && NTL_EXT_DOUBLE && (defined(__i386__) || defined(__i486__) || defined(__i586__)))
+#if (NTL_EXT_DOUBLE && defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__)))
 
 #if (!defined(NTL_X86_FIX) && !defined(NTL_NO_X86_FIX))
 
@@ -113,40 +122,36 @@ END_FIX
 
 
 #if (NTL_BITS_PER_LONG >= NTL_DOUBLE_PRECISION)
+
+
 quad_float to_quad_float(long n)
 {
-START_FIX
    DOUBLE xhi, xlo;
-   DOUBLE u, v;
 
-   xhi = double(n);
+   xhi = TrueDouble(n);
 
    // Because we are assuming 2's compliment integer
    // arithmetic, the following prevents long(xhi) from overflowing.
 
    if (n > 0)
-      xlo = double(n+long(-xhi));
+      xlo = TrueDouble(n+long(-xhi));
    else
-      xlo = double(n-long(xhi));
+      xlo = TrueDouble(n-long(xhi));
 
    // renormalize...just to be safe
-  
-   u = xhi + xlo;
-   v = xhi - u;
-   v = v + xlo;
-END_FIX
-   return quad_float(u, v);
+
+   quad_float z;
+   normalize(z, xhi, xlo);
+   return z;
 }
 
 quad_float to_quad_float(unsigned long n)
 {
-START_FIX
    DOUBLE xhi, xlo, t;
-   DOUBLE u, v;
 
    const double bnd = double(1L << (NTL_BITS_PER_LONG-2))*4.0;
 
-   xhi = double(n);
+   xhi = TrueDouble(n);
    
    if (xhi >= bnd)
       t = xhi - bnd;
@@ -155,19 +160,16 @@ START_FIX
 
    // we use the "to_long" function here to be as portable as possible.
    long llo = to_long(n - (unsigned long)(t));
-   xlo = double(llo);
+   xlo = TrueDouble(llo);
 
-   // renormalize...just to be safe
-
-   u = xhi + xlo;
-   v = xhi - u;
-   v = v + xlo;
-END_FIX
-   return quad_float(u, v);
+   quad_float z;
+   normalize(z, xhi, xlo);
+   return z;
 }
 #endif
 
 
+NTL_CHEAP_THREAD_LOCAL
 long quad_float::oprec = 10;
 
 void quad_float::SetOutputPrecision(long p)
@@ -175,7 +177,7 @@ void quad_float::SetOutputPrecision(long p)
    if (p < 1) p = 1;
 
    if (NTL_OVERFLOW(p, 1, 0)) 
-      Error("quad_float: output precision too big");
+      ResourceError("quad_float: output precision too big");
 
    oprec = p;
 }
@@ -351,33 +353,70 @@ END_FIX
 }
 
 
+
+#if (NTL_FMA_DETECTED)
+
+double quad_float_zero = 0;
+
+static inline
+double Protect(double x) { return x + quad_float_zero; }
+
+#else
+
+
+static inline
+double Protect(double x) { return x; }
+
+
+#endif
+
+// NOTE: this is really sick: some compilers will issue FMA
+// (fused mul add) instructions which will break correctness.
+// C99 standard is supposed to prevent this across separate
+// statements, but C++ standard doesn't guarantee much at all.
+// In any case, gcc does not even implement the C99 standard
+// correctly.  One could disable this by compiling with
+// an appropriate flag: -mno-fma works for gcc, while -no-fma works
+// for icc.  icc and MSVC++ also support pragmas to do this:
+// #pragma fp_contract(off).  There is also a compiler flag for
+// gcc: -ffp-contract=off, but -mno-fma seems more widely supported.
+// These flags work for clang, as well.
+//
+// But in any case, I'd rather not mess with getting these flags right.
+// By calling Protect(a*b), this has the effect of forcing the
+// compiler to compute a*b + 0.  Assuming the compiler otherwise
+// does not perform any re-association, this should do the trick.
+// There is a small performance penalty, but it should be reasonable.
+
+
+
 quad_float operator *(const quad_float& x,const quad_float& y ) {
 START_FIX
   DOUBLE hx, tx, hy, ty, C, c;
   DOUBLE t1, t2;
 
-  C = NTL_QUAD_FLOAT_SPLIT*x.hi;
+  C = Protect(NTL_QUAD_FLOAT_SPLIT*x.hi);
   hx = C-x.hi;
-  c = NTL_QUAD_FLOAT_SPLIT*y.hi;
+  c = Protect(NTL_QUAD_FLOAT_SPLIT*y.hi);
   hx = C-hx;
   tx = x.hi-hx;
   hy = c-y.hi;
-  C = x.hi*y.hi;
+  C = Protect(x.hi*y.hi);
   hy = c-hy;
   ty = y.hi-hy;
 
   // c = ((((hx*hy-C)+hx*ty)+tx*hy)+tx*ty)+(x.hi*y.lo+x.lo*y.hi);
   
-  t1 = hx*hy;
+  t1 = Protect(hx*hy);
   t1 = t1-C;
-  t2 = hx*ty;
+  t2 = Protect(hx*ty);
   t1 = t1+t2;
-  t2 = tx*hy;
+  t2 = Protect(tx*hy);
   t1 = t1+t2;
-  t2 = tx*ty;
+  t2 = Protect(tx*ty);
   c = t1+t2;
-  t1 = x.hi*y.lo;
-  t2 = x.lo*y.hi;
+  t1 = Protect(x.hi*y.lo);
+  t2 = Protect(x.lo*y.hi);
   t1 = t1+t2;
   c = c + t1;
 
@@ -395,28 +434,28 @@ START_FIX
   DOUBLE hx, tx, hy, ty, C, c;
   DOUBLE t1, t2;
 
-  C = NTL_QUAD_FLOAT_SPLIT*x.hi;
+  C = Protect(NTL_QUAD_FLOAT_SPLIT*x.hi);
   hx = C-x.hi;
-  c = NTL_QUAD_FLOAT_SPLIT*y.hi;
+  c = Protect(NTL_QUAD_FLOAT_SPLIT*y.hi);
   hx = C-hx;
   tx = x.hi-hx;
   hy = c-y.hi;
-  C = x.hi*y.hi;
+  C = Protect(x.hi*y.hi);
   hy = c-hy;
   ty = y.hi-hy;
 
   // c = ((((hx*hy-C)+hx*ty)+tx*hy)+tx*ty)+(x.hi*y.lo+x.lo*y.hi);
   
-  t1 = hx*hy;
+  t1 = Protect(hx*hy);
   t1 = t1-C;
-  t2 = hx*ty;
+  t2 = Protect(hx*ty);
   t1 = t1+t2;
-  t2 = tx*hy;
+  t2 = Protect(tx*hy);
   t1 = t1+t2;
-  t2 = tx*ty;
+  t2 = Protect(tx*ty);
   c = t1+t2;
-  t1 = x.hi*y.lo;
-  t2 = x.lo*y.hi;
+  t1 = Protect(x.hi*y.lo);
+  t2 = Protect(x.lo*y.hi);
   t1 = t1+t2;
   c = c + t1;
 
@@ -437,25 +476,25 @@ START_FIX
   DOUBLE t1;
 
   C = x.hi/y.hi;
-  c = NTL_QUAD_FLOAT_SPLIT*C;
+  c = Protect(NTL_QUAD_FLOAT_SPLIT*C);
   hc = c-C;
-  u = NTL_QUAD_FLOAT_SPLIT*y.hi;
+  u = Protect(NTL_QUAD_FLOAT_SPLIT*y.hi);
   hc = c-hc;
   tc = C-hc;
   hy = u-y.hi;
-  U = C * y.hi;
+  U = Protect(C * y.hi);
   hy = u-hy;
   ty = y.hi-hy;
 
   // u = (((hc*hy-U)+hc*ty)+tc*hy)+tc*ty;
 
-  u = hc*hy;
+  u = Protect(hc*hy);
   u = u-U;
-  t1 = hc*ty;
+  t1 = Protect(hc*ty);
   u = u+t1;
-  t1 = tc*hy;
+  t1 = Protect(tc*hy);
   u = u+t1;
-  t1 = tc*ty;
+  t1 = Protect(tc*ty);
   u = u+t1;
 
   // c = ((((x.hi-U)-u)+x.lo)-C*y.lo)/y.hi;
@@ -463,7 +502,7 @@ START_FIX
   c = x.hi-U;
   c = c-u;
   c = c+x.lo;
-  t1 = C*y.lo;
+  t1 = Protect(C*y.lo);
   c = c - t1;
   c = c/y.hi;
   
@@ -481,25 +520,25 @@ START_FIX
   DOUBLE t1;
 
   C = x.hi/y.hi;
-  c = NTL_QUAD_FLOAT_SPLIT*C;
+  c = Protect(NTL_QUAD_FLOAT_SPLIT*C);
   hc = c-C;
-  u = NTL_QUAD_FLOAT_SPLIT*y.hi;
+  u = Protect(NTL_QUAD_FLOAT_SPLIT*y.hi);
   hc = c-hc;
   tc = C-hc;
   hy = u-y.hi;
-  U = C * y.hi;
+  U = Protect(C * y.hi);
   hy = u-hy;
   ty = y.hi-hy;
 
   // u = (((hc*hy-U)+hc*ty)+tc*hy)+tc*ty;
 
-  u = hc*hy;
+  u = Protect(hc*hy);
   u = u-U;
-  t1 = hc*ty;
+  t1 = Protect(hc*ty);
   u = u+t1;
-  t1 = tc*hy;
+  t1 = Protect(tc*hy);
   u = u+t1;
-  t1 = tc*ty;
+  t1 = Protect(tc*ty);
   u = u+t1;
 
   // c = ((((x.hi-U)-u)+x.lo)-C*y.lo)/y.hi;
@@ -507,7 +546,7 @@ START_FIX
   c = x.hi-U;
   c = c-u;
   c = c+x.lo;
-  t1 = C*y.lo;
+  t1 = Protect(C*y.lo);
   c = c - t1;
   c = c/y.hi;
   
@@ -524,7 +563,7 @@ END_FIX
 
 quad_float sqrt(const quad_float& y) {
   if (y.hi < 0.0) 
-    Error("Quad: attempto to take square root of negative number");
+    ArithmeticError("quad_float: square root of negative number");
   if (y.hi == 0.0) return quad_float(0.0,0.0);
 
   double c;
@@ -536,18 +575,18 @@ START_FIX
   DOUBLE p,q,hx,tx,u,uu,cc;
   DOUBLE t1;
 
-  p = NTL_QUAD_FLOAT_SPLIT*c; 
+  p = Protect(NTL_QUAD_FLOAT_SPLIT*c); 
   hx = (c-p); 
   hx = hx+p; 
   tx = c-hx;
-  p = hx*hx;
-  q = hx*tx;
+  p = Protect(hx*hx);
+  q = Protect(hx*tx);
   q = q+q;
 
   u = p+q;
   uu = p-u;
   uu = uu+q;
-  t1 = tx*tx;
+  t1 = Protect(tx*tx);
   uu = uu+t1;
 
 
@@ -652,7 +691,7 @@ void conv(quad_float& z, const ZZ& a)
 
    // The following is just paranoia.
    if (fabs(z.hi) < NTL_FDOUBLE_PRECISION && z.lo != 0)
-      Error("internal error: ZZ to quad_float conversion");
+      LogicError("internal error: ZZ to quad_float conversion");
 } 
 
 void conv(ZZ& z, const quad_float& x)
@@ -688,46 +727,40 @@ ostream& operator<<(ostream& s, const quad_float& a)
       return s;
    }
 
-   long old_p = RR::precision();
-   long old_op = RR::OutputPrecision();
+   RRPush push;
+   RROutputPush opush;
 
    RR::SetPrecision(long(3.33*quad_float::oprec) + 10);
    RR::SetOutputPrecision(quad_float::oprec);
 
-   NTL_THREAD_LOCAL static RR t;
+   NTL_TLS_LOCAL(RR, t);
 
    conv(t, a);
    s << t;
 
-   RR::SetPrecision(old_p);
-   RR::SetOutputPrecision(old_op);
-
    return s;
 }
 
 istream& operator>>(istream& s, quad_float& x)
 {
-   long old_p = RR::precision();
+   RRPush push;
    RR::SetPrecision(4*NTL_DOUBLE_PRECISION);
 
-   NTL_THREAD_LOCAL static RR t;
-   s >> t;
+   NTL_TLS_LOCAL(RR, t);
+   NTL_INPUT_CHECK_RET(s, s >> t);
    conv(x, t);
 
-   RR::SetPrecision(old_p);
    return s;
 }
 
 void random(quad_float& x)
 {
-   long old_p = RR::precision();
+   RRPush push;
    RR::SetPrecision(4*NTL_DOUBLE_PRECISION);
 
-   NTL_THREAD_LOCAL static RR t;
+   NTL_TLS_LOCAL(RR, t);
    random(t);
    conv(x, t);
-
-   RR::SetPrecision(old_p);
 }
 
 quad_float random_quad_float()
@@ -814,14 +847,14 @@ quad_float fabs(const quad_float& x)
 quad_float to_quad_float(const char *s)
 {
    quad_float x;
-   long old_p = RR::precision();
+
+   RRPush push;
    RR::SetPrecision(4*NTL_DOUBLE_PRECISION);
 
-   NTL_THREAD_LOCAL static RR t;
+   NTL_TLS_LOCAL(RR, t);
    conv(t, s);
    conv(x, t);
 
-   RR::SetPrecision(old_p);
    return x;
 }
 
@@ -858,13 +891,11 @@ quad_float exp(const quad_float& x) { // New version 97 Aug 05
   if (x.hi<DBL_MIN_10_EXP*2.302585092994045684017991) 
     return to_quad_float(0.0);
   if (x.hi>DBL_MAX_10_EXP*2.302585092994045684017991) {
-    Error("exp(quad_float): overflow");
+    ResourceError("exp(quad_float): overflow");
   }
 
-  // changed this from "const" to "static" in v5.3, since "const"
-  // causes the initialization to be performed with *every* invocation.
-  NTL_THREAD_LOCAL static quad_float Log2 = 
-    to_quad_float("0.6931471805599453094172321214581765680755");
+  static const quad_float Log2 = to_quad_float("0.6931471805599453094172321214581765680755");
+  // GLOBAL (assumes C++11 thread-safe init)
 
   quad_float y,temp,ysq,sum1,sum2;
   long iy;
@@ -891,7 +922,7 @@ quad_float exp(const quad_float& x) { // New version 97 Aug 05
 
 quad_float log(const quad_float& t) { // Newton method. See Bailey, MPFUN
   if (t.hi <= 0.0) {
-    Error("log(quad_float): argument must be positive");
+    ArithmeticError("log(quad_float): argument must be positive");
   }
   double s1 = log(t.hi);
   ForceToMem(&s1);  // Again, this is fairly paranoid.
@@ -917,3 +948,4 @@ long operator!=(const quad_float& x, const quad_float& y)
 
 
 NTL_END_IMPL
+
